/*   utils.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-2000 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include "tweak.h"
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef DES_ENCRYPTION
#include "des.h"
#endif
#include "commondef.h"

void* alloc(void* buf, long size)
{ if (buf) return realloc(buf,size); else return malloc(size);
}

int strip_crlf(char* buffer)
{ int   n;
  char* ptr;

  n = 0;
  ptr = buffer + strlen(buffer) - 1;
  while ((ptr > buffer) && ((*ptr == '\n') || (*ptr == '\r')))
  { ptr--;
    n++;
  }
  *(ptr+1) = '\0';
  return n;
}

char skipspace(char** x)
{ while (**x && isspace(**x)) (*x)++;
  return **x;
}

char skipnspace(char** x)
{ while (**x && !isspace(**x)) (*x)++;
  return **x;
}

char skiptocolon(char** x)
{ skipspace(x);
  while (**x && !isspace(**x) && (**x!=':')) (*x)++;
  return **x;
}

char skipspctocolon(char** x)
{ while (**x && isspace(**x) && (**x!=':')) (*x)++;
  if (**x == ':')
  { (*x)++;
    skipspace(x);
  }
  return **x;
}

int readline(char* line, int linelen, FILE* file)
{ char* r;

  r = fgets(line,linelen,file);
  if (r) strip_crlf(line);
  return (r != NULL);
}

/* creates filename relative to server_dir */
extern char server_dir[];
void mk_filename(char* buf, char* filename, char* defaultname)
{ if (filename == NULL) filename = defaultname;
  if (filename != NULL)
  { if (*filename == '/') strcpy(buf,filename); else
    { strcpy(buf,server_dir);
      strcat(buf,filename);
    }
  } else strcpy(buf,server_dir);
}

/* read with timeout */
int read_to(int fd, char* buf, int buflen, int timeout)
{ fd_set         fdset;
  int            i;
  struct timeval to;

  to.tv_sec = timeout;
  to.tv_usec = 0;
  FD_ZERO(&fdset);
  FD_SET(fd,&fdset);
  i = select(fd+1,&fdset,NULL,NULL,timeout ? &to : NULL);
  if (i > 0) i = (int)read(fd,buf,buflen); else
  { if (!i)
    { errno = ETIMEDOUT;
      i = -1;
    }
    *buf = '\0';
  }
  return i;
}

/* fgets with timeout */
char* fgets_to(char* buf, int buflen, FILE* stream, int timeout)
{ int            fd;
  fd_set         fdset;
  char*          ptr;
  struct timeval to;

  fd = fileno(stream);
  to.tv_sec = timeout;
  to.tv_usec = 0;
  FD_ZERO(&fdset);
  FD_SET(fd,&fdset);
  fd = select(fd+1,&fdset,NULL,NULL,timeout ? &to : NULL);
  if (fd > 0) ptr = fgets(buf,buflen,stream); else
  { if (!fd) errno = ETIMEDOUT;
    *buf = '\0';
    ptr = NULL;
  }
  return ptr;
}

/* set flag */
void flag_set(int* flags, int n, int value)
{ if (value)
  { *flags |= (1 << n);
  } else
  { *flags &= ~(1 << n);
  }
}

/* get flag */
int flag_get(int flags, int n)
{ return (flags & (1 << n)) ? 1 : 0;
}

/* accept with timeout */
int accept_to(int s, struct sockaddr *addr, int *addrlen, int timeout)
{ fd_set         fdset;
  int            n;
  struct timeval to;

  to.tv_sec = timeout;
  to.tv_usec = 0;
  FD_ZERO(&fdset);
  FD_SET(s,&fdset);
  n = select(s+1,&fdset,NULL,NULL,timeout ? &to : NULL);
  if (!n)
  { errno = ETIMEDOUT;
    n  = -1;
  } else if (n > 0) n = accept(s,addr,addrlen);
  return n;
}

/* connect with timeout */
int connect_to(int s, struct sockaddr *addr, int addrlen, int timeout)
{ fd_set         fdset;
  int            flags;
  int            n;
  struct timeval to;

  if ((flags = fcntl(s,F_GETFL)) >= 0) fcntl(s,F_SETFL,flags | O_NONBLOCK);
  n = connect(s,addr,addrlen);
  if ((n < 0) && (errno == EINPROGRESS))
  { to.tv_sec = timeout;
    to.tv_usec = 0;
    FD_ZERO(&fdset);
    FD_SET(s,&fdset);
    n = select(s+1,NULL,&fdset,NULL,timeout ? &to : NULL);
    if (!n)
    { errno = ETIMEDOUT;
      n = -1;
    } else if (n > 0) n = 0;
  }
  if (flags >= 0) fcntl(s,F_SETFL,flags);
  return n;
}

/* EOF */
