/*   ftpa.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-1999 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include "tweak.h"
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include "defines.h"
#include "rfc931.h"
#include "utils.h"

char server_dir[256];
int  sig_array[sig_size];
int  sig_read  = 0;
int  sig_write = 0;

void signal_handler(int sig)
{ if (sig == SIGTERM) exit(CL_KILL);
  if (((sig_write==sig_size-1) && !sig_read) || (sig_write+1==sig_read))
    return;
  sig_array[sig_write] = sig;
  if (++sig_write >= sig_size) sig_write = 0;
}

void handle_signals(void)
{ while (sig_read != sig_write)
  { switch (sig_array[sig_read])
    { case SIGPIPE :
        exit(CL_PIPEERR);
    }
    if (++sig_read >= sig_size) sig_read = 0;
  }
}

int main(int argc, char** argv, char** env)
{ int                bouncer;
  char               buffer[4096];
  fd_set             fdset;
  char               ident[32];
  unsigned long      ipaddr;
  struct sockaddr_in loc_addr;
  int                n;
  char*              p;
  struct sockaddr_in rem_addr;
#ifndef HAVE_SIGSET
  struct sigaction   sig_act;
#endif
  int                sock;
  int                timeout;

#ifndef HAVE_SIGSET
  sigfillset(&sig_act.sa_mask);
  sig_act.sa_flags = SA_RESTART;
#define sigset(x,y) sig_act.sa_handler = y; sigaction(x,&sig_act,NULL)
#endif
  sigset(SIGHUP,signal_handler);
  sigset(SIGINT,signal_handler);
  sigset(SIGQUIT,signal_handler);
  sigset(SIGALRM,signal_handler);
  sigset(SIGPIPE,signal_handler);
  sigset(SIGTERM,signal_handler);
  sigset(SIGUSR1,signal_handler);
  sigset(SIGUSR2,signal_handler);
  setvbuf(stdout,NULL,_IOLBF,0);

  /* evaluate environment */
  p = getenv("SOCK");
  if (!p) return CL_ERROR;
  sock = atoi(p);
  p = getenv("BOUNCER");
  bouncer = p ? atoi(p) : 0;
  p = getenv("BOUNCERTIMEOUT");
  timeout = p ? atoi(p) : 30;

  /* bouncer and ident check 0x01=bouncer, 0x02=ident */
  *ident = '\0';
  ipaddr = 0;
  if (bouncer & 0x01)
  { /* bouncer handling, bnc4all & rftp compatible
       bnc4all syntax: AUTH [<user>]@<ip>[ <anything_else_ignored>]
       rftp syntax:    IDNT [<user>]@<ip>
     */
    n = read_to(sock,buffer,sizeof(buffer),timeout);
    if (n)
    { p = strchr(buffer,'@');
      if (p)
      { ipaddr = ntohl(inet_addr(p+1));
        *p = '\0';
        p = strchr(buffer,' ');
        if (p)
        { strncpy(ident,p+1,sizeof(ident));
          ident[sizeof(ident)-1] = '\0';
        }
      }
      if (!p)
      { strcpy(buffer,"-AUTH not received\r\n");
        send(sock,buffer,strlen(buffer),0);
      }
    }
  }
  if ((bouncer & 0x02) && !*ident)
  { n = sizeof(loc_addr);
    if (getsockname(sock,(struct sockaddr*)&loc_addr,&n) < 0) bouncer = 0;
    n = sizeof(rem_addr);
    if (getpeername(sock,(struct sockaddr*)&rem_addr,&n) < 0) bouncer = 0;
    if (bouncer)
    { strncpy(ident,rfc931(&rem_addr,&loc_addr),sizeof(ident));
      ident[sizeof(ident)-1] = '\0';
    }
  }
  if (!*ident) strcpy(ident,"UNKNOWN");

  /* send ident@ip to ftpd for check, exit if check fails */
  printf("IDT_CHK %lu %s\n",ipaddr,ident);
  fgets(buffer,sizeof(buffer),stdin);
  if (*buffer != '+') return CL_IDENT;

  /* send startup message and enter command loop */
  FD_ZERO(&fdset);
  do
  { FD_SET(0,&fdset);
    FD_SET(sock,&fdset);
    n = select(sock+1,&fdset,NULL,NULL,NULL);
    if (n > 0)
    { if (FD_ISSET(0,&fdset))
      { n = read(0,buffer,sizeof(buffer));
        if (n > 0) n = send(sock,buffer,n,0);
      }
      if ((n > 0) && FD_ISSET(sock,&fdset))
      { n = recv(sock,buffer,sizeof(buffer),0);
        if (n > 0) n = write(1,buffer,n);
      }
    } else
    { if (errno == EINTR) handle_signals();
        else return CL_ERROR;
    }
  } while (n);

  /* not reached */
  return 0;
}

/* EOF */
