/*   definesd.h
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-1999 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "commondef.h"

struct user_s
{ int32     uid;            /* user ID                              */
  int32     gid;            /* group ID                             */
  int32     gids[8];        /* additional groups                    */
  int32     class;          /* user class                           */
  int64     rights;         /* user rights                          */
  int32     limit;          /* limit (ratio)                        */
  int64     credit;         /* credit in bytes                      */
  int32     dl_speed;       /* min and max download speed in kb/s   */
  int32     ul_speed;       /* min and max upload speed   in kb/s   */
  int64     dl_week;        /* download current week                */
  int64     ul_week;        /* upload current week                  */
  int64     dl_month;       /* download current month               */
  int64     ul_month;       /* upload current month                 */
  int64     dl_year;        /* download current year                */
  int64     ul_year;        /* upload current year                  */
  int64     dl_total;       /* download total                       */
  int64     ul_total;       /* upload total                         */
  int64     dl_tm_week;     /* download time (s) current week       */
  int64     ul_tm_week;     /* upload time (s) current week         */
  int64     dl_tm_month;    /* download time (s) current month      */
  int64     ul_tm_month;    /* upload time (s) current month        */
  int64     dl_tm_year;     /* download time (s) current year       */
  int64     ul_tm_year;     /* upload time (s) current year         */
  int64     dl_tm_total;    /* download time (s) total              */
  int64     ul_tm_total;    /* upload time (s) total                */
  time_t    last_login;     /* time of last login                   */
  char      name[10];       /* username                             */
  char      password[40];   /* encrypted password                   */
  char      info[32];       /* info                                 */
  char      idents[10][20]; /* 12bytes for ident; 8byte for iprange */
} ;

struct user_v223_s
{ char   name[10];
  long   uid;
  long   gid;
  char   password[14];
  long   kb_downloaded;
  long   kb_uploaded;
  time_t last_login;
  char   idents[10][20]; /* 12bytes for ident; 8byte for iprange */
  int    limit;
  char   info[16];
  long   credit;
  long   gids[8];
  long   class;
  long   kb_down_tmp;
  long   kb_up_tmp;
} ;

struct group_v3009
{ int32     gid;            /* group ID                             */
  int32     limit;          /* limit (ratio)                        */
  int64     credit;         /* credit in bytes                      */
  int32     dl_speed;       /* min and max download speed in kb/s   */
  int32     ul_speed;       /* min and max upload speed   in kb/s   */
  int32     logins;         /* max number of logins                 */
  char      name[10];       /* group name                           */
  char      info[32];       /* group information                    */
  char      ident[20];      /* 12bytes for ident; 8byte for iprange */
} ;

struct group_s
{ int32     gid;            /* group ID                             */
  int32     limit;          /* limit (ratio)                        */
  int64     credit;         /* credit in bytes                      */
  int32     dl_speed;       /* min and max download speed in kb/s   */
  int32     ul_speed;       /* min and max upload speed   in kb/s   */
  int32     logins;         /* max number of logins                 */
  char      name[10];       /* group name                           */
  char      info[32];       /* group information                    */
  char      ident[20];      /* 12bytes for ident; 8byte for iprange */
  int	    umask_group;    /* group based umask                    */
} ;

struct child_s
{ time_t         act_time;         /* time of last activity   */
  char           activity[160];    /* last activity           */
  int            admin;            /* flag: admin port        */
  int            bouncer;          /* 0x01=bouncer 0x02=ident */
  char           buffer[1024];     /* command buffer          */
  int            buffer_pos;       /* position in buffer      */
  int            buffer_overrun;   /* flag: buffer overrun    */
  int64          dl;               /* bytes downloaded        */
  int64          dl_tm;            /* download time (s)       */
  int            fd;               /* fd of server socket     */
  int            hidden;           /* hidden activity: visible to superuser only */
  int            idle;             /* flag: idle              */
  unsigned long  ip;               /* logged in from this ip  */
  char		 ident[12];        /* ident                   */
  int            lim_bw;           /* limit: bandwidth        */
  int            lim_direction;    /* limit: direction        */
  int            lim_session;      /* limit: anon or user     */
  pid_t          pid;              /* process id              */
  FILE*          rdp;              /* read pipe               */
  int            sock;             /* socket number           */
  time_t         time;             /* login time              */
  int64          ul;               /* bytes uploaded          */
  int64          ul_tm;            /* upload time (s)         */
  struct user_s* user;             /* user record             */
  FILE*          wrp;              /* write pipe              */
} ;

struct shandler_s
{ int   fd;                      /* file descriptor to listen on */
  void* ptr;                     /* handler-specific pointer     */
  int   (*handler)(int,void*);   /* service handler              */
} ;

struct config_s
{ char  key[32];    /* key                     */
  int   type;       /* type (bool, int, char)  */
  void* value;      /* pointer to value        */
  int   min;        /* minimum value           */
  int   max;        /* maximum value           */
  void* dflvalue;   /* default value           */
} ;

struct limit_s
{ int sday;        /* start day of week                */
  int shour;       /* start hour                       */
  int smin;        /* start minute                     */
  int eday;        /* end day of week                  */
  int ehour;       /* end hour                         */
  int emin;        /* end minute                       */
  int max_anon;    /* number of anon                   */
  int max_anon_dom;/* number of anon per domain        */
  int max_anon_ip; /* number of anon per ip            */
  int max_user;    /* number of users                  */
  int max_sum;     /* flag: sum of anon & user         */
  int sp_dl;       /* download speed limit             */
  int sp_ul;       /* upload speed limit               */
  int sp_sum;      /* flag: sum of upload and download */
  int sp_dl_fact;  /* download speed factor            */
  int sp_ul_fact;  /* upload speed factor              */
  int sp_dl_avg;   /* average download speed           */
  int sp_dl_dbl;   /* double download speed            */
  int sp_dl_half;  /* half download speed              */
  int sp_ul_avg;   /* average upload speed             */
  int sp_ul_dbl;   /* double upload speed              */
  int sp_ul_half;  /* half upload speed                */
} ;

struct watch_s
{ struct sockaddr_in addr;   /* destination address */
  struct child_s*    child;  /* connected client    */
} ;

#define sig_size 16

#define STATE_MASK               0x0F
#define STATE_CLOSE              0x01
#define STATE_OPEN               0x00
#define STATE_SHUTDOWN           0x02
#define STATE_FLAG_MASK          0xF0
#define STATE_FLAG_NORMAL        0x00
#define STATE_FLAG_TRANSACTIONAL 0x10
#define STATE_FLAG_IMMEDIATE     0x20

#if (INTSIZE >= 8)

/* constants for 64 bit processors */

#define R_STATE             0x00000001
#define R_PORTS             0x00000002
#define R_GUI               0x00000004
#define R_MONITOR           0x00000080
#define R_FILEACCESS        0x00000010
#define R_DUMP_FILES        0x00000020
#define R_REGISTRY          0x00000040
#define R_MESSAGES          0x00000080
#define R_CONFIG            0x00000100

#define R_GROUP_ALL         0x000F0000
#define R_GROUP_ADD         0x00010000
#define R_GROUP_EDIT        0x00020000
#define R_GROUP_DELETE      0x00040000
#define R_GROUP_LIST        0x00080000

#define R_USER_ALL          0x7FFFFF700000000
#define R_USER_DISABLE      0x000000100000000
#define R_USER_KICK         0x000000200000000
#define R_USER_PASSWD       0x000000400000000
#define R_USER_ADD          0x000001000000000
#define R_USER_EDIT         0x000002000000000
#define R_USER_DELETE       0x000004000000000
#define R_USER_READ         0x000008000000000
#define R_USER_ADD_SO       0x000010000000000
#define R_USER_EDIT_SO      0x000020000000000
#define R_USER_DELETE_SO    0x000040000000000
#define R_USER_READ_SO      0x000080000000000
#define R_USER_ADD_GO       0x000100000000000
#define R_USER_EDIT_GO      0x000200000000000
#define R_USER_DELETE_GO    0x000400000000000
#define R_USER_READ_GO      0x000800000000000
#define R_USER_ADD_ANON     0x001000000000000
#define R_USER_ADD_TMP      0x002000000000000
#define R_USER_EDIT_IP      0x004000000000000
#define R_USER_EDIT_GRPS    0x008000000000000
#define R_USER_EDIT_CREDIT  0x010000000000000
#define R_USER_EDIT_STATS   0x020000000000000
#define R_USER_EDIT_RIGHTS  0x040000000000000
#define R_USER_EDIT_INFO    0x080000000000000
#define R_USER_EDIT_UID     0x100000000000000
#define R_USER_EDIT_LOGINS  0x200000000000000
#define R_USER_EDIT_SPEEDS  0x400000000000000
#define R_ALL               0x7FFFFFFFFFFFFFF

#else

/* same constants for 32 bit processors */

#define R_STATE             0x00000001LL
#define R_PORTS             0x00000002LL
#define R_GUI               0x00000004LL
#define R_MONITOR           0x00000008LL
#define R_FILEACCESS        0x00000010LL
#define R_DUMP_FILES        0x00000020LL
#define R_REGISTRY          0x00000040LL
#define R_MESSAGES          0x00000080LL
#define R_CONFIG            0x00000100LL

#define R_GROUP_ALL         0x000F0000LL
#define R_GROUP_ADD         0x00010000LL
#define R_GROUP_EDIT        0x00020000LL
#define R_GROUP_DELETE      0x00040000LL
#define R_GROUP_LIST        0x00080000LL

#define R_USER_ALL          0x7FFFFF700000000LL
#define R_USER_DISABLE      0x000000100000000LL
#define R_USER_KICK         0x000000200000000LL
#define R_USER_PASSWD       0x000000400000000LL
#define R_USER_ADD          0x000001000000000LL
#define R_USER_EDIT         0x000002000000000LL
#define R_USER_DELETE       0x000004000000000LL
#define R_USER_READ         0x000008000000000LL
#define R_USER_ADD_SO       0x000010000000000LL
#define R_USER_EDIT_SO      0x000020000000000LL
#define R_USER_DELETE_SO    0x000040000000000LL
#define R_USER_READ_SO      0x000080000000000LL
#define R_USER_ADD_GO       0x000100000000000LL
#define R_USER_EDIT_GO      0x000200000000000LL
#define R_USER_DELETE_GO    0x000400000000000LL
#define R_USER_READ_GO      0x000800000000000LL
#define R_USER_ADD_ANON     0x001000000000000LL
#define R_USER_ADD_TMP      0x002000000000000LL
#define R_USER_EDIT_IP      0x004000000000000LL
#define R_USER_EDIT_GRPS    0x008000000000000LL
#define R_USER_EDIT_CREDIT  0x010000000000000LL
#define R_USER_EDIT_STATS   0x020000000000000LL
#define R_USER_EDIT_RIGHTS  0x040000000000000LL
#define R_USER_EDIT_INFO    0x080000000000000LL
#define R_USER_EDIT_UID     0x100000000000000LL
#define R_USER_EDIT_LOGINS  0x200000000000000LL
#define R_USER_EDIT_SPEEDS  0x400000000000000LL
#define R_ALL               0x7FFFFFFFFFFFFFFLL

#endif

/* EOF */
