/*   f4adp_group.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-2000 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *                       
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "externsd.h"
#include "ftpd.h"
#include "group.h"
#include "serverd.h"
#include "user.h"
#include "utils.h"

/* group add */
int f4adp_grp_add(struct child_s* c, char** p, int n)
{ struct group_s* g;
  struct group_s  grp;

  memset(&grp,0,sizeof(grp));
  if ((strlen(p[0]) >= sizeof(grp.name)) || !sscanf(p[1],SI32,&grp.gid))
  { server_reply(c,"-Syntax error");
    return 0;
  }
  if (get_group_by_name(p[0]))
  { server_reply(c,"-Group already exists");
    return 0;
  }
  if (!strcmp(p[1],"-0")) for (n=0, g=group; n<group_count; n++, g++)
    if (g->gid >= grp.gid) grp.gid = g->gid + 1;
  strcpy(grp.name,p[0]);
  if (group_count == group_size)
  { group_size += 8;
    g = (struct group_s*)alloc(group,group_size*sizeof(struct group_s));
    if (!g)
    { group_size -= 8;
      server_reply(c,"-Out of memory");
      return 0;
    }
    group = g;
  }
  group[group_count++] = grp;
  group_dirty++;
  server_reply(c,"+Group added");
  return 0;
}

/* group change */
int f4adp_grp_chg(struct child_s* c, char** p, int n)
{ struct group_s* g;
  struct group_s  g_new;
  int32           i;
  int             j;
  int64           l;
  char*           q;
  int             se = 0;

  g = get_group_by_name(p[0]);
  if (!g)
  { server_reply(c,"-Group does not exist");
    return 0;
  }
  g_new = *g;
  for (j=1; j<n; j++)
  { q = strchr(p[j],'=');
    if (q)
    { *(q++) = '\0';
      if (!strcasecmp(p[j],"NAME"))
      { if (strlen(q) >= sizeof(g_new.name)) se = 1; else
        { if (strcasecmp(g_new.name,q)) if (get_group_by_name(q))
          { server_reply(c,"-Group already exists");
            return 0;
          }
          strcpy(g_new.name,q);
        }
      } else if (!strcasecmp(p[j],"GID"))
      { if (!sscanf(q,SI32,&i)) se = 1; else g_new.gid = i;
      } else if (!strcasecmp(p[j],"LIMIT"))
      { if (!sscanf(q,SI32,&i)) se = 1; else g_new.limit = i;
      } else if (!strcasecmp(p[j],"CREDIT"))
      { if (!sscanf(q,SI64,&l)) se = 1; else g_new.credit = l * 1024 * 1024;
      } else if (!strcasecmp(p[j],"DL_MAX"))
      { if (!sscanf(q,SI32,&i)) se = 1; else
          g_new.dl_speed = (g_new.dl_speed & 0xFFFF) | (i << 16);
      } else if (!strcasecmp(p[j],"DL_MIN"))
      { if (!sscanf(q,SI32,&i)) se = 1; else
          g_new.dl_speed = (g_new.dl_speed & 0xFFFF0000) | i;
      } else if (!strcasecmp(p[j],"DL_SPEED"))
      { if (!sscanf(q,SI32,&i)) se = 1; else g_new.dl_speed = i;
      } else if (!strcasecmp(p[j],"UL_MAX"))
      { if (!sscanf(q,SI32,&i)) se = 1; else
          g_new.ul_speed = (g_new.ul_speed & 0xFFFF) | (i << 16);
      } else if (!strcasecmp(p[j],"UL_MIN"))
      { if (!sscanf(q,SI32,&i)) se = 1; else
          g_new.ul_speed = (g_new.ul_speed & 0xFFFF0000) | i;
      } else if (!strcasecmp(p[j],"UL_SPEED"))
      { if (!sscanf(q,SI32,&i)) se = 1; else g_new.ul_speed = i;
      } else if (!strcasecmp(p[j],"UMASK"))
      { if (!sscanf(q,SI32,&i)) se = 1; else g_new.umask_group = i;
      } else if (!strcasecmp(p[j],"LOGINS"))
      { if (!sscanf(q,SI32,&i)) se = 1; else g_new.logins = i;
      } else if (!strcasecmp(p[j],"INFO"))
      { if (strlen(q) >= sizeof(g_new.info)) se = 1; else
          strcpy(g_new.info,strcmp(q,"-") ? q : "");
      } else if (!strcasecmp(p[j],"IDENT"))
      { if (get_iprange(&q,g_new.ident)) se = 1;
      } else se = 1;
    } else se = 1;
    if (se) j = n;
  }
  if (se)
  { server_reply(c,"-Syntax error");
    return 0;
  }
  *g = g_new;
  group_dirty++;
  server_reply(c,"+Group changed");
  return 0;
}

/* group to string */
void sprintf_group(char* str, struct group_s* g)
{ sprintf(str,"+%s\t"SI32"\t"SI32"\t"SI64"\t"SI32"\t"SI32"\t"SI32"\t"SI32"\t%s\t",
    g->name,g->gid,g->limit,g->credit,g->dl_speed,g->ul_speed,g->umask_group,g->logins,
    g->info[0] ? g->info : "-");
  print_ident(str+strlen(str),g->ident);
}

/* group get */
int f4adp_grp_get(struct child_s* c, char** p, int n)
{ struct group_s* g;
  char            str[256];

  g = get_group_by_name(p[0]);
  if (!g)
  { server_reply(c,"-Group does not exist");
    return 0;
  }
  sprintf_group(str,g);
  server_reply(c,str);
  return 0;
}

int f4adp_grp_ls(struct child_s* c, char** p, int n)
{ struct group_s* g;
  char            str[256];

  for (n=0, g=group; n<group_count; n++, g++)
  { sprintf_group(str,g);
    server_reply(c,str);
  }
  server_reply(c,"+");
  return 0;
}

/* group list restricted */
int f4adp_grp_lsr(struct child_s* c, char** p, int n)
{ struct group_s* g;
  char            str[64];

  sprintf(str,"%i",group_count);
  server_reply(c,str);
  if (!group_count) return 0;
  if (!fgets_to(str,sizeof(str),c->rdp,5)) return 0;
  if (*str != '+') return 0;
  for (n=0, g=group; n<group_count; n++, g++)
  { sprintf(str,"%s "SI32,g->name,g->gid);
    server_reply(c,str);
  }
  return 0;
}

/* group move down */
int f4adp_grp_mdn(struct child_s* c, char** p, int n)
{ struct group_s* g;
  struct group_s  grp;

  if (!allowed(c->user,R_GROUP_ADD,NULL) && !allowed(c->user,R_GROUP_EDIT,NULL))
  { server_reply(c,"-Permission denied");
    return 0;
  }
  g = get_group_by_name(p[0]);
  if (!g)
  { server_reply(c,"-Group does not exist");
    return 0;
  }
  if (g == group+group_count-1)
  { server_reply(c,"-This is the last group");
    return 0;
  }
  grp    = *g;
  *g     = *(g+1);
  *(g+1) = grp;
  group_dirty++;
  server_reply(c,"+Group moved down");
  return 0;
}

/* group move up */
int f4adp_grp_mup(struct child_s* c, char** p, int n)
{ struct group_s* g;
  struct group_s  grp;

  if (!allowed(c->user,R_GROUP_ADD,NULL) && !allowed(c->user,R_GROUP_EDIT,NULL))
  { server_reply(c,"-Permission denied");
    return 0;
  }
  g = get_group_by_name(p[0]);
  if (!g)
  { server_reply(c,"-Group does not exist");
    return 0;
  }
  if (g == group)
  { server_reply(c,"+This is the first group");
    return 0;
  }
  grp    = *g;
  *g     = *(g-1);
  *(g-1) = grp;
  group_dirty++;
  server_reply(c,"+Group moved up");
  return 0;
}

/* group remove */
int f4adp_grp_rm(struct child_s* c, char** p, int n)
{ struct group_s* g;

  g = get_group_by_name(p[0]);
  if (!g)
  { server_reply(c,"-Group does not exist");
    return 0;
  }
  memmove(g,g+1,((--group_count)-(g-group))*sizeof(struct group_s));
  group_dirty++;
  server_reply(c,"+Group removed");
  return 0;
}

/* group set */
int f4adp_grp_set(struct child_s* c, char** p, int n)
{ int64           credit;
  int32           dl_speed;
  struct group_s* g;
  int32           gid;
  int32           gumask;
  char            ident[sizeof(g->ident)];
  int32           limit;
  int32           logins;
  int32           ul_speed;

  if ((strlen(p[1]) >= sizeof(g->name)) || !sscanf(p[2],SI32,&gid)
      || !sscanf(p[3],SI32,&limit) || !sscanf(p[4],SI64,&credit)
      || !sscanf(p[5],SI32,&dl_speed) || !sscanf(p[6],SI32,&ul_speed)
      || !sscanf(p[7],SI32,&gumask) || !sscanf(p[8],SI32,&logins)
      || (strlen(p[9]) >= sizeof(g->info)) || get_iprange(p+10,ident))
  { server_reply(c,"-Syntax error");
    return 0;
  }
  g = get_group_by_name(p[0]);
  if (!g)
  { server_reply(c,"-Group does not exist");
    return 0;
  }
  if (strcasecmp(p[0],p[1])) if (get_group_by_name(p[1]))
  { server_reply(c,"-Group already exists");
    return 0;
  }
  strcpy(g->name,p[1]);
  if (strcmp(p[2],"-0")) g->gid = gid;
  g->limit    = limit;
  g->credit   = credit;
  g->dl_speed = dl_speed;
  g->ul_speed = ul_speed;
  g->umask_group = gumask;
  g->logins   = logins;
  strcpy(g->info,strcmp(p[9],"-") ? p[9] : "");
  memmove(g->ident,ident,sizeof(g->ident));
  group_dirty++;
  server_reply(c,"+Group changed");
  return 0;
}

/* EOF */
