/*   f4adp_message.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-1999 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *                       
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "externsd.h"
#include "ftpd.h"
#include "log.h"
#include "serverd.h"
#include "utils.h"

void tolowerstr(char* str)
{ while (*str) *(str++) = tolower(*str);
}

int get_msg_filename(char* filename, char** p)
{ char* q  = filename;
  int   se = 0;

  strcpy(filename,server_dir);
  switch (p[0][0])
  { case 'd' : strcat(filename,DFL_MSGDIR);
               q = filename + strlen(filename);
               break;
    case 'g' : if (p[0][1] != ':') se = 1; else
               { strcat(filename,DFL_MSGGDIR);
                 mkdir(filename,0777);
                 q = filename + strlen(filename);
                 strcat(filename,p[0]+2);
                 strcat(filename,"/");
               }
               break;
    case 'u' : if (p[0][1] != ':') se = 1; else
               { strcat(filename,DFL_MSGUDIR);
                 mkdir(filename,0777);
                 q = filename + strlen(filename);
                 strcat(filename,p[0]+2);
                 strcat(filename,"/");
               }
               break;
    default  : se = 1;
  }
  if (!se)
  { tolowerstr(q);
    mkdir(filename,0777);
    strcat(filename,p[1]);
    tolowerstr(q);
  }
  return se;
}

/* get available messages */
int f4adp_msg_avl(struct child_s* c, char** p, int n)
{ struct dirent* d;
  DIR*           dir;
  char           filename[256];
  int            l;
  char*          name;
  int            se;
  struct stat    st;
  char           str[512];

  p[1] = "";
  se = get_msg_filename(filename,p);
  if (se)
  { server_reply(c,"-Syntax error");
    return 0;
  }
  strcpy(str,"+");
  n = 1;
  dir = opendir(filename);
  if (dir)
  { name = filename + strlen(filename);
    *(name++) = '/';
    while ((d = readdir(dir)) != NULL) if (d->d_name[0] != '.')
    { strcpy(name,d->d_name);
      if (stat(filename,&st) < 0)
      { st.st_mode |= S_IFDIR;
        errlog("stat(%s): %s (%i)",filename,strerror(errno),errno);
      }
      if (!S_ISDIR(st.st_mode))
      { l = strlen(d->d_name);
        if (n+1+l < sizeof(str))
        { str[n++] = ' ';
          strcpy(str+n,d->d_name);
          n += l;
        } else break;
      }
    }
    closedir(dir);
  }
  server_reply(c,str);
  return 0;
}

/* get default message names */
int f4adp_msg_dfl(struct child_s* c, char** p, int n)
{ int  se = 0;
  char str[256];

  if (strlen(p[0]) > 1) se = 1; else switch (p[0][0])
  { case 'd' :
      sprintf(str,"+cdup chdir download goodbye list mkdir rmdir startup "
        "toomany upload welcome");
      break;
    case 'g' :
    case 'u' :
      sprintf(str,"+cdup chdir download goodbye list mkdir rmdir "
        "toomany upload welcome");
      break;
    default  : se = 1;
  }
  if (se)
  { server_reply(c,"-Syntax error");
    return 0;
  }
  server_reply(c,str);
  return 0;
}

/* get message */
int f4adp_msg_get(struct child_s* c, char** p, int n)
{ FILE*       file = NULL;
  char        filename[256];
  int         se;
  struct stat st;
  char        str[256];

  se = get_msg_filename(filename,p);
  if (se)
  { server_reply(c,"-Syntax error");
    return 0;
  }
  if (stat(filename,&st) == -1)
  { se = 1;
    sprintf(str,"+ -1");
  } else
  { file = fopen(filename,"rt");
    if (!file) st.st_size = 0;
    sprintf(str,"+ %u",(unsigned int)st.st_size);
    se = (st.st_size == 0);
  }
  server_reply(c,str);
  if (se) return 0;
  while (st.st_size > 0)
  { if (!fgets(str,sizeof(str),file))
    { se = (st.st_size >= sizeof(str)) ? sizeof(str)-1 : st.st_size;
      memset(str,'x',se-1);
      str[se++] = '\n';
      str[se] = 0;
    }
    st.st_size -= strlen(str);
    fprintf(c->wrp,"%s",str);
  }
  fflush(c->wrp);
  fclose(file);
  return 0;
}

/* remove message */
int f4adp_msg_rm(struct child_s* c, char** p, int n)
{ char filename[256];
  int  se;
  char str[256];

  se = get_msg_filename(filename,p);
  if (se)
  { server_reply(c,"-Syntax error");
    return 0;
  }
  if (unlink(filename) < 0) strcpy(str,"-Remove message failed");
    else strcpy(str,"+Message removed");
  server_reply(c,str);
  return 0;
}

/* set message */
int f4adp_msg_set(struct child_s* c, char** p, int n)
{ FILE*       file;
  char        filename[256];
  int         se;
  int         size;
  char        str[256];

  se = get_msg_filename(filename,p);
  if (!sscanf(p[2],"%i",&size)) se = 1;
  if (se)
  { server_reply(c,"-Syntax error");
    return 0;
  }
  strcat(filename,".tmp");
  file = fopen(filename,"wb");
  if (!file)
  { server_reply(c,"-Cannot open file");
    return 0;
  }
  server_reply(c,"+Ok");
  while (size > 0)
  { if ((se = read_to(fileno(c->rdp),str,size > sizeof(str) ?
      sizeof(str) : size,5)) <= 0) break;
    size -= se;
    fwrite(str,1,se,file);
  }
  fclose(file);
  if (size > 0) unlink(filename); else
  { strcpy(str,filename);
    str[strlen(str)-4] = 0;
    if (rename(filename,str) < 0)
      errlog("rename(%s,%s): %s (%i)",filename,str,strerror(errno),errno);
  }
  return 0;
}

/* EOF */
