/*   f4adp_server.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-2000 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *                       
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include "externsd.h"
#include "ftpd.h"
#include "serverd.h"
#include "user.h"

void port_change(int* so, int* po, int* sn, int *pn, int force)
{ struct child_s* ch;
  int             i;

  if (!*pn) return;
  if (*so >= 0)
  { if (!force) for (i=0, ch=child; i<children; i++, ch++)
      if (ch->fd == *so) return;
    close(*so);
    shandler_del(*so);
  }
  *so = *sn;
  *sn = -1;
  *po = *pn == -1 ? 0 : *pn;
  *pn = 0;
}

int f4adp_port_get(struct child_s* c, char** p, int n)
{ char str[128];

  sprintf(str,"+%i %i %i",
    admin_port_new ? (admin_port_new == -1 ? 0 : admin_port_new) : admin_port,
    port_new ? (port_new == -1 ? 0 : port_new) : port,
    port2_new ? (port2_new == -1 ? 0 : port2_new) : port2);
  server_reply(c,str);
  return 0;
}

int f4adp_port_set(struct child_s* c, char** p, int n)
{ struct sockaddr_in addr;
  int                busy;
  int                prt;
  int                se;
  int                sock;
  char               tmpstr[512];

  se = ((strlen(p[0]) != 1) || !sscanf(p[1],"%i",&prt));
  p[0][0] = tolower(p[0][0]);
  if (!se)
  { busy = 0;
    switch (p[0][0])
    { case 'a' :
        if (admin_port_new) busy = 1;
        break;
      case 'p' :
        if (port_new) busy = 1;
        break;
      case 's' :
        if (port2_new) busy = 1;
        break;
      default  :
        se = 1;
    }
    if (!se && !busy)
    { if (prt)
      { sock = get_socket(SOCK_STREAM,prt,p[0][0] == 'a' ? 4 : 16,&addr);
        if (sock < 0)
        { sprintf(tmpstr,"-Unable to get socket: %s",strerror(errno));
          server_reply(c,tmpstr);
          return 0;
        }
        if (!shandler_add(sock,NULL,srv_accept))
        { close(sock);
          server_reply(c,"-Unable to register service handler");
          return 0;
        }
      } else
      { prt = -1;
        sock = -1;
      }
      switch (p[0][0])
      { case 'a' :
          admin_port_new = prt;
          admin_socket_new = sock;
          port_change(&admin_socket,&admin_port,&admin_socket_new,
            &admin_port_new,1);
          break;
        case 'p' :
          port_new = prt;
          ss_new = sock;
          port_change(&ss,&port,&ss_new,&port_new,1);
          break;
        case 's' :
          port2_new = prt;
          ss2_new = sock;
          port_change(&ss2,&port2,&ss2_new,&port2_new,1);
          break;
      }
    }
  }
  if (se)
  { server_reply(c,"-Syntax error");
    return 0;
  } else if (busy)
  { server_reply(c,"-Last port change is still pending");
    return 0;
  }
  server_reply(c,prt == -1 ? "+Port unbound" : "+Port changed");
  return 0;
}

int f4adp_server_get_state(struct child_s* c, char** p, int n)
{ char str[128];

  sprintf(str,"+%i",state);
  server_reply(c,str);
  return 0;
}

int f4adp_server_set_state(struct child_s* c, char** p, int n)
{ struct child_s* ch;
  int             i;
  int             nstate = 0;
  int             se = 0;
  int             sno;
  char            str[128];

  if (strlen(p[0]) != 2) se = 1; else
  { switch (tolower(p[0][0]))
    { case 'c' : nstate = STATE_CLOSE;    break;
      case 'o' : nstate = STATE_OPEN;     break;
      case 's' : nstate = STATE_SHUTDOWN; break;
      default  : se = 1;
    }
    switch (tolower(p[0][1]))
    { case 'i' : nstate |= STATE_FLAG_IMMEDIATE;     break;
      case 'n' : nstate |= STATE_FLAG_NORMAL;        break;
      case 't' : nstate |= STATE_FLAG_TRANSACTIONAL; break;
      default  : se = 1;
    }
    if (!se)
    { sno = -1;
      if (nstate == (STATE_SHUTDOWN | STATE_FLAG_IMMEDIATE))
      { raise(SIGTERM);
        server_reply(c,"+Server shutdown initiated");
        sleep(5);
        return 0;
      }
      switch (nstate & STATE_MASK)
      { case STATE_CLOSE    :
          sprintf(str,"+Server closed");
          if ((nstate & STATE_FLAG_MASK) == STATE_FLAG_IMMEDIATE)
            sno = SIGTERM;
          else if ((nstate & STATE_FLAG_MASK) == STATE_FLAG_TRANSACTIONAL)
            sno = SIGINT;
          break;
        case STATE_OPEN     :
          nstate = STATE_OPEN;
          sprintf(str,"+Server opened");
          if ((state & STATE_MASK) == STATE_SHUTDOWN)
            strcat(str,", shutdown cancelled");
          sno = SIGUSR1;
          break;
        case STATE_SHUTDOWN :
          sprintf(str,"+Server shutdown initiated");
          if ((nstate & STATE_FLAG_MASK) == STATE_FLAG_TRANSACTIONAL)
            sno = SIGHUP;
          break;
      }
      if (sno != -1) for (i=0, ch=child; i<children; i++, ch++)
      { if (ch->user && (ch->pid != -1))
          if (!allowed(ch->user,R_STATE,NULL)) kill(ch->pid,sno);
      }
      state = nstate;
    }
  }
  if (se)
  { server_reply(c,"-Syntax error");
    return 0;
  }
  server_reply(c,str);
  return 0;
}

/* EOF */
