/*   f4adp_user.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-2000 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *                       
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <ctype.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "externsd.h"
#include "ftpd.h"
#include "group.h"
#include "serverd.h"
#include "user.h"
#include "utils.h"

/* user add/copy */
int f4adp_usr_add(struct child_s* c, char** p, int n)
{ struct group_s* g = NULL;
  int             i;
  struct user_s*  u;
  struct user_s   usr;

  if (strlen(p[0]) >= sizeof(usr.name))
  { server_reply(c,"-Syntax error");
    return 0;
  }
  if (get_user_by_name(p[0]))
  { server_reply(c,"-User already exists");
    return 0;
  }
  if (n == 2)
  { u = get_user_by_name(p[1]);
    if (!u) if (!groupop(c->user)) g = get_group_by_name(p[1]);
    if (!u && !g)
    { server_reply(c,"-Template does not exist");
      return 0;
    }
  } else u = NULL;
  if (u)
  { usr = *u;
    usr.class &= ~UC_MASK;
  } else
  { memset(&usr,0,sizeof(usr));
    for (i=0; i<sizeof(usr.gids)/sizeof(int32); i++) usr.gids[i] = -1;
  }
  if (g || groupop(c->user))
  { if (!g) g = get_group_by_gid(c->user->gid);
    if (!g)
    { server_reply(c,"-Group does not exist");
      return 0;
    }
    usr.gid      = g->gid;
    usr.limit    = g->limit;
    usr.credit   = g->credit;
    usr.dl_speed = g->dl_speed;
    usr.ul_speed = g->ul_speed;
    usr.class    = g->logins << 8;
    strcpy(usr.info,g->info);
    memmove(usr.idents[0],g->ident,sizeof(g->ident));
  } else if (!u)
  { usr.gid = -1;
    usr.class = 0x0100;
  }
  strcpy(usr.name,p[0]);
  usr.uid = get_new_uid();
  usr.class |= UC_NORMAL;
  if (!insert_user(&usr))
  { server_reply(c,"-Out of memory");
    return 0;
  }
  server_reply(c,"+User added");
  top_ok = 0;
  return 0;
}

/* user add quick */
int f4adp_usr_adq(struct child_s* c, char** p, int n)
{ struct group_s* g;
  int 	          i;
  char            salt[2];
  struct user_s   usr;

  /* Name */
  if (strlen(p[0]) >= sizeof(usr.name))
  { server_reply(c,"-Syntax error (name)");
    return 0;
  }
  if (get_user_by_name(p[0]))
  { server_reply(c,"-User already exists");
    return 0;
  }
  memset(&usr,0,sizeof(usr));
  memset(usr.gids,-1,sizeof(usr.gids));
  strcpy(usr.name,p[0]);
  /* Group */
  if (groupop(c->user)) g = get_group_by_gid(c->user->gid);
    else g = get_group_by_name(p[1]);
  if (!g)
  { server_reply(c,"-Group does not exist");
    return 0;
  }
  usr.gid      = g->gid;
  usr.limit    = g->limit;
  usr.credit   = g->credit;
  usr.dl_speed = g->dl_speed;
  usr.ul_speed = g->ul_speed;
  usr.class    = g->logins << 8;
  strcpy(usr.info,g->info);
  memmove(usr.idents[0],g->ident,sizeof(g->ident));
  /* Password */
  salt[0] = salt_a[time(NULL) & 0x3F];
  salt[1] = salt_a[(time(NULL) >> 6) & 0x3F];
  strcpy(usr.password,crypt(p[2],salt));
  /* Limit */
  if (allowed(c->user,R_USER_EDIT_CREDIT,NULL))
  { if (!strcasecmp(p[3],"leech")) usr.limit = 0;
    else if (sscanf(p[3],"1:%i",&i)) usr.limit = 100 / i;
    else if (sscanf(p[3],"%i",&i)) usr.limit = i;
    else
    { server_reply(c,"-Syntax error (limit)");
      return 0;
    }
  }
  /* IPs */
  for (i=0; i<(n-4); i++) if (get_iprange(p+4+i,usr.idents[i]) != 0) 
  { server_reply(c,"-Syntax error (IP)");
    return 0;
  }
  /* Others */
  usr.uid = get_new_uid();
  usr.class |= UC_NORMAL;
  if (!insert_user(&usr))
  { server_reply(c,"-Out of memory");
    return 0;
  }
  server_reply(c,"+User added");
  top_ok = 0;
  return 0;
}

/* user change */
int f4adp_usr_chg(struct child_s* c, char** p, int n)
{ struct child_s* ch;    /* children (aux)          */
  int             i;     /* integer (aux)           */
  int32           j;     /* integer (aux)           */
  int64           l;     /* integer (aux)           */
  int             pd;    /* flag: permission denied */
  int             rf;    /* range full (for ip+)    */
  char*           q;     /* argument after "="      */
  int             se;    /* flag: syntax error      */
  struct user_s*  u;     /* user to change          */
  struct user_s   usr;   /* user to change (temp)   */
  char            salt[2]; /* for PASSWD set        */

  u = get_user_by_name(p[0]);
  if (!u)
  { server_reply(c,"-User does not exist");
    return 0;
  }
  pd = !allowed(c->user,R_USER_EDIT,u);
  if (pd) n = 1; else usr = *u;
  se = 0;
  rf = 0;
  while (--n)
  { p++;
    q = strchr(*p,'=');
    if (q)
    { *(q++) = '\0';
      if (!strcasecmp(*p,"NAME"))
      { if (strlen(q) >= sizeof(usr.name)) se = 1; else
        { if (get_user_by_name(q))
          { server_reply(c,"-User already exists");
            return 0;
          }
          strcpy(usr.name,q);
        }
      } else if (!strcasecmp(*p,"UID"))
      { if (!allowed(c->user,R_USER_EDIT_UID,u)) pd = 1; else
        if (!sscanf(q,SI32,&usr.uid)) se = 1;
      } else if (!strcasecmp(*p,"GID"))
      { if (!allowed(c->user,R_USER_EDIT_GRPS,u)) pd = 1; else
        if (!sscanf(q,SI32,&usr.gid)) se = 1;
      } else if (!strcasecmp(*p,"IPRANGE") || !strcasecmp(*p,"IP0"))
      { if (!allowed(c->user,R_USER_EDIT_IP,u)) pd = 1; else
        if (get_iprange(&q,usr.idents[0]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP1"))
      { if (!allowed(c->user,R_USER_EDIT_IP,u)) pd = 1; else
        if (get_iprange(&q,usr.idents[1]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP2"))
      { if (!allowed(c->user,R_USER_EDIT_IP,u)) pd = 1; else
        if (get_iprange(&q,usr.idents[2]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP3"))
      { if (!allowed(c->user,R_USER_EDIT_IP,u)) pd = 1; else
        if (get_iprange(&q,usr.idents[3]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP4"))
      { if (!allowed(c->user,R_USER_EDIT_IP,u)) pd = 1; else
        if (get_iprange(&q,usr.idents[4]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP5"))
      { if (!allowed(c->user,R_USER_EDIT_IP,u)) pd = 1; else
        if (get_iprange(&q,usr.idents[5]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP6"))
      { if (!allowed(c->user,R_USER_EDIT_IP,u)) pd = 1; else
        if (get_iprange(&q,usr.idents[6]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP7"))
      { if (!allowed(c->user,R_USER_EDIT_IP,u)) pd = 1; else
        if (get_iprange(&q,usr.idents[7]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP8"))
      { if (!allowed(c->user,R_USER_EDIT_IP,u)) pd = 1; else
        if (get_iprange(&q,usr.idents[8]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP9"))
      { if (!allowed(c->user,R_USER_EDIT_IP,u)) pd = 1; else
        if (get_iprange(&q,usr.idents[9]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP+"))
      { if (!allowed(c->user,R_USER_EDIT_IP,u)) pd = 1; else
        for (i=0; i<10; i++)
        { if (!memcmp(usr.idents[i]+12,"\0\0\0\0\0\0\0\0",8))
          { if (get_iprange(&q,usr.idents[i]) != 0) se = 1;
            i = 10; rf = 2;
          }
 /* finixing the IP-range full bug
  * that existed with IP+
  * apparently there was a bug in the line below
  * rf was staying at 2 or 1 accordngly without EVER being set to 0
  * therefore IP+ always gave IP_ranbge full even if hte user had
  * no ip's added
  * Hopefull this will fix it
  */
        } if ((i == 10) && (rf == 2)) rf = 1; else rf =0;
      } else if (!strcasecmp(*p,"IP-"))
      { if (!allowed(c->user,R_USER_EDIT_IP,u)) pd = 1; else
        if (sscanf(q,SI32,&i))
        { if ((i >= 0) && (i <= 9))
          { memset(usr.idents[i],0,sizeof(usr.idents[0]));
          } else se = 1;
        } else se = 1;
      } else if (!strcasecmp(*p,"LIMIT"))
      { if (!allowed(c->user,R_USER_EDIT_CREDIT,u)) pd = 1; else
        if (!sscanf(q,SI32,&usr.limit)) se = 1;
      } else if (!strcasecmp(*p,"INFO"))
      { if (!allowed(c->user,R_USER_EDIT_INFO,u)) pd = 1; else
        if (strlen(q) < sizeof(usr.info)) strcpy(usr.info,q); else se = 1;
      } else if (!strcasecmp(*p,"CREDIT"))
      { if (!allowed(c->user,R_USER_EDIT_CREDIT,u)) pd = 1; else
        if (!sscanf(q,SI64,&usr.credit)) se = 1;
      } else if (!strcasecmp(*p,"CREDIT+"))
      { if (!allowed(c->user,R_USER_EDIT_CREDIT,u)) pd = 1; else
        if (sscanf(q,SI64,&l)) usr.credit += l; else se = 1;
      } else if (!strcasecmp(*p,"CREDIT-"))
      { if (!allowed(c->user,R_USER_EDIT_CREDIT,u)) pd = 1; else
        if (sscanf(q,SI64,&l)) usr.credit -= l; else se = 1;
      } else if (!strcasecmp(*p,"MBCRED"))
      { if (!allowed(c->user,R_USER_EDIT_CREDIT,u)) pd = 1; else
        if (sscanf(q,SI64,&l)) usr.credit = l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"MBCRED+"))
      { if (!allowed(c->user,R_USER_EDIT_CREDIT,u)) pd = 1; else
        if (sscanf(q,SI64,&l)) usr.credit += l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"MBCRED-"))
      { if (!allowed(c->user,R_USER_EDIT_CREDIT,u)) pd = 1; else
        if (sscanf(q,SI64,&l)) usr.credit -= l * 0x100000; else se = 1;
      }
/***************************************************************************** 
 * some additions to user change so the user can edit 
 * MBUP / MBDOWN WKUP/WKDOWN MONTHUP/MONTHDOWN 
 * etc in user change and also so the nuker that currently exists
 * can work with only a slight modification
 * I assume that the siteops who give speed edit to 
 * their staff want it to enable em to change stats
 * and not just the speed(doesn't make much sense otherwise)
 * of course an R_USER_EDIT_STATS variable can be added in later stages
 *****************************************************************************/
        else if(!strcasecmp(*p,"MBUP+"))
      {  if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
        if (sscanf(q,SI64,&l)) usr.ul_total+=l * 0x100000; else se = 1;
      } else if(!strcasecmp(*p,"MBUP-"))
      {  if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
        if (sscanf(q,SI64,&l)) usr.ul_total-=l * 0x100000; else se = 1;
      } else if(!strcasecmp(*p,"MBUP"))
      {  if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
        if (sscanf(q,SI64,&l)) usr.ul_total=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"MBDOWN+"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.dl_total+=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"MBDOWN-"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.dl_total-=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"MBDOWN"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.dl_total=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"WKUP-"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.ul_week-=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"WKDOWN-"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.dl_week-=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"WKUP+"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.ul_week+=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"WKDOWN+"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.dl_week+=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"WKUP"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.ul_week=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"WKDOWN"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.dl_week=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"MUP-"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.ul_month-=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"MUP+"))   
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.ul_month+=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"MUP"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.ul_month=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"MDOWN+"))   
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.dl_month+=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"MDOWN-"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.dl_month-=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"MDOWN"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.dl_month=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"YUP-"))   
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.ul_year-=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"YUP+"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.ul_year+=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"YUP"))   
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.ul_year=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"YDOWN+"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.dl_year+=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"YDOWN-"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.dl_year-=l * 0x100000; else se = 1;
      } else if (!strcasecmp(*p,"YDOWN"))
      {   if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
         if (sscanf(q,SI64,&l)) usr.dl_year=l * 0x100000; else se = 1;
/* Now to allow changing of users transfer speeds only need =number, not + and -*/
      } else if(!strcasecmp(*p,"KBSUP"))
      {  if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
        if (sscanf(q,SI64,&l)) usr.ul_tm_total=(usr.ul_total/1024)/l; else se = 1;
      } else if(!strcasecmp(*p,"KBSDOWN"))
      {  if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
        if (sscanf(q,SI64,&l)) usr.dl_tm_total=(usr.dl_total/1024)/l; else se = 1;
      } else if(!strcasecmp(*p,"KBSWKDOWN"))
      {  if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
        if (sscanf(q,SI64,&l)) usr.dl_tm_week=(usr.dl_week/1024)/l; else se = 1;
      } else if(!strcasecmp(*p,"KBSWKUP"))
      {  if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
        if (sscanf(q,SI64,&l)) usr.ul_tm_week=(usr.ul_week/1024)/l; else se = 1;
      } else if(!strcasecmp(*p,"KBSMDOWN"))
      {  if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
        if (sscanf(q,SI64,&l)) usr.dl_tm_month=(usr.dl_month/1024)/l; else se = 1;
      } else if(!strcasecmp(*p,"KBSMUP"))
      {  if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
        if (sscanf(q,SI64,&l)) usr.ul_tm_month=(usr.ul_month/1024)/l; else se = 1;
      } else if(!strcasecmp(*p,"KBSYDOWN"))
      {  if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
        if (sscanf(q,SI64,&l)) usr.dl_tm_year=(usr.dl_year/1024)/l; else se = 1;
      } else if(!strcasecmp(*p,"KBSYUP"))
      {  if(!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
        if (sscanf(q,SI64,&l)) usr.ul_tm_year=(usr.ul_year/1024)/l; else se = 1;
/* end of speeds section */
      } else if (!strcasecmp(*p,"LOGINS"))
      { if (!allowed(c->user,R_USER_EDIT_LOGINS,u)) pd = 1; else
        if (sscanf(q,SI32,&i) && (i >= 0) && (i <= 255))
          usr.class = (usr.class & ~UL_MASK) | (i << 8); else se = 1;
      } else if (!strcasecmp(*p,"CLASS") && (strlen(q) == 1))
      { usr.class &= ~UC_MASK;
        switch (tolower(*q))
        { case 'a' :
            if (!allowed(c->user,R_USER_ADD_ANON,NULL)) pd = 1;
              else usr.class |= UC_ANON;
            break;
          case 'g' :
            if (!allowed(c->user,R_USER_ADD_GO,NULL)) pd = 1;
              else usr.class |= UC_GROUPOP;
            break;
          case 'n' :
            if (!allowed(c->user,R_USER_ADD,NULL)) pd = 1;
              else usr.class |= UC_NORMAL;
            break;
          case 'o' :
            if (!allowed(c->user,R_USER_ADD_SO,NULL)) pd = 1;
              else usr.class |= UC_SITEOP;
            break;
          case 'r' :
          case 's' :
            if (!superuser(c->user)) pd = 1;
              else usr.class |= UC_ROOT;
            break;
          case 't' :
            if (!allowed(c->user,R_USER_ADD_TMP,NULL)) pd = 1;
              else usr.class |= UC_TEMPLATE;
            break;
          default  :
            se = 1;
        }
      } else if (!strcasecmp(*p,"FLAGS+") && (strlen(q) == 1))
      { if (!allowed(c->user,R_USER_EDIT,u)) pd = 1; else
        switch (tolower(*q))
        { case 'c' : usr.class |= UF_COLOR; break;
          case 'd' : usr.class |= UF_DISABLED; break;
          case 'i' : usr.class |= UF_INVISIBLE; break;
          case 'p' : usr.class |= UF_NOPASSCH; break;
          case '-' : break;
          default  : se = 1;
        }
      } else if (!strcasecmp(*p,"FLAGS-") && (strlen(q) == 1))
      { if (!allowed(c->user,R_USER_EDIT,u)) pd = 1; else
        switch (tolower(*q))
        { case 'c' : usr.class &= ~UF_COLOR; break;
          case 'd' : usr.class &= ~UF_DISABLED; break;
          case 'i' : usr.class &= ~UF_INVISIBLE; break;
          case 'p' : usr.class &= ~UF_NOPASSCH; break;
          case '-' : break;
          default  : se = 1;
        }
      } else if (!strcasecmp(p[0],"UL_MIN"))
      { if (!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
        if (!sscanf(q,SI32,&j)) se = 1;
          else usr.ul_speed = (usr.ul_speed & 0xFFFF0000) | j;
      } else if (!strcasecmp(p[0],"UL_MAX"))
      { if (!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
        if (!sscanf(q,SI32,&j)) se = 1;
          else usr.ul_speed = (usr.ul_speed & 0xFFFF) | (j << 16);
      } else if (!strcasecmp(p[0],"DL_MIN"))
      { if (!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
        if (!sscanf(q,SI32,&j)) se = 1;
          else usr.dl_speed = (usr.dl_speed & 0xFFFF0000) | j;
      } else if (!strcasecmp(p[0],"DL_MAX"))
      { if (!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; else
        if (!sscanf(q,SI32,&j)) se = 1;
          else usr.dl_speed = (usr.dl_speed & 0xFFFF) | (j << 16);
      } else if (!strcasecmp(*p,"PASSWD"))
      { if (!allowed(c->user,R_USER_PASSWD,u) || (strlen(q) < sizeof(usr.password)+1))
        if (strlen(q) == 0) usr.password[0] = '\0'; else
        { salt[0] = salt_a[time(NULL) & 0x3F];
          salt[1] = salt_a[(time(NULL) >> 6) & 0x3F];
          strcpy(usr.password,crypt(q,salt));
        } else { pd = 1; }
      } else se = 1;
    } else se = 1;
    if (pd || se) n = 1;
  }
  if (pd)
  { server_reply(c,"-Permission denied.");
    return 0;
  } else if (se)
  { server_reply(c,"-Syntax error.");
    return 0;
  } else if (rf)
  { server_reply(c,"-IP-range full.");
    return 0;
  }

  /* if uid is to change, re-sort users */
  if (usr.uid != u->uid)
  { user_count--;
    if (u-user < user_count)
      memmove(u,u+1,(user_count-(u-user))*sizeof(*u));
    for (i=0, ch=child; i<children; i++, ch++)
      if (ch->user == u) ch->user = (struct user_s*)-1;
      else if (ch->user > u) ch->user--;
    u = insert_user(&usr);
    for (i=0, ch=child; i<children; i++, ch++)
      if (ch->user == (struct user_s*)-1) ch->user = u;
  } else
  { *u = usr;
    user_dirty++;
  }
  server_reply(c,"+User changed.");
  return 0;
}

/* print user */
void print_user(struct user_s* u, char* str)
{ int i;

  sprintf(str,
    "+%s\t"SI32"\t"SI32"\t"SI32"\t"SI32"\t"SI32"\t"
    SI64"\t"SI64"\t"SI64"\t"SI64"\t"SI64"\t"SI64"\t"
    SI64"\t"SI64"\t"SI64"\t"SI64"\t"SI64"\t"
    SI64"\t"SI64"\t"SI64"\t"SI64"\t"SI64"\t"
    SI64"\t"SI32"\t"SI64"\t%s\t%li",
    u->name,u->uid,u->gid,u->limit,u->ul_speed,u->dl_speed,
    u->ul_week,u->dl_week,u->ul_month,u->dl_month,u->ul_year,u->dl_year,
    u->ul_total,u->dl_total,u->ul_tm_week,u->dl_tm_week,u->ul_tm_month,
    u->dl_tm_month,u->ul_tm_year,u->dl_tm_year,u->ul_tm_total,u->dl_tm_total,
    u->credit,u->class,u->rights,u->info[0] ? u->info : "-",u->last_login);
  str += strlen(str);
  for (i=0; i<8; i++)
  { sprintf(str,"\t"SI32,u->gids[i]);
    str += strlen(str);
  }
  for (i=0; i<10; i++)
  { *(str++) = '\t';
    print_ident(str,u->idents[i]);
    str += strlen(str);
  }
}

/* user get */
int f4adp_usr_get(struct child_s* c, char** p, int n)
{ char           str[1024];
  struct user_s* u;

  u = get_user_by_name(p[0]);
  if (!u)
  { server_reply(c,"-User does not exist");
    return 0;
  }
  if (!allowed(c->user,R_USER_READ,u))
  { server_reply(c,"-Permission denied");
    return 0;
  }
  print_user(u,str);
  server_reply(c,str);
  return 0;
}

/* user kick (kill all sessions) */
int usr_kick(struct child_s** c, struct user_s* u, int signo)
{ struct child_s* ch;
  int             found;
  int             i;

  for (found=0, i=0, ch=child; i<children; ) if (ch->user == u)
  { found++;
    if (ch->pid != -1) kill(ch->pid,signo);
    if (*c)
    { if (ch == *c) *c = NULL; else if (ch < *c) (*c)--;
    }
    child_logout(ch,CL_KICK);
  } else
  { ch++;
    i++;
  }
  return found;
}

/* user kick (kill all sessions) */
int f4adp_usr_kick(struct child_s* c, char** p, int n)
{ struct child_s* ch;
  int             signo;
  struct user_s*  u;

  if (n == 2)
  { if (!sscanf(p[1],"%i",&signo)) signo = -1;
  } else signo = SIGQUIT;
  if (signo == -1)
  { server_reply(c,"-Syntax error");
    return 0;
  }
  u = get_user_by_name(p[0]);
  if (!u)
  { server_reply(c,"-User does not exist");
    return 0;
  }
  if (!allowed(c->user,R_USER_KICK,u))
  { server_reply(c,"-Permission denied");
    return 0;
  }
  ch = c;
  n = usr_kick(&ch,u,signo);
  if (ch)
  { server_reply(c,n ? "+User kicked" : "-User is not logged in");
    return c - ch;
  }
  return -1;
}

/* user kill session */
int f4adp_usr_kill(struct child_s* c, char** p, int n)
{ struct child_s* ch;
  int             i;
  pid_t           pid;
  int             signo;

  if (n == 2)
  { if (!sscanf(p[1],"%i",&signo)) signo = -1;
  } else signo = SIGQUIT;
  if ((signo == -1) || !sscanf(p[0],"%i",&pid))
  { server_reply(c,"-Syntax error");
    return 0;
  }
  for (i=0, ch=child; i<children; i++, ch++) if (ch->pid == pid)
  { if (!allowed(c->user,R_USER_KICK,ch->user))
    { server_reply(c,"-Permission denied");
      return 0;
    }
    if (pid != -1) kill(pid,signo);
    server_reply(c,"+Session killed");
    child_logout(ch,CL_KICK);
    return c == ch ? -1 : (c < ch ? 0 : 1);
  }
  server_reply(c,"-Session does not exist");
  return 0;
}

/* user list */
int f4adp_usr_ls(struct child_s* c, char** p, int n)
{ char           str[1024];
  struct user_s* u;

  for (n=0, u=user; n<user_count; n++, u++) if (allowed(c->user,R_USER_READ,u))
  { print_user(u,str);
    server_reply(c,str);
  }
  server_reply(c,"+");
  return 0;
}

/* user list restricted */
int f4adp_usr_lsr(struct child_s* c, char** p, int n)
{ char           str[32];
  struct user_s* u;   

  sprintf(str,"%i",user_count);
  server_reply(c,str);
  if (!user_count) return 0;
  if (!fgets_to(str,sizeof(str),c->rdp,5)) return 0;
  if (*str != '+') return 0;
  for (n=0, u=user; n<user_count; n++)
  { sprintf(str,"%s "SI32,u->name,u->uid);
    server_reply(c,str);
    u++;
  }
  return 0;
}

/* user password */
int f4adp_usr_pwd(struct child_s* c, char** p, int n)
{ char           salt[2];
  struct user_s* u;

  if (strcasecmp(p[0],c->user->name))
  { u = get_user_by_name(p[0]);
    if (!u)
    { server_reply(c,"-User does not exist");
      return 0;
    }
    if (!allowed(c->user,R_USER_PASSWD,u))
    { server_reply(c,"-Permission denied");
      return 0;
    }
  } else
  { u = c->user;
    if (anonuser(u) || no_pass_change(u))
    { server_reply(c,"-Permission denied");
      return 0;
    }
  }
  if ((n == 1) || !p[1][0])
  { /* reset password, if no password given or password is null string */
    u->password[0] = '\0';
    server_reply(c,"+Password cleared");
  } else
  { /* store crypted password */
    salt[0] = salt_a[time(NULL) & 0x3F];
    salt[1] = salt_a[(time(NULL) >> 6) & 0x3F];
    strcpy(u->password,crypt(p[1],salt));
    server_reply(c,"+Password changed");
  }
  user_dirty++;
  return 0;
}

/* user remove */
int f4adp_usr_rm(struct child_s* c, char** p, int n)
{ struct child_s* ch = child;
  int             i;
  struct user_s*  u;

  u = get_user_by_name(p[0]);
  if (!u)
  { server_reply(c,"-User does not exist");
    return 0;
  }
  if (!allowed(c->user,R_USER_DELETE,u))
  { server_reply(c,"-Permission denied");
    return 0;
  }
  for (i=0; i<children; i++, ch++) if (ch->user)
  { if (ch->user == u)
    { if (ch->pid != -1) kill(ch->pid,SIGKILL);
      child_logout(ch,CL_KICK);
    } else if (ch->user > u) ch->user--;
  }
  memmove(u,u+1,(--user_count-(u-user))*sizeof(struct user_s));
  top_ok = 0;
  user_dirty++;
  server_reply(c,"+User removed");
  return 0;
}

/* user reset */
int f4adp_usr_rst(struct child_s* c, char** p, int n)
{ int            r_time = 0;
  int            r_what = 0;
  int            se     = 0;
  struct user_s* u;

  if (!strcasecmp(p[0],"all")) r_what = 0x07; else
    if (!strcasecmp(p[0],"up")) r_what = 0x01; else
    if (!strcasecmp(p[0],"down")) r_what = 0x02; else
    if (!strcasecmp(p[0],"credit")) r_what = 0x04; else se = 1;
  if (n == 2)
  { if (r_what & 0x03)
    { if (!strcasecmp(p[1],"all")) r_time = 0x0F;
        else if (!strcasecmp(p[1],"week")) r_time = 0x01; else
        if (!strcasecmp(p[1],"month")) r_time = 0x02; else
        if (!strcasecmp(p[1],"year")) r_time = 0x04; else
        if (!strcasecmp(p[1],"total")) r_time = 0x08; else se = 1;
      if (r_time != 0x0F) r_what &= 0x03;
    } else se = 1;
  }
  if (se)
  { server_reply(c,"-Syntax error");
    return 0;
  }
  for (n=0, u=user; n<user_count; n++, u++) if (!templateuser(u))
  { if (r_what & 0x01)
    { if (r_time & 0x01) u->ul_week = u->ul_tm_week = 0;
      if (r_time & 0x02) u->ul_month = u->ul_tm_month = 0;
      if (r_time & 0x04) u->ul_year = u->ul_tm_year = 0;
      if (r_time & 0x08) u->ul_total = u->ul_tm_total = 0;
    }
    if (r_what & 0x02)
    { if (r_time & 0x01) u->dl_week = u->dl_tm_week = 0;
      if (r_time & 0x02) u->dl_month = u->dl_tm_month = 0;
      if (r_time & 0x04) u->dl_year = u->dl_tm_year = 0;
      if (r_time & 0x08) u->dl_total = u->dl_tm_total = 0;
    }
    if (r_what & 0x04) u->credit = 0;
  }
  user_dirty++;
  if (r_what & 0x03) top_ok = 0;
  server_reply(c,"+Reset successful");
  return 0;
}

/* user set */
int f4adp_usr_set(struct child_s* c, char** p, int n)
{ struct child_s* ch;
  int             i;
  int64           l;
  int             pd;
  struct user_s*  u;
  struct user_s   usr;

  u = get_user_by_name(p[0]);
  if (!u)
  { server_reply(c,"-User does not exist");
    return 0;
  }
  pd = !allowed(c->user,R_USER_EDIT,u);
  if (!pd)
  { if (strcasecmp(p[0],p[1])) if (get_user_by_name(p[1]))
    { server_reply(c,"-User already exists");
      return 0;
    }
    usr = *u;
    strcpy(usr.name,p[1]);
    sscanf(p[2],SI32,&usr.uid);
    if (usr.uid != u->uid) if (!allowed(c->user,R_USER_EDIT_UID,NULL)) pd = 1;
    sscanf(p[3],SI32,&usr.gid);
    if (groupop(c->user) && !groupopof(c->user,usr.gid)) pd = 1;
    sscanf(p[4],SI32,&usr.limit);
    if (usr.limit != u->limit) if (!allowed(c->user,R_USER_EDIT_CREDIT,NULL)) pd = 1;
    if (sscanf(p[5],SI64,&l))
    { if (l) if (!allowed(c->user,R_USER_EDIT_CREDIT,NULL)) pd = 1;
      usr.credit += l;
    } else
    { if (l) if (!allowed(c->user,R_USER_EDIT_CREDIT,NULL)) pd = 1;
      usr.credit = 0;
    }
    sscanf(p[6],SI32,&usr.class);
    if ((usr.class & UC_MASK) != (u->class & UC_MASK))
      switch (usr.class & UC_MASK)
    { case UC_NORMAL   :
        if (!allowed(c->user,R_USER_ADD,NULL)) pd = 1; break;
      case UC_GROUPOP  :
        if (!allowed(c->user,R_USER_ADD_GO,NULL)) pd = 1; break;
      case UC_SITEOP   :
        if (!allowed(c->user,R_USER_ADD_SO,NULL)) pd = 1; break;
      case UC_TEMPLATE :
        if (!allowed(c->user,R_USER_ADD_TMP,NULL)) pd = 1; break;
      case UC_ANON :
        if (!allowed(c->user,R_USER_ADD_ANON,NULL)) pd = 1; break;
      default          :
        if (!superuser(c->user)) pd = 1;
    }
    if (usr.class >> 8 != u->class >> 8)
      if (!allowed(c->user,R_USER_EDIT_LOGINS,NULL)) pd = 1;
    sscanf(p[7],SI64,&usr.rights);
    if (usr.rights != u->rights) if (!superuser(c->user))
    { if (c->user == u) pd = 1; else
      if (!allowed(c->user,R_USER_EDIT_RIGHTS,NULL)) pd = 1; else
      /* you must have the rights that you want to grant/revoke */
      if ((c->user->rights | (usr.rights ^ u->rights)) != c->user->rights)
        pd = 1;
    }
    if (strlen(p[8]) >= sizeof(usr.info))
    { server_reply(c,"-User information is too long.");
      return 0;
    }
    strcpy(usr.info,strcmp(p[8],"-") ? p[8] : "");
    if (strcmp(usr.info,u->info))
      if (!allowed(c->user,R_USER_EDIT_INFO,NULL)) pd = 1;
    sscanf(p[9],SI32,&usr.ul_speed);
    if (usr.ul_speed != u->ul_speed) if (!allowed(c->user,R_USER_EDIT_SPEEDS,NULL)) pd = 1;
    sscanf(p[10],SI32,&usr.dl_speed);
    if (usr.dl_speed != u->dl_speed) if (!allowed(c->user,R_USER_EDIT_SPEEDS,NULL)) pd = 1;
    for (n=0; n<8; n++)
    { sscanf(p[n+11],SI32,&usr.gids[n]);
      if (usr.gids[n] != u->gids[n])
        if (!allowed(c->user,R_USER_EDIT_GRPS,NULL)) pd = 1;
    }
    for (n=0; n<10; n++)
    { get_iprange(&p[n+19],usr.idents[n]);
      if (memcmp(usr.idents[n],u->idents[n],sizeof(usr.idents[0])))
        if (!allowed(c->user,R_USER_EDIT_IP,NULL)) pd = 1;
    }
  }
  if (pd)
  { server_reply(c,"-Permission denied");
    return 0;
  }
  /* if uid is to change, re-sort users */
  if (usr.uid != u->uid)
  { user_count--;
    if (u-user < user_count) memmove(u,u+1,(user_count-(u-user))*sizeof(*u));
    for (i=0, ch=child; i<children; i++, ch++)
      if (ch->user == u) ch->user = (struct user_s*)-1;
      else if (ch->user > u) ch->user--;
    u = insert_user(&usr);
    for (i=0, ch=child; i<children; i++, ch++)
      if (ch->user == (struct user_s*)-1) ch->user = u;
  } else
  { *u = usr;
    user_dirty++;
  }
  server_reply(c,"+User changed");
  top_ok = 0;
  return 0;
}

/* EOF */
