/*   log.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-1999 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "externsd.h"
#include "f4adp_admin.h"
#include "registry.h"
#include "serverd.h"
#include "utils.h"

void errlog(char* fmt, ...)
{ va_list ap;
  char    msg[1024];
  time_t  t;

  strcpy(msg,"+E ");
  time(&t);
  strftime(msg+3,64,"%a %d/%m/%Y %H:%M:%S",localtime(&t));
  sprintf(msg+strlen(msg)," [%i] ",getpid());
  va_start(ap,fmt);
  vsprintf(msg+strlen(msg),fmt,ap);
  va_end(ap);
  if (errlogf) fprintf(errlogf,"%s\n",msg+3);
  notify_watchers(msg,0,NULL);
}

void errlog2(char* fmt, ...)
{ va_list ap;
  char    msg[1024];
  time_t  t;

  strcpy(msg,"+E ");
  time(&t);
  strftime(msg+3,64,"%a %d/%m/%Y %H:%M:%S",localtime(&t));
  sprintf(msg+strlen(msg)," [%i] ",getpid());
  va_start(ap,fmt);
  vsprintf(msg+strlen(msg),fmt,ap);
  va_end(ap);
  fprintf(stderr,"%s\n",msg+3);
  fflush(stderr);
}

void parse_log(void)
{ int   n;
  char* p;
  char* q;
  int   v;

  logc = 0;
  p = reg_get("log");
  if (p)
  { while (skipspace(&p))
    { if (*p == '!')
      { n = 0x01;
        p++;
      } else n = 0x00;
      q = p;
      if (skipnspace(&q))
      { n |= 0x02;
        *q = '\0';
      }
      if (!strcasecmp(p,"all")) v = LG_ALL; else
      if (!strcasecmp(p,"cd")) v = LG_CD; else
      if (!strcasecmp(p,"chmod")) v = LG_CHMOD; else
      if (!strcasecmp(p,"del")) v = LG_DEL; else
      if (!strcasecmp(p,"get")) v = LG_GET; else
      if (!strcasecmp(p,"login")) v = LG_LOGIN; else
      if (!strcasecmp(p,"logout")) v = LG_LOGOUT; else
      if (!strcasecmp(p,"ls")) v = LG_LS; else
      if (!strcasecmp(p,"md")) v = LG_MD; else
      if (!strcasecmp(p,"none")) v = 0; else
      if (!strcasecmp(p,"put")) v = LG_PUT; else
      if (!strcasecmp(p,"rd")) v = LG_RD; else
      if (!strcasecmp(p,"ren")) v = LG_REN; else
      if (!strcasecmp(p,"site")) v = LG_SITE; else v = -1;
      if (v != -1)
      { if (n & 0x01) logc &= ~v; else logc |= v;
      }
      if (n & 0x02) *(q++) = ' ';
      p = q;  
    }
  }
}

void open_logfiles(void)
{ int   n;
  char* param[16];
  int   pipefd[2];
  char  tmpbuf[256];

  if (log_pid)
  { fclose(log_program_f);
    log_program_f = NULL;
    log_pid = 0;
  }
  mk_filename(tmpbuf,errlog_file,DFL_ERRLOGFILE);
  errlogf = freopen(tmpbuf,"at",errlogf ? errlogf : stdout);
  if (errlogf) setvbuf(errlogf,NULL,_IOLBF,0); else
    printf("ftpd[%i] freopen(%s): %s (%i)\n",getpid(),tmpbuf,strerror(errno),errno);
  setvbuf(stderr,NULL,_IOLBF,0);
  mk_filename(tmpbuf,ftpserrlog_file,DFL_FTPSERRLOGFILE);
  if (!freopen(tmpbuf,"at",stderr)) errlog("freopen(%s): %s (%i)\n",
    tmpbuf,strerror(errno),errno);
  if (logf)
  { fclose(logf);
    logf = NULL;
  }
  parse_log();
  if (logc)
  { mk_filename(tmpbuf,log_file,DFL_LOGFILE);
    if ((logf = fopen(tmpbuf,"at")) == NULL)
      errlog("cannot open logfile (%s)",tmpbuf);
    if (logf)
    { if (fcntl(fileno(logf),F_SETFD,1) < 0)
        errlog("fcntl: %s (%i)",strerror(errno),errno);
      setvbuf(logf,NULL,_IOLBF,0);
    }
    if (log_program)
    { if (logf) fflush(logf);
      if (errlogf) fflush(errlogf);
      fflush(stderr);
      if (pipe(pipefd) < 0) errlog("pipe(): %s (%i)",strerror(errno),errno);
      else switch (log_pid = fork())
      { case  0 :
          close(0);
          close(1);
          close(2);
	  close(pipefd[1]);
	  if (dup2(pipefd[0],0) != -1) close(pipefd[0]);
	  n = 0;
	  parse_command(log_program,param,15,16,&n);
	  param[n] = NULL;
	  n = execvp(param[0],param);
	  exit(n);
	case -1 :
	  errlog("fork: %s",strerror(errno));
	  log_pid = 0;
	  close(pipefd[0]);
	  close(pipefd[1]);
	  break;
	default :
	  close(pipefd[0]);
	  log_program_f = fdopen(pipefd[1],"wt");
	  if (!log_program_f)
	  { errlog("fdopen(%i): %s (%i)",pipefd[1],strerror(errno),errno);
	    close(pipefd[1]);
          }
      }
    }
  }
}

void server_log_local(struct child_s* c, long l, char* str, ...)
{ va_list ap;
  time_t  t;
  char    tstr[32];

  if (!logf || !(logc & l)) return;
  time(&t);
  strftime(tstr,sizeof(tstr),"%a %d/%m/%Y %H:%M:%S",localtime(&t));
  fprintf(logf,"%s %5i %-9s ",tstr,c ? c->pid : getpid(),
    c && c->user ? c->user->name : "???");
  va_start(ap,str);
  vfprintf(logf,str,ap);
  fprintf(logf,"\n");
  if (log_pid)
  { fprintf(log_program_f,"%s %5i %-9s ",tstr,c ? c->pid : getpid(),
      c && c->user ? c->user->name : "???");
    vfprintf(log_program_f,str,ap);
    fprintf(log_program_f,"\n");
    fflush(log_program_f);
  }
}

void client_log(struct child_s* c, int idle, char** p, int n)
{ int    i;
  int    lno;
  char   tstr[32];
  char** q;

  if (!sscanf(p[0],"%i",&c->hidden)) c->hidden = 0;
  if (!sscanf(p[1],"%i",&lno)) lno = 0;
  c->idle = idle;
  /* log */
  if ((logc & lno) && (logf || log_program_f))
  { strftime(tstr,sizeof(tstr),"%a %d/%m/%Y %H:%M:%S",localtime(&c->act_time));
    if (logf) fprintf(logf,"%s %5i %-9s",tstr,c->pid,
      c->user ? c->user->name : "???");
    if (log_program_f) fprintf(log_program_f,"%s %5i %-9s",tstr,c->pid,
      c->user ? c->user->name : "???");
    i = n - 2;
    q = p + 2;
    while (i--)
    { if (logf) fprintf(logf," %s",*q);
      if (log_program_f) fprintf(log_program_f," %s",*q);
      q++;
    }
    if (logf) fprintf(logf,"\n");
    if (log_program_f)
    { fprintf(log_program_f,"\n");
      fflush(log_program_f);
    }
  }
  /* client activity */
  strcpy(c->activity,"");
  i = n - 2;
  q = p + 2;
  while (i--)
  { if (strlen(c->activity)+strlen(*q)+1 < sizeof(c->activity))
    { if (*c->activity) strcat(c->activity," ");
      strcat(c->activity,*q);
    } else
    { if (strlen(c->activity) < sizeof(c->activity)-2)
      { strcat(c->activity," ");
        strncpy(c->activity+strlen(c->activity),*q,
          sizeof(c->activity)-strlen(c->activity)-1);
        *(c->activity+sizeof(c->activity)-1) = '\0';
      }
      i = 0;
    }
    q++;
  }
  server_reply(c,"+ok\n");
  notify_watchers_activity(c);
}

int f4adp_log_begin(struct child_s* c, char** p, int n)
{ client_log(c,0,p,n);
  return 0;
}

int f4adp_log_end(struct child_s* c, char** p, int n)
{ client_log(c,1,p,n);
  return 0;
}

/* EOF */
