/*   registry.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-2000 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "tweak.h"
#include "externsd.h"
#include "group.h"
#include "log.h"
#include "reject.h"
#include "serverd.h"
#include "user.h"
#include "utils.h"

char*  registry            = NULL;  /* the registry                         */
int    registry_entries    = 0;     /* number of entries                    */
int    registry_maxentries = 0;     /* max number of entries
                                       (size of registry_keys)              */
char** registry_keys       = NULL;  /* array of keys, sorted alphabetically */
int    registry_len        = 0;     /* length of registry in bytes          */
int    registry_size       = 0;     /* site of registry (char* registry)    */
int    registry_dirty      = 0;     /* registry modified ? (counter)        */

#define CT_BOOL  1
#define CT_INT   2
#define CT_SHORT 3
#define CT_INT32 4
#define CT_INT64 5
#define CT_CHAR  6
#define CT_DIR   7

struct config_s config_vars[] =
{ { "adminport",                CT_INT , &admin_port        , 0, 0xFFFF            , NULL },
  { "alternativeserverprogram", CT_CHAR, &alternative_srv   , 0, 0                 , NULL },
  { "basedir"                 , CT_DIR , &base_dir          , 0, 0                 , default_base_dir },
  { "bouncer"                 , CT_CHAR, &bouncer           , 0, 0                 , NULL },
  { "bouncertimeout"          , CT_INT,  &bouncertimeout    , 0, 300               , NULL },
  { "checktimeout"            , CT_INT,  &checktimeout      , 0, 0xFFFF            , NULL },
  { "chmod_allowed"           , CT_BOOL, &flags             , 1, FLAG_CHMODALLOWED , NULL },
  { "desencryption"           , CT_BOOL, &flags             , 1, FLAG_DESENCRYPTION, NULL },
  { "destroylogin"            , CT_CHAR, &destroy_login     , 0, 0                 , NULL },
  { "destroypassword"         , CT_CHAR, &destroy_password  , 0, 0                 , NULL },
  { "dupecheck"               , CT_CHAR, &dupechk           , 0, 0                 , NULL },
  { "dupepath"                , CT_CHAR, &dupe_path         , 0, 0                 , NULL },
  { "errlogfile"              , CT_DIR , &errlog_file       , 0, 0                 , DFL_ERRLOGFILE },
  { "filenameconversion"      , CT_INT , &fconv_value       , 0, 0xFFFF            , NULL },
  { "ffpath"                  , CT_CHAR, &ff_path           , 0, 0                 , NULL },
  { "ftpserrlogfile"          , CT_DIR , &ftpserrlog_file   , 0, 0                 , DFL_FTPSERRLOGFILE },
  { "groupfile"               , CT_DIR,  &group_file        , 0, 0                 , DFL_GROUPFILE },
  { "ident"                   , CT_BOOL, &flags             , 0, FLAG_IDENTMODE    , NULL },
  { "log"                     , CT_CHAR, NULL               , 0, 0                 , NULL },
  { "logfile"                 , CT_DIR , &log_file          , 0, 0                 , DFL_LOGFILE },
  { "loginkick"               , CT_BOOL, &flags             , 1, FLAG_LOGINKICK    , NULL },
  { "loginretries"            , CT_INT , &login_retries     , 0, 0xFFFF            , NULL },
  { "logprogram"              , CT_CHAR, &log_program       , 0, 0                 , NULL },
  { "nukebrokenuploads"       , CT_BOOL, &flags             , 0, FLAG_NUKEBROKENUP , NULL },
  { "other_ip"                , CT_CHAR, &other_ip          , 0, 0                 , NULL },
  { "passivemode"             , CT_BOOL, &flags             , 1, FLAG_PASSIVEMODE  , NULL },
  { "permissionfile"          , CT_CHAR, &permission_file   , 0, 0                 , default_permission_file },
  { "pidfile"                 , CT_DIR , &pid_file          , 0, 0                 , DFL_PIDFILE },
  { "port"                    , CT_INT,  &port              , 0, 0xFFFF            , NULL },
  { "port2"                   , CT_INT,  &port2             , 0, 0xFFFF            , NULL },
  { "pre_download"            , CT_CHAR, &pre_download      , 0, 0                 , NULL },
  { "pre_upload"              , CT_CHAR, &pre_upload        , 0, 0                 , NULL },
  { "private"                 , CT_BOOL, &flags             , 0, FLAG_PRIVATEMODE  , NULL },
  { "programname"             , CT_CHAR, &program_name      , 0, 0                 , NULL },
  { "readmefile"              , CT_CHAR, &readme_file       , 0, 0                 , NULL },
  { "reject"                  , CT_CHAR, &rejected_ips      , 0, 0                 , NULL },
  { "resolver"                , CT_BOOL, &flags             , 1, FLAG_RESOLVER     , NULL },
  { "resolvesymlinks"         , CT_BOOL, &flags             , 0, FLAG_RESOLVESYMLINKS, NULL },
  { "serverprogram"           , CT_CHAR, &server_program_ptr, 0, 0                 , DFL_SERVERPRG },
  { "setgid"                  , CT_INT , &set_gid           ,-1, 0x7FFFFFFF        , NULL },
  { "setuid"                  , CT_INT , &set_uid           ,-1, 0x7FFFFFFF        , NULL },
  { "shortcut"                , CT_CHAR, &shortcut          , 0, 0                 , NULL },
  { "statdir"                 , CT_DIR , &stat_dir          , 0, 0                 , NULL },
  { "statfile"                , CT_DIR , &stat_file         , 0, 0                 , DFL_STATFILE },
  { "gstatfile"               , CT_DIR , &gstat_file        , 0, 0                 , DFL_GSTATFILE },
  { "timeoutanon"             , CT_INT , &timeout_anon      , 0, 0xFFFF            , NULL },
  { "timeoutuser"             , CT_INT , &timeout_user      , 0, 0xFFFF            , NULL },
  { "touch"                   , CT_BOOL, &flags             , 0, FLAG_TOUCH        , NULL },
  { "umask"                   , CT_INT , &umask_value       , 0, 0777              , NULL },
  { "umaskanon"               , CT_INT , &umask_anon        , 0, 0xFFFFFF          , NULL },
  { "umaskroot"               , CT_INT , &umask_root        , 0, 0xFFFFFF          , NULL },
  { "umaskuser"               , CT_INT , &umask_user        , 0, 0xFFFFFF          , NULL },
  { "userfile"                , CT_DIR , &user_file         , 0, 0                 , DFL_USERFILE },
  { ""                        , CT_INT,  NULL               , 0, 0                 , NULL } };

void reg_updatekeys(char* from, int n)
{ int i;

  for (i=0; i<registry_entries; i++) if (registry_keys[i] >= from)
    registry_keys[i] += n;
}

/* find registry key */
char** reg_find(char* key)
{ int i;
  int j;
  int k;
  int r;

  i = 0;
  k = registry_entries - 1;
  while (i <= k)
  { j = (i + k) >> 1;
    r = strcmp(key,registry_keys[j]);
    if (r < 0) k = j - 1; else if (r > 0) i = j + 1;
      else return registry_keys+j;
  }
  return NULL;
}

/* delete registry key & entry */
int reg_delete(char* key)
{ char** entry;
  int    n;
  int    pos;
  char*  ptr;

  entry = reg_find(key);
  if (entry)
  { ptr = *entry;
    ptr += strlen(ptr) + 1;
    ptr += strlen(ptr) + 1;
    pos = ptr - registry;
    n = ptr - *entry;
    if (pos < registry_len) memmove(*entry,ptr,registry_len-pos);
    reg_updatekeys(ptr,-n);
    registry_len -= n;
    pos = entry - registry_keys;
    registry_entries--;
    if (pos < registry_entries) memmove(entry,entry+1,
      (registry_entries-pos)*sizeof(char*));
    registry_dirty++;
    return 0;
  } else return -1;
}

/* insert registry key */
char** reg_insert(char* key)
{ int    i;
  int    j;
  int    k;
  char** newkeys;
  int    r;

  if (registry_entries >= registry_maxentries)
  { i = registry_maxentries + 16;
    newkeys = alloc(registry_keys,i*sizeof(char*));
    if (!newkeys)
    { errlog("out of memory (registry)");
      return NULL;
    }
    registry_keys = newkeys;
    registry_maxentries = i;
  }
  i = 0;
  k = registry_entries - 1;
  while (i <= k)
  { j = (i + k) >> 1;
    r = strcmp(key,registry_keys[j]);
    if (r < 0) k = j - 1; else if (r > 0) i = j + 1; else
    { registry_keys[j] = key;
      return registry_keys+j;
    }
  }
  k++;
  if (k < registry_entries) memmove(registry_keys+k+1,registry_keys+k,
    (registry_entries-k)*sizeof(char*));
  registry_keys[k] = key;
  registry_entries++;
  return registry_keys+k;
}

char* reg_get(char* key)
{ char** entry;

  entry = reg_find(key);
  return entry ? (*entry + strlen(*entry) + 1) : NULL;
}

int reg_getb(char* key, int defval)
{ int   i;
  char* val;

  val = reg_get(key);
  if (val)
  { if (!strcmp(val,"0") || !strcasecmp(val,"no") || !strcasecmp(val,"off")
      || !strcasecmp(val,"false") || !strcasecmp(val,"disabled"))
      i = 0; else
    if (!strcmp(val,"1") || !strcasecmp(val,"yes") || !strcasecmp(val,"on")
      || !strcasecmp(val,"true") || !strcasecmp(val,"enabled"))
      i = 1; else i = defval;
  } else i = defval;
  return i;
}

int reg_geti(char* key, int defval)
{ int   i;
  char* val;

  val = reg_get(key);
  if (val)
  { if (!sscanf(val,"%i",&i)) i = defval;
  } else i = defval;
  return i;
}

int32 reg_geti32(char* key, int32 defval)
{ int32 i;
  char* val;

  val = reg_get(key);
  if (val)
  { if (!sscanf(val,SI32,&i)) i = defval;
  } else i = defval;
  return i;
}

int64 reg_geti64(char* key, int64 defval)
{ int64 i;
  char* val;

  val = reg_get(key);
  if (val)
  { if (!sscanf(val,SI64,&i)) i = defval;
  } else i = defval;
  return i;
}

long reg_getl(char* key, long defval)
{ long  l;
  char* val;

  val = reg_get(key);
  if (val)
  { if (!sscanf(val,"%li",&l)) l = defval;
  } else l = defval;
  return l;
}

int reg_put(char* key, char* value)
{ char** entry;
  int    lk;
  int    lov;
  int    lv;
  char*  ptr;

  lk = strlen(key);
  lv = strlen(value);
  if (registry_len + lk + lv + 2 > registry_size)
  { lov = registry_size + 1024;
    ptr = alloc(registry,lov);
    if (!ptr)
    { errlog("out of memory (registry)");
      return -1;
    }
    reg_updatekeys(registry,ptr-registry);
    registry = ptr;
    registry_size = lov;
  }
  entry = reg_find(key);
  if (entry)
  { ptr = *entry + lk + 1;
    lov = strlen(ptr);
    if (lov < lv)
    { lov = lv - lov;
      memmove(ptr+lov,ptr,registry_len-(ptr-registry));
      registry_len += lov;
      reg_updatekeys(ptr,lov);
    } else if (lov > lv)
    { lov -= lv;
      memmove(ptr,ptr+lov,registry_len-(ptr-registry)-lov);
      registry_len -= lov;
      reg_updatekeys(ptr,-lov);
    }
    strcpy(ptr,value);
  } else
  { ptr = registry+registry_len;
    strcpy(ptr,key);
    if (!reg_insert(ptr)) return -1;
    registry_len += lk + 1;
    strcpy(registry+registry_len,value);
    registry_len += lv + 1;
  }
  registry_dirty++;
  return 0;
}

int load_registry(char* filename, int reset)
{ FILE* file;
  char* p;
  char* q;
  char  str[1024];

  mk_filename(str,filename,DFL_REGFILE);
  if ((file = fopen(str,"rt")) == NULL)
  { errlog("fopen(%s): %s (%i)",str,strerror(errno),errno);
    return -1;
  }
  if (reset) registry_entries = registry_len = 0;
  while (readline(str,sizeof(str),file))
  { p = str;
    skipspace(&p);
    if (*p && (*p != '#'))
    { q = p;
      skipnspace(&q);
      if (*q) *(q++) = '\0';
      skipspace(&q);
      reg_put(p,q);
    }
  }
  fclose(file);
  if (reset) registry_dirty = 0;
  return 0;
}

int save_registry(char* filename)
{ FILE* file;
  char  filename_tmp[256];
  int   i = 0;
  char* p;
  char* q;
  char  str[256];

  mk_filename(filename_tmp,filename,DFL_REGFILE);
  strcat(filename_tmp,".tmp");
  if ((file = fopen(filename_tmp,"wt")) == NULL)
  { errlog("fopen(%s): %s (%i)",filename_tmp,strerror(errno),errno);
    return -1;
  }
  while (i < registry_len)
  { p = registry+i;
    i += strlen(registry+i) + 1;
    q = registry+i;
    i += strlen(registry+i) + 1;
    fprintf(file,"%-30s %s\n",p,q);
  }
  fclose(file);
  strcpy(str,filename_tmp);
  str[strlen(str)-4] = '\0';
  if (rename(filename_tmp,str) == -1)
  { errlog("rename(%s,%s): %s (%i)",filename_tmp,str,strerror(errno),errno);
    return -1;
  }
  return 0;
}

int load_config(char* filename, int reconfig, int reset)
{ int   i;
  int   n;
  char* ptr;
  int   r;

  if (!reconfig) r = load_registry(filename,reset);
  r = 0;
  for (i=0; config_vars[i].key[0]; i++)
  { switch (config_vars[i].type)
    { case CT_INT  :
        n = reg_geti(config_vars[i].key,*(int*)config_vars[i].value);
        if ((n < config_vars[i].min) || (n > config_vars[i].max))
        { errlog("value %i of variable %s not in range from %i to %i",
            n,config_vars[i].key,config_vars[i].min,config_vars[i].max);
        } else *(int*)config_vars[i].value = n;
        break;
      case CT_BOOL :
        n = reg_getb(config_vars[i].key,config_vars[i].min);
        flag_set((int*)config_vars[i].value,config_vars[i].max,n);
        break;
      case CT_CHAR :
      case CT_DIR  :
        ptr = reg_get(config_vars[i].key);
        if (config_vars[i].value)
        { *(char**)(config_vars[i].value) = ptr ? ptr :
            (char*)config_vars[i].dflvalue;
        }
        break;
    }
  }
  mk_filename(admin_program,DFL_ADMINPRG,NULL);
  mk_filename(server_program,server_program_ptr,DFL_SERVERPRG);
  reject_parse(rejected_ips);
  if (!reconfig)
  { if (set_gid >= 0) if (setegid(set_gid) < 0)
      errlog("setegid(%i): %s (%i)",set_gid,strerror(errno),errno);
    if (set_uid >= 0) if (seteuid(set_uid) < 0)
      errlog("seteuid(%i): %s (%i)",set_uid,strerror(errno),errno);
  }
  umask(umask_value);
  open_logfiles();
  if (!reconfig)
  { if (load_user_file(user_file,0) == -1) r = -1;
    if (load_group_file(group_file,0) == -1) r = -1;
  }
  if (chdir(base_dir) == -1)
  { errlog("chdir(%s): %s (%i)",
      base_dir,strerror(errno),errno);
    r = -1;
  }
  return r;
}

void print_configuration(FILE* out)
{ fprintf(out,"Base directory : %s\n",base_dir ? base_dir : "NULL");
  fprintf(out,"Readme file    : %s\n",readme_file ? readme_file : "(none)");
  fprintf(out,"Permission file: %s\n",permission_file ? permission_file : "(none)");
  fprintf(out,"Errlog file    : %s\n",errlog_file ? errlog_file : "(none)");
  fprintf(out,"Log file       : %s\n",log_file ? log_file : "(none)");
  fprintf(out,"Log program    : %s\n",log_program ? log_program : "(none)");
  fprintf(out,"Server program : %s\n",server_program ? server_program : "NULL");
}

/* get */
int f4adp_reg_get(struct child_s* c, char** p, int n)
{ char  str[512];
  char* val;

  val = reg_get(p[0]);
  if (val) sprintf(str,"+%s",val); else strcpy(str,"-Key does not exist");
  server_reply(c,str);
  return 0;
}

/* keys */
int f4adp_reg_keys(struct child_s* c, char** p, int n)
{ int  i;
  char str[512];

  for (i=0; config_vars[i].key[0]; i++)
  { sprintf(str,"+%s\n",config_vars[i].key);
    server_reply(c,str);
  }
  server_reply(c,"+");
  return 0;
}

/* list */
int f4adp_reg_ls(struct child_s* c, char** p, int n)
{ int  i;
  char str[512];

  sprintf(str,"+entries: %i/%i   size: %i/%i",
    registry_entries,registry_maxentries,registry_len,registry_size);
  server_reply(c,str);
  for (i=0; i<registry_entries; i++)
  { sprintf(str,"+%-20s %s\n",registry_keys[i],
      registry_keys[i]+strlen(registry_keys[i])+1);
    server_reply(c,str);
  }
  server_reply(c,"+");
  return 0;
}

/* remove */
int f4adp_reg_rm(struct child_s* c, char** p, int n)
{ char str[128];

  if (reg_delete(p[0]) == -1) strcpy(str,"-Key does not exist"); else
  { strcpy(str,"+Key removed");
    load_config(NULL,1,0);
  }
  server_reply(c,str);
  return 0;
}

/* put */
int f4adp_reg_set(struct child_s* c, char** p, int n)
{ char str[512];

  if (reg_put(p[0],p[1]) < 0) strcpy(str,"-Set failed (out of memory ?)");
  else
  { strcpy(str,"+Set succeeded");
    load_config(NULL,1,0);
  }
  server_reply(c,str);
  return 0;
}

/* EOF */
