/*   stat.c
 *                                                                              
 *   This file is part of FTP4ALL                                               
 *   
 *   Copyright (C) 1996-1999 by Crescent
 *   
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or   
 *   any later version.                                                  
 *   
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *   GNU General Public License for more details.                   
 *   
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "externsd.h"
#include "utils.h"

struct stat_s
{ struct user_s* u;    /* user   */
  int64          n;    /* amount */
  int64          t;    /* time   */
} ;

struct statg_s
{ struct group_s* g;   /* group  */
  int64           n;   /* amount */
  int64           t;   /* time   */
} ;

struct stat_s*  s          = NULL;  /* array of users sorted by amount  */
int             stat_size  = 0;     /* capacity of array s              */
struct statg_s* sg         = NULL;  /* array of groups sorted by amount */
int             statg_size = 0;     /* capacity of array sg             */

/* allocate memory for user stat array */
int alloc_stat_mem()
{ int            i;
  struct stat_s* u;

  if (stat_size < user_size)
  { i = user_size + 16;
    u = alloc(s,i*sizeof(struct stat_s));
    if (!u) return 0;
    s = u;
    stat_size = i;
  }
  return 1;
}

/* allocate memory for group stat array */
int alloc_statg_mem()
{ struct statg_s* g;
  int             i;

  if (statg_size < group_size)
  { i = group_size + 4;
    g = alloc(sg,i*sizeof(struct statg_s));
    if (!g) return 0;
    sg = g;
    statg_size = i;
  }
  return 1;
}

/* reset month stats of all users */
void reset_stats_month()
{ int            i;
  struct user_s* u;

  for (i=0, u=user; i<user_count; i++, u++)
    u->dl_month = u->ul_month = u->dl_tm_month = u->ul_tm_month = 0;
  top_ok = 0;
}

/* reset total stats of all users */
void reset_stats_total()
{ int            i;
  struct user_s* u;

  for (i=0, u=user; i<user_count; i++, u++)
    u->dl_total = u->ul_total = u->dl_tm_total = u->ul_tm_total = 0;
  top_ok = 0;
}

/* reset week stats of all users */
void reset_stats_week()
{ int            i;
  struct user_s* u;

  for (i=0, u=user; i<user_count; i++, u++)
    u->dl_week = u->ul_week = u->dl_tm_week = u->ul_tm_week = 0;
  top_ok = 0;
}

/* reset year stats of all users */
void reset_stats_year()
{ int            i;
  struct user_s* u;

  for (i=0, u=user; i<user_count; i++, u++)
    u->dl_year = u->ul_year = u->dl_tm_year = u->ul_tm_year = 0;
  top_ok = 0;
}

/* sort users by month download into stat array */
void sort_dl_month()
{ int            i,j,k,n;
  struct user_s* u;

  for (n=0, u=user; n<user_count; u++, n++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (u->dl_month <= s[j].n) i = j + 1; else k = j - 1;
    }
    if (++k < n) memmove(s+k+1,s+k,(n-k)*sizeof(struct stat_s));
    s[k].u = u;
    s[k].n = u->dl_month;
    s[k].t = u->dl_tm_month;
  }
}

/* sort users by month upload into stat array */
void sort_ul_month()
{ int            i,j,k,n;
  struct user_s* u;

  for (n=0, u=user; n<user_count; u++, n++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (u->ul_month <= s[j].n) i = j + 1; else k = j - 1;
    }
    if (++k < n) memmove(s+k+1,s+k,(n-k)*sizeof(struct stat_s));
    s[k].u = u;
    s[k].n = u->ul_month;
    s[k].t = u->ul_tm_month;
  }
}

/* sort users by total download into stat array */
void sort_dl_total()
{ int            i,j,k,n;
  struct user_s* u;

  for (n=0, u=user; n<user_count; u++, n++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (u->dl_total <= s[j].n) i = j + 1; else k = j - 1;
    }
    if (++k < n) memmove(s+k+1,s+k,(n-k)*sizeof(struct stat_s));
    s[k].u = u;
    s[k].n = u->dl_total;
    s[k].t = u->dl_tm_total;
  }
}

/* sort users by total upload into stat array */
void sort_ul_total()
{ int            i,j,k,n;
  struct user_s* u;

  for (n=0, u=user; n<user_count; u++, n++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (u->ul_total <= s[j].n) i = j + 1; else k = j - 1;
    }
    if (++k < n) memmove(s+k+1,s+k,(n-k)*sizeof(struct stat_s));
    s[k].u = u;
    s[k].n = u->ul_total;
    s[k].t = u->ul_tm_total;
  }
}

/* sort users by week download into stat array */
void sort_dl_week()
{ int            i,j,k,n;
  struct user_s* u;

  for (n=0, u=user; n<user_count; u++, n++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (u->dl_week <= s[j].n) i = j + 1; else k = j - 1;
    }
    if (++k < n) memmove(s+k+1,s+k,(n-k)*sizeof(struct stat_s));
    s[k].u = u;
    s[k].n = u->dl_week;
    s[k].t = u->dl_tm_week;
  }
}

/* sort users by week upload into stat array */
void sort_ul_week()
{ int            i,j,k,n;
  struct user_s* u;

  for (n=0, u=user; n<user_count; u++, n++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (u->ul_week <= s[j].n) i = j + 1; else k = j - 1;
    }
    if (++k < n) memmove(s+k+1,s+k,(n-k)*sizeof(struct stat_s));
    s[k].u = u;
    s[k].n = u->ul_week;
    s[k].t = u->ul_tm_week;
  }
}

/* sort users by year download into stat array */
void sort_dl_year()
{ int            i,j,k,n;
  struct user_s* u;

  for (n=0, u=user; n<user_count; u++, n++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (u->dl_year <= s[j].n) i = j + 1; else k = j - 1;
    }
    if (++k < n) memmove(s+k+1,s+k,(n-k)*sizeof(struct stat_s));
    s[k].u = u;
    s[k].n = u->dl_year;
    s[k].t = u->dl_tm_year;
  }
}

/* sort users by year upload into stat array */
void sort_ul_year()
{ int            i,j,k,n;
  struct user_s* u;

  for (n=0, u=user; n<user_count; u++, n++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (u->ul_year <= s[j].n) i = j + 1; else k = j - 1;
    }
    if (++k < n) memmove(s+k+1,s+k,(n-k)*sizeof(struct stat_s));
    s[k].u = u;
    s[k].n = u->ul_year;
    s[k].t = u->ul_tm_year;
  }
}

/* copy groups to stat array and sort them by gid */
int copy_groups()
{ struct group_s* g;
  int             i,j,k,m,n;

  if (!alloc_statg_mem()) return 0;
  for (m=0, n=0, g=group; m<group_count; m++, g++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (g->gid > sg[j].g->gid) i = j + 1; else k = j - 1;
    }
    if (++k < n) memmove(sg+k+1,sg+k,(n-k)*sizeof(struct statg_s));
      else sg[k].g = NULL;
    if (!sg[k].g || (g->gid != sg[k].g->gid))
    { sg[k].g = g;
      sg[k].n = sg[k].t = 0;
      n++;
    }
  }
  return n;
}

/* sort group stat array by n */
void sort_groups(int n)
{ int             i;
  int             j;
  struct statg_s* max;
  struct statg_s  tmp;

  for (i=0; i<n; i++)
  { max = sg+i;
    for (j=i+1; j<n; j++) if (sg[j].n > max->n) max = sg+j;
    if (max != sg+i)
    { tmp = *max;
      *max = sg[i];
      sg[i] = tmp;
    }
  }
}

/* calculate group month download and sort groups by it */
int sort_dl_g_month()
{ int            i,j,k,m,n;
  struct user_s* u;

  n = copy_groups();
  if (!n) return 0;
  for (m=0, u=user; m<user_count; u++, m++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (u->gid > sg[j].g->gid) i = j + 1; else
      if (u->gid < sg[j].g->gid) k = j - 1; else
      { sg[j].n += u->dl_month;
        sg[j].t += u->dl_tm_month;
        i = k + 1;
      }
    }
  }
  sort_groups(n);
  return n;
}

/* calculate group month upload and sort groups by it */
int sort_ul_g_month()
{ int            i,j,k,m,n;
  struct user_s* u;

  n = copy_groups();
  if (!n) return 0;
  for (m=0, u=user; m<user_count; u++, m++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (u->gid > sg[j].g->gid) i = j + 1; else
      if (u->gid < sg[j].g->gid) k = j - 1; else
      { sg[j].n += u->ul_month;
        sg[j].t += u->ul_tm_month;
        i = k + 1;
      }
    }
  }
  sort_groups(n);
  return n;
}

/* calculate group total download and sort groups by it */
int sort_dl_g_total()
{ int            i,j,k,m,n;
  struct user_s* u;

  n = copy_groups();
  if (!n) return 0;
  for (m=0, u=user; m<user_count; u++, m++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (u->gid > sg[j].g->gid) i = j + 1; else
      if (u->gid < sg[j].g->gid) k = j - 1; else
      { sg[j].n += u->dl_total;
        sg[j].t += u->dl_tm_total;
        i = k + 1;
      }
    }
  }
  sort_groups(n);
  return n;
}

/* calculate group total upload and sort groups by it */
int sort_ul_g_total()
{ int            i,j,k,m,n;
  struct user_s* u;

  n = copy_groups();
  if (!n) return 0;
  for (m=0, u=user; m<user_count; u++, m++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (u->gid > sg[j].g->gid) i = j + 1; else
      if (u->gid < sg[j].g->gid) k = j - 1; else
      { sg[j].n += u->ul_total;
        sg[j].t += u->ul_tm_total;
        i = k + 1;
      }
    }
  }
  sort_groups(n);
  return n;
}

/* calculate group week download and sort groups by it */
int sort_dl_g_week()
{ int            i,j,k,m,n;
  struct user_s* u;

  n = copy_groups();
  if (!n) return 0;
  for (m=0, u=user; m<user_count; u++, m++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (u->gid > sg[j].g->gid) i = j + 1; else
      if (u->gid < sg[j].g->gid) k = j - 1; else
      { sg[j].n += u->dl_week;
        sg[j].t += u->dl_tm_week;
        i = k + 1;
      }
    }
  }
  sort_groups(n);
  return n;
}

/* calculate group week upload and sort groups by it */
int sort_ul_g_week()
{ int            i,j,k,m,n;
  struct user_s* u;

  n = copy_groups();
  if (!n) return 0;
  for (m=0, u=user; m<user_count; u++, m++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (u->gid > sg[j].g->gid) i = j + 1; else
      if (u->gid < sg[j].g->gid) k = j - 1; else
      { sg[j].n += u->ul_week;
        sg[j].t += u->ul_tm_week;
        i = k + 1;
      }
    }
  }
  sort_groups(n);
  return n;
}

/* calculate group year download and sort groups by it */
int sort_dl_g_year()
{ int            i,j,k,m,n;
  struct user_s* u;

  n = copy_groups();
  if (!n) return 0;
  for (m=0, u=user; m<user_count; u++, m++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (u->gid > sg[j].g->gid) i = j + 1; else
      if (u->gid < sg[j].g->gid) k = j - 1; else
      { sg[j].n += u->dl_year;
        sg[j].t += u->dl_tm_year;
        i = k + 1;
      }
    }
  }
  sort_groups(n);
  return n;
}

/* calculate group year upload and sort groups by it */
int sort_ul_g_year()
{ int            i,j,k,m,n;
  struct user_s* u;

  n = copy_groups();
  if (!n) return 0;
  for (m=0, u=user; m<user_count; u++, m++)
  { i = 0;
    k = n - 1;
    while (i <= k)
    { j = (i + k) >> 1;
      if (u->gid > sg[j].g->gid) i = j + 1; else
      if (u->gid < sg[j].g->gid) k = j - 1; else
      { sg[j].n += u->ul_year;
        sg[j].t += u->ul_tm_year;
        i = k + 1;
      }
    }
  }
  sort_groups(n);
  return n;
}

void write_stats_file(char* basename)
{ FILE* file;
  char  filename[256];
  int   i;

  mk_filename(filename,stat_dir,DFL_STATDIR);
  i = strlen(filename);
  if (filename[i-1] != '/') filename[i++] = '/';
  strcpy(filename+i,basename);
  file = fopen(filename,"wt");
  if (!file) return;
  for (i=0; i<user_count; i++)
    fprintf(file,"%-10s %12"I64" %12"I64"\n",s[i].u->name,s[i].n,s[i].t);
  fclose(file);
}

void write_statsg_file(char* basename, int n)
{ FILE* file;
  char  filename[256];
  int   i;

  mk_filename(filename,stat_dir,DFL_STATDIR);
  i = strlen(filename);
  if (filename[i-1] != '/') filename[i++] = '/';
  strcpy(filename+i,basename);
  file = fopen(filename,"wt");
  if (!file) return;
  for (i=0; i<n; i++)
    fprintf(file,"%-10s %12"I64" %12"I64"\n",sg[i].g->name,sg[i].n,sg[i].t);
  fclose(file);
}

void write_stats_month()
{ char base[32];

  if (!alloc_stat_mem()) return;
  sprintf(base,"dl_u_m_%04i%02i",stat_year,stat_month);
  sort_dl_month();
  write_stats_file(base);
  base[0] = 'u';
  sort_ul_month();
  write_stats_file(base);
}

void write_stats_total()
{ char base[32];

  if (!alloc_stat_mem()) return;
  sprintf(base,"dl_u_t");
  sort_dl_total();
  write_stats_file(base);
  base[0] = 'u';
  sort_ul_total();
  write_stats_file(base);
}

void write_stats_week(int year)
{ char base[32];

  if (!alloc_stat_mem()) return;
  sprintf(base,"dl_u_w_%04i%02i",year,stat_week);
  sort_dl_week();
  write_stats_file(base);
  base[0] = 'u';
  sort_ul_week();
  write_stats_file(base);
}

void write_stats_year()
{ char base[32];

  if (!alloc_stat_mem()) return;
  sprintf(base,"dl_u_y_%04i",stat_year);
  sort_dl_year();
  write_stats_file(base);
  base[0] = 'u';
  sort_ul_year();
  write_stats_file(base);
}

void write_statsg_month()
{ char base[32];
  int  n;

  n = sort_dl_g_month();
  if (n)
  { sprintf(base,"dl_g_m_%04i%02i",stat_year,stat_month);
    write_statsg_file(base,n);
  }
  n = sort_ul_g_month();
  if (n)
  { sprintf(base,"ul_g_m_%04i%02i",stat_year,stat_month);
    write_statsg_file(base,n);
  }
}

void write_statsg_total()
{ char base[32];
  int  n;

  n = sort_dl_g_total();
  if (n)
  { sprintf(base,"dl_g_t");
    write_statsg_file(base,n);
  }
  n = sort_ul_g_total();
  if (n)
  { sprintf(base,"ul_g_t");
    write_statsg_file(base,n);
  }
}

void write_statsg_week(int year)
{ char base[32];
  int  n;

  n = sort_dl_g_week();
  if (n)
  { sprintf(base,"dl_g_w_%04i%02i",year,stat_week);
    write_statsg_file(base,n);
  }
  n = sort_ul_g_week();
  if (n)
  { sprintf(base,"ul_g_w_%04i%02i",year,stat_week);
    write_statsg_file(base,n);
  }
}

void write_statsg_year()
{ char base[32];
  int  n;

  n = sort_dl_g_year();
  if (n)
  { sprintf(base,"dl_g_y_%04i",stat_year);
    write_statsg_file(base,n);
  }
  n = sort_ul_g_year();
  if (n)
  { sprintf(base,"ul_g_y_%04i",stat_year);
    write_statsg_file(base,n);
  }
}

void init_stats()
{ struct tm* t;
  time_t     tim;

  time(&tim);
  t = localtime(&tim);
  t->tm_mon++;
  t->tm_year += 1900;
  if (t->tm_wday >= t->tm_yday) stat_week = 53;
    else stat_week = ((t->tm_yday-t->tm_wday) / 7) + 1;
  stat_month = t->tm_mon;
  stat_year = t->tm_year;
}

void write_stats(int force_week, int force_month, int force_year,
  int force_total, time_t tim)
{ int        new_month;
  int        new_week;
  int        new_year;
  struct tm* t;
  int        week;
  int        week_year;

  t = localtime(&tim);
  t->tm_mon++;
  t->tm_year += 1900;
  week_year = t->tm_year;
  if (t->tm_wday >= t->tm_yday) week_year--;
  if (force_week || force_month || force_year || force_total)
  { if (force_week)
    { write_stats_week(week_year);
      write_statsg_week(week_year);
    }
    if (force_month)
    { write_stats_month();
      write_statsg_month();
    }
    if (force_year)
    { write_stats_year();
      write_statsg_year();
    }
    if (force_total)
    { write_stats_total();
      write_statsg_total();
    }
  } else
  { new_year = (t->tm_year != stat_year);
    new_month = new_year || (t->tm_mon != stat_month);
    if (t->tm_wday >= t->tm_yday) week = 53;
      else week = ((t->tm_yday-t->tm_wday) / 7) + 1;
    new_week = (week != stat_week);
    if (new_week)
    { write_stats_week(week_year);
      write_statsg_week(week_year);
      reset_stats_week();
      stat_week = week;
      write_stats_total();
      write_statsg_total();
    }
    if (new_month)
    { write_stats_month();
      write_statsg_month();
      reset_stats_month();
      stat_month = t->tm_mon;
      write_stats_total();
      write_statsg_total();
    }
    if (new_year)
    { write_stats_year();
      write_statsg_year();
      reset_stats_year();
      stat_year = t->tm_year;
    }
  }
}

/* EOF */
