/*   commands.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996,1997 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include "externs.h"
#include "common.h"
#include "log.h"
#include "perm.h"

int cmd_smnt(char** p, int n)
{ return send_reply(502,"SMNT command not implemented.");
}

int cmd_rein(char** p, int n)
{ return send_reply(502,"REIN command not implemented.");
}

int cmd_stou(char** p, int n)
{ return send_reply(502,"STOU command not implemented.");
}

int cmd_allo(char** p, int n)
{ return send_reply(502,"ALLO command not implemented.");
}

int cmd_rnto(char** p, int n)
{ return send_reply(503,"Bad sequence of commands, RNFR not issued.");
}

int cmd_rnfr(char** p, int n, char* str, int strlen)
{ char*       dir;
  char*       name;
  char*       name_alt;
  char        new[256];
  int         ok;
  int         perm;
  struct stat st;
  char        wd[256];

  if (n != 1) return send_reply(501,"Syntax: RNFR <filename>");
  ok = add_dir(cwd,*p,wd,sizeof(wd));
  if (ok != -1)
  { split_dir(wd,&dir,&name,1);
    ok = permission_ok(dir,P_CHDIR);
    perm = ok < 0 ? 0 : ok;
    if (ok != -1) ok = file_permission_ok(dir,name,FP_WRITE);
    if (ok != -1)
    { if (*name)
      { *(name-1) = '/';
        name_alt = name;
      } else name_alt = NULL;
      ok = stat(wd+1,&st);
      if (ok != -1)
      { if (send_reply(350,"File exists, ready for destination name.") == -1)
          return -1;
        switch (read_ftp_command(str,strlen,p,4,&n))
        { case -1 : return -1;
          case  0 : break;
          default : return 0;
        }
        if (strcasecmp(*p,"RNTO"))
        { return send_reply(503,"Bad sequence of commands, RNTO expected.");
        }
        if (n != 2)
        { return send_reply(501,"Syntax: RNTO <filename>");
        }
        ok = add_dir(cwd,*(p+1),new,sizeof(new));
      }
      if (ok != -1)
      { split_dir(new,&dir,&name,1);
        ok = permission_ok(dir,P_WRITE);
        if (ok > 0) perm = ok;
      }
      if (ok != -1)
      { if (*name) *(name-1) = '/'; else
          if (name_alt) ok = add_dir(new,name_alt,new,sizeof(new));
        if (ok != -1)
        { ok = rename(wd+1,new+1);
          if (ok < 0) log_action(perm,LG_REN,"RENAME","%s to %s FAILED: %s",
            wd,new,strerror(errno));
          else log_action(perm,LG_REN,"RENAME","%s to %s",wd,new);
        }
      }
    }
  }
  if (ok == -1) return send_error(553,"rename",errno);
  if (!find_cdpath(wd))
  { printf("FRN- %s\n",wd);
    fgets(wd,sizeof(wd),stdin);
    if (*wd == '+')
    { printf("%s\n",new);
      fgets(wd,sizeof(wd),stdin);
    }
  }
  return send_reply(250,"RNTO command successful.");
}

int cmd_abor(char** p, int n)
{ return send_reply(226,"ABOR command successful.");
}

int cmd_dele(char** p, int n)
{ char*       dir;
  char*       name;
  int         ok;
  int32       perm;
  int         permok;
  int32       rf_gid;
  int32       rf_uid;
  struct stat st;
  char        wd[256];

  if (n != 1) return send_reply(501,"Syntax: DELE <filename>");
  ok = add_dir(cwd,*p,wd,sizeof(wd));
  if (ok != -1)
  { split_dir(wd,&dir,&name,0);
    ok = permission_ok(dir,P_CHDIR);
    permok = ok < 0 ? 0 : ok;
    if (ok != -1) ok = file_permission_ok(dir,name,FP_WRITE);
    if (ok != -1)
    { *(name-1) = '/';
      if (stat(wd+1,&st) < 0) st.st_size = 0;
      ok = unlink(wd+1);
      if (ok < 0) log_action(permok,LG_DEL,"DELETE","%s FAILED: %s",
        wd,strerror(errno));
      else log_action(permok,LG_DEL,"DELETE","%s (%lu bytes)",wd,st.st_size);
    }
  }
  if (ok == -1) return send_error(553,"delete",errno);
  read_file_permission(dir ? dir : "/",name,&rf_uid,&rf_gid,&perm);
  if (!find_cdpath(wd))
  { printf("FRM- %s\n",wd);
    fgets(wd,sizeof(wd),stdin);
  }
  if ((rf_gid != -2) && st.st_size)
  { printf("UUR %s %u\n",usr.name,(unsigned int)st.st_size);
    fgets(wd,sizeof(wd),stdin);
  }
  return send_message(253,1,dele_msg.msg,*p);
}

int cmd_syst(char** p, int n)
{ if (n) return send_reply(501,"Syntax: SYST"); else
    return send_reply(215,"UNIX Type: L8");
}

int cmd_stat(char** p, int n)
{ return send_message(211,1,
    "FTP status of server FTP4ALL "VERSION":\n"
    "Logged in as %s from %i (%h)\n"
    "There are %a anonymous (max. %A)\n"
    "and %u registered users (max. %U) logged in\n"
    "You uploaded %p and downloaded %d in the current session\n"
    "You uploaded %P and downloaded %D in all previous sessions\n"
    "Ratio of current session is %r%%, total is %R%%\n"
    "Your limit is %L, you have %c credit",usr.name);
}

int cmd_help(char** p, int n)
{ return send_reply(214,
    "The following commands are recognized (* =>'s unimplemented).\n"
    "USER    PORT    STOR    MSAM*   RNTO    NLST    MKD     CDUP\n"
    "PASS    PASV    APPE    MRSQ*   ABOR    SITE    XMKD    XCUP*\n"
    "ACCT*   TYPE    MLFL*   MRCP*   DELE    SYST    RMD     STOU*\n"
    "SMNT*   STRU    MAIL*   ALLO    CWD     STAT    XRMD    SIZE\n"
    "REIN*   MODE    MSND*   REST    XCWD    HELP    PWD     MDTM\n"
    "QUIT    RETR    MSOM*   RNFR    LIST    NOOP    XPWD    IDLE*\n"
    "Direct comments to the site administrator.");
}

int cmd_noop(char** p, int n)
{ return send_reply(200,"NOOP command successful.");
}

/* EOF */
