/*   common.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-1999 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include "tweak.h"
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include "externs.h"
#include "common.h"
#include "ftps.h"
#include "log.h"
#include "perm.h"
#include "utils.h"

#ifndef HAVE_SETENV
void setenv(char* var, char* val, int new)
{ char str[256];

  sprintf(str,"%s=%s",var,val);
  putenv(str);
}
#endif /* HAVE_SETENV */

int parse_command(char* p, char** param, int max, int* n)
{ int q;

  while (skipspace(&p))
  { if (*n >= max) return 1;
    *(param+(*n)) = p;
    (*n)++;
    q = 0;
    while (!(q & 0x04)) switch (*p)
    { case '"'  :
        if (q & 0x02) p++; else
        { q ^= 0x01;
          memmove(p,p+1,strlen(p));
        }
        break;
      case '\'' :
        if (q & 0x01) p++; else
        { q ^= 0x02;
          memmove(p,p+1,strlen(p));
        }
        break;
      case '\\' :
        memmove(p,p+1,strlen(p));
        if (*p) p++;
        break;
      case '\t' :
      case ' ' :
        if (q) p++; else q |= 0x04;
        break;
      case '\0' :
        q |= 0x04;
        break;
      default :
        p++;
        break;
    }
    if (*p) *(p++) = '\0';
  }
  return 0;
}

void sprint_bytes(char* str, int64 bytes, int aligned)
{ if ((bytes <= -0x0400) || (bytes >= 0x0400))
  { if ((bytes <= -0x00100000) || (bytes >= 0x00100000))
    { if ((bytes <= -0x40000000) || (bytes >= 0x40000000))
        sprintf(str,aligned ? "%6.1f GB" : "%3.1f GB",((float)bytes)/0x40000000);
      else
        sprintf(str,aligned ? "%6.1f MB" : "%3.1f MB",((float)bytes)/0x00100000);
    } else sprintf(str,aligned ? "%6"I64" KB" : "%"I64" KB",bytes/0x0400);
  } else sprintf(str,aligned ? "%6"I64" BY" : "%"I64" Bytes",bytes);
}

/***************************************************************/
/* enhancement for replacing color-cookies with ansi-sequences */
/***************************************************************/

void expand_color(char* p, char* buf, char* str)
{ if (!(usr.class & UF_COLOR))
  { if (*p == '!') strcpy(buf,"!"); else
    if (strchr("abcdefghABCDEFGH012345678X",*p)) buf[0] = '\0'; else
    { strncpy(buf,p-1,2);
      buf[2] = '\0';
    }
  } else switch (*p)
  { /* all attributes off */
    case '0' : strcpy(buf,"\x1B[0m"); break;
    /*  a standard ! */
    case '!' : strcpy(buf,"!"); break;
    /* blinking */
    case 'X' : strcpy(buf,"\x1B[5m"); break;
    /*********************/
    /* FOREGROUND COLORS */
    /*********************/
    /* FG black */
    case 'a' : strcpy(buf,"\x1B[0;30m"); break;
    /* FG red */
    case 'b' : strcpy(buf,"\x1B[0;31m"); break;
    /* FG green */
    case 'c' : strcpy(buf,"\x1B[0;32m"); break;
    /* FG brown */
    case 'd' : strcpy(buf,"\x1B[0;33m"); break;
    /* FG blue */
    case 'e' : strcpy(buf,"\x1B[0;34m"); break;
    /* FG magenta */
    case 'f' : strcpy(buf,"\x1B[0;35m"); break;
    /* FG cyan */
    case 'g' : strcpy(buf,"\x1B[0;36m"); break;
    /* FG light gray */
    case 'h' : strcpy(buf,"\x1B[0;37m"); break;
    /* FG dark gray */
    case 'A' : strcpy(buf,"\x1B[1;30m"); break;
    /* FG light red */
    case 'B' : strcpy(buf,"\x1B[1;31m"); break;
    /* FG light green */
    case 'C' : strcpy(buf,"\x1B[1;32m"); break;
    /* FG yellow */
    case 'D' : strcpy(buf,"\x1B[1;33m"); break;
    /* FG light blue */
    case 'E' : strcpy(buf,"\x1B[1;34m"); break;
    /* FG light magenta */
    case 'F' : strcpy(buf,"\x1B[1;35m"); break;
    /* FG light cyan */
    case 'G' : strcpy(buf,"\x1B[1;36m"); break;
    /* FG white */
    case 'H' : strcpy(buf,"\x1B[1;37m"); break;
    /*********************/
    /* BACKGROUND COLORS */
    /*********************/
    /* BG black */
    case '1' : strcpy(buf,"\x1B[1;40m"); break;
    /* BG red */
    case '2' : strcpy(buf,"\x1B[1;41m"); break;
    /* BG green */
    case '3' : strcpy(buf,"\x1B[1;42m"); break;
    /* BG brown */
    case '4' : strcpy(buf,"\x1B[1;43m"); break;
    /* BG blue */
    case '5' : strcpy(buf,"\x1B[1;44m"); break;
    /* BG magenta */
    case '6' : strcpy(buf,"\x1B[1;45m"); break;
    /* BG cyan */
    case '7' : strcpy(buf,"\x1B[1;46m"); break;
    /* BG white */
    case '8' : strcpy(buf,"\x1B[1;47m"); break;
    default  :
      strncpy(buf,p-1,2);
      buf[2] = '\0';
      break;
  }  
}

void server_expand_variable(char c, char* buf)
{ printf("VAR_EXP %c\n",c);
  readline(buf,32,stdin);
  if (*buf == '+') memmove(buf,buf+1,strlen(buf)); else sprintf(buf,"%%%c",c);
}

void expand_variable(char** p, int* len, char* buf, char* str)
{ struct sockaddr_in addr;
  struct hostent*    hp;
  int                n;
  time_t             t;
  struct top_s*      top;

  switch (**p)
  { case '%' :
      strcpy(buf,"%");
      break;
    case 'a' : /* num anon */
      server_expand_variable(**p,buf);
      break;   /* max anon */
    case 'A' :
      server_expand_variable(**p,buf);
      break;
    case 'c' : /* credit */
      if (usr.limit < 0) strcpy(buf,"upload-only"); else
        if (usr.limit) sprint_bytes(buf,usr.credit,0); else
        strcpy(buf,"unlimited");
      break;
    case 'd' : /* download current session */
      sprint_bytes(buf,bytes_downloaded,0);
      break;
    case 'D' : /* download all sessions */
      sprint_bytes(buf,usr.downloaded,0);
      break;
    case 'h' : /* hostname */
      addr.sin_addr.s_addr = htonl(ipaddr);
      hp = gethostbyaddr((char*)&addr.sin_addr,sizeof(addr.sin_addr),
        AF_INET);
      if (hp) strcpy(buf,hp->h_name); else
        strcpy(buf,inet_ntoa(addr.sin_addr));
      break;
    case 'i' : /* ip addr */
      addr.sin_addr.s_addr = htonl(ipaddr);
      strcpy(buf,inet_ntoa(addr.sin_addr));
      break;
    case 'k' : /* last speed */
      sprintf(buf,"%3.1f",((float)last_speed)/1024);
      break;
    case 'l' : /* last login */
      strftime(buf,32,"%a %b %d, %Y %H:%M",localtime(&usr.last_login));
      break;
    case 'L' : /* limit */
      n = usr.limit;
      if (n < 0) strcpy(buf,"upload-only"); else
        if (!n) strcpy(buf,"none"); else
        if ((n < 100) && ((100 % n) <= 1)) sprintf(buf,"1:"SI32,100/n); else
        sprintf(buf,SI32,usr.limit);
      break;
    case 'n' : /* num users */
      server_expand_variable(**p,buf);
      break;
    case 'N' : /* max users */
      server_expand_variable(**p,buf);
      break;
    case 'p' : /* upload current session */
      sprint_bytes(buf,bytes_uploaded,0);
      break;
    case 'P' : /* download current session */
      sprint_bytes(buf,usr.uploaded,0);
      break;
    case 'q' :
      if (!bytes_uploaded) strcpy(buf,"n/a"); else
        sprintf(buf,"%3.1f",(float)bytes_downloaded/bytes_uploaded);
      break;
    case 'Q' :
      if (!usr.uploaded) strcpy(buf,"n/a"); else
        sprintf(buf,"%3.1f",(float)usr.downloaded/usr.uploaded);
      break;
    case 'r' :
      if (!bytes_downloaded) strcpy(buf,"n/a"); else
        sprintf(buf,"%3.1f",(float)bytes_uploaded*100/bytes_downloaded);
      break;
    case 'R' :
      if (!usr.downloaded) strcpy(buf,"n/a"); else
        sprintf(buf,"%3.1f",(float)usr.uploaded*100/usr.downloaded);
      break;
    case 's' : /* special */
      strcpy(buf,str ? str : "%s");
      break;
    case 'S' : /* username */
      strcpy(buf,usr.name);
      break;
    case 't' : /* local time */
      time(&t);
      strftime(buf,32,"%a %b %d, %Y %H:%M",localtime(&t));
      break;
    case 'T' : /* top ten, format: %T{0-9}[m|t|w|y]{d|e|l|u}{a|n|i|m} */
      if ((*len >= 3) && (*(*p+1) >= '0' && *(*p+1) <= '9'))
      { n = *(*p+1) - '0';
        if (!n) n = 10;
        n--;
        /* get time range (week, month, year, total) */
        switch (*(*p+2))
        { case 'm' : top = top_m; break;
          case 't' : top = top_t; break;
          case 'w' : top = top_w; break;
          case 'y' : top = top_y; break;
          default  : top = NULL;
        }
        if (top)
        { (*len)--;
          (*p)++;
        } else top = top_t;
        /* get list (download, upload, elite, lame) */
        switch (*(*p+2))
        { case 'd' :
            switch (*(*p+3))
            { case 'a' : sprint_bytes(buf,top[n].d_amount,1); break;
              case 'n' : sprintf(buf,"%-9s",top[n].d_name); break;
              case 'i' : sprintf(buf,"%-15s",top[n].d_info); break;
              case 'm' : strcpy(buf,top[n].d_name);
                         sprintf(buf+strlen(buf)," %-*s",24-strlen(buf),
                           top[n].d_info);
                         break;
              default  : strncpy(buf,*p-1,5); *(buf+5) = '\0'; break;
            }
            break;
          case 'e' :
            switch (*(*p+3))
            { case 'a' : sprintf(buf,"%7.1f",top[n].e_amount); break;
              case 'n' : sprintf(buf,"%-9s",top[n].e_name); break;
              case 'i' : sprintf(buf,"%-15s",top[n].e_info); break;
              case 'm' : strcpy(buf,top[n].e_name);
                         sprintf(buf+strlen(buf)," %-*s",24-strlen(buf),
                           top[n].e_info);
                         break;
              default  : strncpy(buf,*p-1,5); *(buf+5) = '\0'; break;
            }
            break;
          case 'l' :
            switch (*(*p+3))
            { case 'a' : sprintf(buf,"%7.1f",top[n].l_amount); break;
              case 'n' : sprintf(buf,"%-9s",top[n].l_name); break;
              case 'i' : sprintf(buf,"%-15s",top[n].l_info); break;
              case 'm' : strcpy(buf,top[n].l_name);
                         sprintf(buf+strlen(buf)," %-*s",24-strlen(buf),
                           top[n].l_info);
                         break;
              default  : strncpy(buf,*p-1,5); *(buf+5) = '\0'; break;
            }
            break;
          case 'u' :
            switch (*(*p+3))
            { case 'a' : sprint_bytes(buf,top[n].u_amount,1); break;
              case 'n' : sprintf(buf,"%-9s",top[n].u_name); break;
              case 'i' : sprintf(buf,"%-15s",top[n].u_info); break;
              case 'm' : strcpy(buf,top[n].u_name);
                         sprintf(buf+strlen(buf)," %-*s",24-strlen(buf),
                           top[n].u_info);
                         break;
              default  : strncpy(buf,*p-1,5); *(buf+5) = '\0'; break;
            }
            break;
          default  : strncpy(buf,*p-1,5); *(buf+5) = '\0'; break;
        }
        *len -= 3;
        *p += 3;
      } else
      { strncpy(buf,*p-1,2);
        *(buf+2) = '\0';
      }
      break;
    case 'u' : /* num users */
      server_expand_variable(**p,buf);
      break;
    case 'U' : /* max users */
      server_expand_variable(**p,buf);
      break;
    case 'v' : /* short version */
      strcpy(buf,SVERSION);
      break;
    default :
      strncpy(buf,(*p)-1,2);
      buf[2] = '\0';
      break;
  }
}

int send_message_part(int rno, char* p, int len, char* str, int finish,
  int expand)
{ char  buf[1024];
  int   n;
  char* param[32];

  if ((len > 3) && !strncmp(p,"%#!",3))
  { len -= 3;
    p += 3;
    strcpy(buf,"");
    while (len)
    { if ((*p == '%') && (expand & 0x01) && (len > 1))
      { len--;
        p++;
        expand_variable(&p,&len,buf+strlen(buf),str);
        if (len)
        { len--;
          p++;
        }
      } else if ((*p == '!') && (expand & 0x02) && (len > 1))
      { len--;
        p++;
        expand_color(p,buf+strlen(buf),str);
        if (len)
        { len--;
          p++;
        }
      } else
      { switch (*p)
        { case '\n' :
            strcat(buf," ");
          case '\r' :
            break;
          default :
            *(buf+strlen(buf)+1) = '\0';
            *(buf+strlen(buf)) = *p;
        }
        len--;
        p++;
      }
    }
    n = 0;
    if (parse_command(buf,param,32,&n))
      fprintf(sockf,"%i%c%s\r\n",rno,finish ? '-' : ' ',buf);
      else run_program(*param,param,n,rno,finish,cwd+1,&n,NULL);
    return 0;
  }

  fprintf(sockf,"%i%c",rno,finish ? ' ' : '-');
  while (len)
  { if ((*p == '%') && (expand & 0x01) && (len > 1))
    { len--;
      p++;
      expand_variable(&p,&len,buf,str);
      fprintf(sockf,"%s",buf);
      if (len)
      { len--;
        p++;
      }
    } else if ((*p == '!') && (expand & 0x02) && (len > 1))
    { len--;
      p++;
      expand_color(p,buf,str);
      fprintf(sockf,"%s",buf);
      if (len)
      { len--;
        p++;
      }
    } else 
    { switch (*p)
      { case '\n' :
          fprintf(sockf,"\r\n");
          break;
        case '\r' :
          break;
        default :
          fputc(*p,sockf);
      }
      len--;
      p++;
    }
  }
  if (*(p-1) != '\n') fprintf(sockf,"\r\n");
  return 0;
}

int send_message_exp(int rno, int finish, char* buffer, char* str, int expand)
{ char* p = buffer;

/*errlog("sending <%s> <%s>",buffer,str);*/
  if (!buffer) { if (finish) return -1; else return 0; }
  do
  { while (*p && (*p != '\n')) p++;
    if (*p) p++;
    if (send_message_part(rno,buffer,p-buffer,str,finish && !*p,expand) == -1)
      return -1;
    buffer = p;
  } while (*p);
  return 0;
}

int send_message(int rno, int finish, char* buffer, char* str)
{ return send_message_exp(rno,finish,buffer,str,3);
}

int send_reply_exp(int rno, char* reply, int expand)
{ return send_message_exp(rno,1,reply,NULL,expand);
}

int send_reply(int rno, char* reply)
{ return send_message_exp(rno,1,reply,NULL,3);
}

int send_error(int rno, char* cmd, int errn)
{ char str[256];

  strcpy(str,strerror(errn));
  sprintf(str+strlen(str)," (%s).",cmd);
  return send_reply_exp(rno,str,0);
}

char class_char(void)
{ switch (usr.class & UC_MASK)
  { case UC_ROOT    : return 's';
    case UC_ANON    : return 'a';
    case UC_NORMAL  : return 'n';
    case UC_GROUPOP : return 'g';
    case UC_SITEOP  : return 'o';
    default         : return '-';
  }
}

int run_program(char* prog, char** param, int n, int rno, int finish, char* cd,
  int* exitcode, char* buffer)
{ char* c;
  int   err;
  FILE* file;
  pid_t pid;
  int   pipefd[2];
  char  str[256];
  char  str2[256];
  char envbuffer[6][20];
  char* temp;

  err = socketpair(AF_UNIX,SOCK_STREAM,0,pipefd);
  if (err == -1) return -1;
  fflush(stdout);

  sprintf(envbuffer[0],"F4A_USER=%s",usr.name);
  putenv(envbuffer[0]);
  sprintf(envbuffer[1],"F4A_CLASS=%c",class_char());
  putenv(envbuffer[1]);
  sprintf(envbuffer[2],"F4A_INVISIBLE=%c",invisible() ? 'i' : '-');
  putenv(envbuffer[2]);
  sprintf(envbuffer[3],"F4A_COLOR=%c",coloruser() ? 'c' : '-');
  putenv(envbuffer[3]);
  sprintf(envbuffer[4],"F4A_GROUP=%s",get_groupname(usr.gid));
  putenv(envbuffer[4]);
  sprintf(envbuffer[5],"F4A_APASS=%s",anonpass);
  putenv(envbuffer[5]);
  temp = server_dir + strlen(server_dir)+9;
  strcpy(temp,"F4A_STD=");strcat(temp,server_dir);
  putenv(temp);

  pid = fork();
  if (pid == -1) return -1;
  if (pid)
  { close(pipefd[1]);
    file = fdopen(pipefd[0],"rt");
    if (!file) return -1;
    if (buffer) *buffer = '\0';
    if (finish)
    { c = fgets(str,sizeof(str)-1,file);
      if (!c) strcpy(str,"(no output from program)");
    } else c = (char*)-1;
    while (c)
    { c = fgets(str2,sizeof(str2)-1,file);
      if (c && strlen(str2))
      { if (buffer) strcat(buffer,str2); else
        { if (finish)
          { if (send_message_exp(rno,0,str,NULL,3) == -1) return -1;
            strcpy(str,str2);
          } else
          { if (send_message_exp(rno,0,str2,NULL,3) == -1) return -1;
          }
        }
      }
    }
    fclose(file);
    close(pipefd[0]);
    pid = wait(exitcode);
    if (finish) return send_reply_exp(rno,str,3);
    return 0;
  } else
  { fclose(sockf);
    fclose(sockfr);
    if (cd && *cd) chdir(cd);
    close(pipefd[0]);
    dup2(pipefd[1],0);
    dup2(pipefd[1],1);
    dup2(pipefd[1],2);
    close(pipefd[1]);
    *(param+n) = NULL;
    execvp(prog,param);
    exit(-1);
  }
}

int run_prog_xt(char* prog, char **param, int* exitcode, char* returner)
{ int   err;
  FILE* file;
  pid_t pid;
  int   pipefd[2];
  char  str[255];
  char* c;
  char* p;
  char envbuffer[6][20];
  char* temp;

  err = socketpair(AF_UNIX,SOCK_STREAM,0,pipefd);
  if (err == -1) return -1;
  fflush(stdout);

  sprintf(envbuffer[0],"F4A_USER=%s",usr.name);
  putenv(envbuffer[0]);
  sprintf(envbuffer[1],"F4A_CLASS=%c",class_char());
  putenv(envbuffer[1]);
  sprintf(envbuffer[2],"F4A_INVISIBLE=%c",invisible() ? 'i' : '-');
  putenv(envbuffer[2]);
  sprintf(envbuffer[3],"F4A_COLOR=%c",coloruser() ? 'c' : '-');
  putenv(envbuffer[3]);
  sprintf(envbuffer[4],"F4A_GROUP=%s",get_groupname(usr.gid));
  putenv(envbuffer[4]);
  sprintf(envbuffer[5],"F4A_APASS=%s",anonpass);
  putenv(envbuffer[5]);
  temp = server_dir + strlen(server_dir)+9;
  strcpy(temp,"F4A_STD=");strcat(temp,server_dir);
  putenv(temp);

  pid = fork();
  if (pid == -1) return -1;
  if (pid)
  { close(pipefd[1]);
    file = fdopen(pipefd[0],"rt");
    if (!file) return -1;
    c = fgets(str,sizeof(str)-1,file);          /* get one line of the script */
    if (!c) sprintf(str,"no special reason\n");
    if ((p = strrchr(str,'\n'))) *p = 0;
    strcpy(returner,str);  /* paste it to the calling-proc without "\n" */
    fclose(file);
    close(pipefd[0]);
    pid = wait(exitcode);
    return 0;
  } else
  { fclose(sockf);
    fclose(sockfr);
    close(pipefd[0]);
    dup2(pipefd[1],0);
    dup2(pipefd[1],1);
    dup2(pipefd[1],2);
    close(pipefd[1]);
    *(param+2) = NULL;
    execvp(prog,param);
    exit(-1);
  }
}

int convert_buffer(char* buffer, int len)
{ int i = 0;

  while (i < len) switch ((unsigned char)*buffer)
  { case 0xF2 :
    case 0xF4 :
    case 0xFF :
      len--;
      if (len > i) memmove(buffer,buffer+1,len-i);
      break;
    default :
      buffer++;
      i++;
  }
  return i;
}

int recvb(char* buffer, int len)
{ fd_set         fd;
  int            i;
  struct timeval t;

  FD_ZERO(&fd);
  FD_SET(fileno(sockfr),&fd);
  t.tv_sec = timeout;
  t.tv_usec = 0;
  i = select(fileno(sockfr)+1,&fd,NULL,NULL,&t);
  if (i == -1) return -1;
  if (!i)
  { quit(CL_TIMEOUT);
    errno = 0;
    return -1;
  }
  if (!fgets(buffer,len-1,sockfr))
  { quit(CL_NORMAL);
    errno = 0;
    return -1;
  }
  return convert_buffer(buffer,strlen(buffer));
}

int read_ftp_command(char* buffer, int bufferlen, char** param, int max, int* n)
{ int   len = 0;
  char* p;
  int   q;

  *n = 0;
  do
  { if ((q = recvb(buffer+len,bufferlen-len)) == -1) return -1;
    len += q;
    if ((len >= bufferlen) && (*(buffer+len-1) != '\n'))
    { do
      { q = recvb(buffer,bufferlen);
        if (q == -1) return -1;
      } while (*(buffer+q-1) != '\n');
      if (send_reply_exp(500,"Command string too long.",0) == -1) return -1;
      return 1;
    }
  } while (!len && (*(buffer+len-1) != '\n'));
  if (len && (*(buffer+len-1) == '\n')) len--;
  if (len && (*(buffer+len-1) == '\r')) len--;
  *(buffer+len) = '\0';
/*errlog("received <%s>",buffer);*/
  p = buffer;
  if (!skipspace(&p))
  { if (send_reply_exp(500,"Command string missing command.",0) == -1) return -1;
    return 1;
  }
  *param = p;
  *n = 1;
  if (skipnspace(&p)) *(p++) = '\0'; else return 0;
  if (skipspace(&p))
  { *(param+1) = p;
    (*n)++;
  }
  return 0;
}

/* EOF */
