/*   config.c
 * 
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-1999 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include "externs.h"
#include "common.h"
#include "log.h"
#include "utils.h"

int get_cdpath(char* buf)
{ char* cd;
  int32 gid;
  int   n;
  char* p[8];
  int32 perm;
  int32 uid;

  n = 0;
  parse_command(buf,p,8,&n);
  if (n != 4) return -1;
  if (!sscanf(p[1],SI32,&uid) || !sscanf(p[2],SI32,&gid) ||
      !sscanf(p[3],SI32,&perm)) return -1;
  while (cdpath_len+strlen(p[0])+32 >= cdpath_size)
  { n = cdpath_size + 1024;
    cd = alloc(cdpath,n);
    if (!cd)
    { errlog("get_cdpath(): %s (%i)",strerror(errno),errno);
      return -2;
    }
    cdpath = cd;
    cdpath_size = n;
  }
  strcpy(cdpath+cdpath_len,p[0]);
  cd = cdpath+cdpath_len+((strlen(p[0])+8) & ~0x07);
  *(int32*)cd = uid;
  *(int32*)(cd+sizeof(int32)) = gid;
  *(int32*)(cd+(sizeof(int32)*2)) = perm;
  cd += sizeof(int32) * 3;
  cdpath_len = cd-cdpath;
  return 0;
}

int load_cdpath(void)
{ char  buf[512];
  FILE* file;

  mk_filename(buf,"etc/cdpath.cfg",NULL);
  file = fopen(buf,"rt");
  if (!file) return -1;
  cdpath_len = 0;
  while (readline(buf,sizeof(buf),file))
  { if (*buf != '#') get_cdpath(buf);
  }
  fclose(file);
  return 0;
}

int get_check(char* p, struct check_s** check, int* check_count,
  int* check_size)
{ struct check_s* ch;
  char            name[sizeof(ch->name)];
  char            param[sizeof(ch->param)];
  char*           q = p;
  char            wildcard[sizeof(ch->wildcard)];

  if (skipnspace(&q))
  { if (q-p >= sizeof(wildcard)) return -1;
    strncpy(wildcard,p,q-p);
    *(wildcard+(q-p)) = '\0';
    if (!skipspace(&q)) return -1;
    p = q;
  } else return -1;
  skipnspace(&q);
  if (q-p >= sizeof(name)) return -1;
  strncpy(name,p,q-p);
  *(name+(q-p)) = '\0';
  if (*q)
  { if (skipspace(&q))
    { p = q;
      skipnspace(&q);
      if (q-p >= sizeof(param)) return -1;
      strncpy(param,p,q-p);
      *(param+(q-p)) = '\0';
      if (skipspace(&q)) return -1;
    } else *param = '\0';
  } else *param = '\0';
  if (*check_count == *check_size)
  { ch = (struct check_s*)alloc((char*)*check,
      (*check_size+4)*sizeof(struct check_s));
    if (!ch)
    { errlog("get_check(): %s (%i)",strerror(errno),errno);
      return -1;
    }
    *check = ch;
    *check_size += 4;
  }
  ch = *check + *check_count;
  (*check_count)++;
  strcpy(ch->wildcard,wildcard);
  strcpy(ch->name,name);
  strcpy(ch->param,param);
  return 0;
}

int load_check(void)
{ char  buf[512];
  FILE* file;

  mk_filename(buf,"etc/check.cfg",NULL);
  file = fopen(buf,"rt");
  if (!file) return -1;
  check_count = 0;
  while (readline(buf,sizeof(buf),file))
  { if (*buf != '#') get_check(buf,&check,&check_count,&check_size);
  }
  fclose(file);
  return 0;
}

int load_dcheck(void)
{ char  buf[512];
  FILE* file;

  mk_filename(buf,"etc/checkdupe.cfg",NULL);
  file = fopen(buf,"rt");
  if (!file) return -1;
  dcheck_count = 0;
  while (readline(buf,sizeof(buf),file))
  { if (*buf != '#') get_check(buf,&dcheck,&dcheck_count,&dcheck_size);
  }
  fclose(file);
  return 0;
}

int get_dshortcut(char* buf)
{ char* dsh;
  int   l1;
  int   l2;
  int   n;
  char* p[8];

  n = 0;
  parse_command(buf,p,8,&n);
  if (n != 2) return -1;
  l1 = strlen(p[0]);
  l2 = strlen(p[1]);
  while (dshortcut_len+l1+l2+2 >= dshortcut_size)
  { dsh = alloc(dshortcut,dshortcut_size+256);
    if (!dsh)
    { errlog("get_dshortcut(): %s (%i)",strerror(errno),errno);
      return -1;
    }
    dshortcut = dsh;
    dshortcut_size += 256;
  }
  dsh = dshortcut + dshortcut_len;
  strcpy(dsh,p[0]);
  dsh += l1 + 1;
  strcpy(dsh,p[1]);
  dshortcut_len += l1 + l2 + 2;
  return 0;
}

int load_dshortcut(void)
{ char  buf[512];
  FILE* file;

  mk_filename(buf,"etc/dirshortcut.cfg",NULL);
  file = fopen(buf,"rt");
  if (!file) return -1;
  dshortcut_len = 0;
  while (readline(buf,sizeof(buf),file))
  { if (*buf != '#') get_dshortcut(buf);
  }
  fclose(file);
  return 0;
}

/* EOF */
