/*   dir.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-1999 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include "tweak.h"
#include "externs.h"
#include "common.h"
#include "ff.h"
#include "list.h"
#include "log.h"
#include "perm.h"
#include "utils.h"

int show_readmefile(void)
{ FILE* file;
  char  line[256];

  if (readme_file) if ((file = file_open(cwd,readme_file,"rt",0)) != NULL)
  { while (fgets(line,sizeof(line),file))
    { if (send_message(250,0,line,cwd) == -1)
      { fclose(file);
        return -1;
      }
    }
    fclose(file);
  }
  return 0;
}

/* link extension by Buba */
void is_link(char *fullp, char *split)
{ char link[256];
  char *p;

  memset(split,0,256);
  while(strcmp(fullp,""))
  { memset(link,0,sizeof(link));
    if (readlink(fullp,link,255)!=-1)
    { if (link[0] != '/')
      { *(strrchr(fullp,'/')+1) = 0;
        strcat(fullp,link);
      } else strcpy(fullp,link);
      strcat(fullp,split);
      strcpy(split,"");
    } else
    { strcpy(link,strrchr(fullp,'/')); /* Use of link here to save vars */
      strcat(link,split);
      strcpy(split,link);
      *strrchr(fullp,'/') = 0;
    }
  }
  /* split var should'n have now sym links
     Now strip .. or . from split (Usefull when someone made a link
     ../../.... out of sitedir path) */
  while ((p = strstr(split,"/..")) != NULL)
  { *p = 0;
    strcpy(link,p+3);
    *strrchr(split,'/') = 0;
    strcat(split,link);
  }
  while ((p = strstr(split,"/.")) != NULL)
  { *p=0;
    strcat(split,p+2);
  }
}

int find_matching_dir(char* path, char* wildcard)
{ struct dirent* d;
  DIR*           dir;
  char*          name;
  struct stat    st;

  dir = opendir(strcmp(path,"/") ? path+1 : ".");
  if (!dir) return 0;
  if (strcmp(path,"/")) strcat(path,"/");
  name = path + strlen(path);
  while ((d = readdir(dir)) != NULL)
  { strcpy(name,d->d_name);
    if (strcmp(name,".") && strcmp(name,"..") && !stat(path+1,&st) &&
      S_ISDIR(st.st_mode))
    { if (wildcard_matching(wildcard,name,0))
      { closedir(dir);
        return 1;
      }
    }
  }
  closedir(dir);
  return 0;
}

char* is_shortcut(char* shortcutpath)
{ int n;

  n = 0;
  while (n < dshortcut_len)
  { if (!strcasecmp(shortcutpath,dshortcut+n))
    { n += strlen(dshortcut+n) + 1;
      return dshortcut + n;
    }
    n += strlen(dshortcut+n) + 1;
    n += strlen(dshortcut+n) + 1;
  }
  return NULL;
}

int cmd_cwd(char** p, int n)
{ char  basedir[256];
  char* dir;
  char  fullpath[256];
  char* name;
  char* shortcutpath;
  char  wd[256];

  if (n != 1) return send_reply(501,"Syntax: CWD <directory>");
  if ((shortcutpath = is_shortcut(*p)) != NULL) strcpy(*p,shortcutpath); else
  if (flag_get(flags,FLAG_RESOLVESYMLINKS) && strcmp(*p,"/") && strcmp(*p,".."))
  { /* symlink extension by Buba */
    getcwd(basedir,sizeof(basedir)); /* Always return the same as ftpd basedir var */
    if (**p == '/')
    { /* When doing cd /path/to... */
      sprintf(fullpath,"%s%s",basedir,*p);
    } else
    { /* When doing cd path/to... */
      sprintf(fullpath,"%s%s/%s",basedir,strcmp(cwd,"/") ? cwd : "",*p);
    }
    is_link(fullpath,wd); /* function that resolve links and allocate true
                             path in link var */
    if (!strncmp(wd,basedir,strlen(basedir)-1)) /* link is in site dir */
      strcpy(*p,wd+strlen(basedir));
    /* If link is out of site dir we do nothing */
  }
  n = add_dir(cwd,*p,wd,sizeof(wd));
  if (n != -1)
  { n = permission_ok(*p,P_CHDIR);
    if (n == -1) if (errno == ENOENT)
    { split_dir(wd,&dir,&name,0);
      strcpy(*p,name);
      if (!dir) strcpy(wd,"/");
      if (find_matching_dir(wd,*p)) n = permission_ok(wd,P_CHDIR); else
      { n = -1;
        errno = ENOENT;
      }
    }
  }
  if (n == -1) return send_error(550,"chdir",errno);
  strcpy(cwd,wd);
  if (show_readmefile() == -1) return -1;
  return send_message(250,1,chdir_msg.msg,wd);
}
 
int cmd_cdup(char** p, int n)
{ char wd[256];

  if (n) return send_reply(501,"Syntax: CDUP");
  strcpy(*p,"..");
  n = add_dir(cwd,*p,wd,sizeof(wd));
  if ((n == -1) || (permission_ok(*p,P_CHDIR) == -1))
    return send_error(550,"cdup",errno);
  strcpy(cwd,wd);
  if (show_readmefile() == -1) return -1;
  return send_message(250,1,cdup_msg.msg,cwd);
}

int cmd_mkd(char** p, int n)
{ char* dir;
  char* name;
  int   ok;
  int   perm;
  char  wd[256];

  if (n != 1) return send_reply(501,"Syntax: MKD <directory>");
  ok = add_dir(cwd,*p,wd,sizeof(wd));
  if (ok >= 0)
  { split_dir(wd,&dir,&name,0);
    convert_filename(name,fconv_mode >> 4);
    ok = permission_ok(dir,P_MKDIR);
    if (ok >= 0)
    { perm = ok;
      *(name-1) = '/';
      ok = mkdir(wd+1,0777);
      if (ok < 0) log_action(perm,LG_MD,"MKDIR","%s FAILED: %s",
        wd,strerror(errno)); else
      { log_action(perm,LG_MD,"MKDIR","%s",wd);
        ok = write_permission_file(wd,usr.uid,usr.gid,
          umask_value | (perm << 24));
        if ((ok < 0) && (errno == EACCES)) ok = 0;
      }
    }
  }
  if (ok < 0) return send_error(550,"mkdir",errno);
  return send_message(257,1,mkdir_msg.msg,name);
}

int directory_empty(char* wd)
{ DIR*           d;
  struct dirent* de;

  if (!strcmp(wd,"/"))
  { errno = EACCES;
    return -1;
  }
  d = opendir(wd+1);
  if (!d) return -1;
  de = readdir(d);
  while (de)
  { if (strcmp(de->d_name,".") && strcmp(de->d_name,"..") &&
        strcmp(de->d_name,permission_file))
    { if (!readme_file || strcmp(de->d_name,readme_file))
      { closedir(d);
        return 0;
      }
    }
    de = readdir(d);
  }
  closedir(d);
  return 1;
}

int cmd_rmd(char** p, int n)
{ int   ok;
  int   perm;
  char  wd[256];

  if (n != 1) return send_reply(501,"Syntax: RMD <directory>");
  ok = add_dir(cwd,*p,wd,sizeof(wd));
  if (ok >= 0)
  { if (!strcmp(wd,cwd))
      return send_reply(553,"Cannot remove current directory.");
    ok = permission_ok(wd,P_CHMOD);
    if (ok >= 0)
    { perm = ok;
      ok = directory_empty(wd);
      if (ok != -1)
      { if (!ok) return send_reply(553,"Directory is not empty.");
        ok = remove_permission_file(wd);
        if (ok != -1) ok = rmdir(wd+1);
        if (ok < 0) log_action(perm,LG_RD,"RMDIR","%s FAILED: %s",
          wd,strerror(errno));
        else log_action(perm,LG_RD,"RMDIR","%s",wd);
      }
    }
  }
  if (ok < 0) return send_error(553,"rmdir",errno);
  return send_message(253,1,rmdir_msg.msg,*p);
}

int cmd_pwd(char** p, int n)
{ if (n) return send_reply(501,"Syntax: PWD");
  return send_message(257,1,"\"%s\" is current directory.",cwd);
}

/* EOF */
