/*   list.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-1999 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include "tweak.h"
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include "externs.h"
#include "common.h"
#include "log.h"
#include "perm.h"
#include "utils.h"

#define LS_DOT       0x01
#define LS_LONG      0x02
#define LS_RECURSIVE 0x04
#define LS_TYPECHAR  0x08
#define LS_REVERSE   0x10
#define LS_DIRECTORY 0x20

#define SORT_MASK    0x0700
#define SORT_NONE    0x0000
#define SORT_NAME    0x0100
#define SORT_SIZE    0x0200
#define SORT_MTIME   0x0300
#define SORT_ATIME   0x0400
#define SORT_CTIME   0x0500

struct fileinfo_s
{ long size;
  long time;
  long nlink;
  long mode;
} ;

/* wildcard matching, case insensitive */
int wm_ci(unsigned char* name, unsigned char* wc)
{ unsigned char  c1;
  unsigned char  c2;
  unsigned char* ptr;
  int            r;

  while (*name && *wc)
  { switch (*wc)
    { case '?' :
        break;
      case '*' :
        if (!(*(++wc))) return 1;
        do
        { r = wm_ci(name,wc);
          if (r) return r;
        } while (*(++name));
        return 0;
        break;
      case '[' :
        ptr = ++wc;
        while (*ptr && (*ptr != ']') && (*(ptr-1) != '\\')) ptr++;
        if (!(*ptr)) return -1;
        r = 1;
        ptr = wc;
        c1 = *(ptr++);
        if (c1 == '\\') c1 = *(ptr++);
          else if (c1 == ']') return -1;
        if (*(ptr++) == '-')
        { c2 = *(ptr++);
          if (c2 == '\\') c2 = *(ptr++);
            else if (c2 == ']') r = 0;
          if (*ptr != ']') r = 0;
        } else
        { c2 = '\0';
          r = 0;
        }
        if (r)
        { /* format [a-b] */
          c1 = tolower(c1);
          c2 = tolower(c2);
          if ((tolower(*name) < c1) || (tolower(*name) > c2))
          { c1 = toupper(c1);
            c2 = toupper(c2);
            if ((toupper(*name) < c1) || (toupper(*name) > c2)) return 0;
          }
          wc = ptr;
        } else
        { /* format [abc] */
          while (*wc != ']')
          { if (*wc == '\\') wc++;
            if (tolower(*name) == tolower(*wc))
            { while ((*wc != ']') && (*(wc-1) != '\\')) wc++;
            } else wc++;
          }
        }
        break;
      default  :
        if (*wc == '\\') if (!(*(++wc))) return -1;
        if (tolower(*wc) != tolower(*name)) return 0;
        break;
    }
    name++;
    wc++;
  }
  return (*name == *wc) || !strcmp(wc,"*") ? 1 : 0;
}

/* wildcard matching, case sensitive */
int wm_cs(unsigned char* name, unsigned char* wc)
{ unsigned char  c1;
  unsigned char  c2;
  unsigned char* ptr;
  int            r;

  while (*name && *wc)
  { switch (*wc)
    { case '?' :
        break;
      case '*' :
        if (!(*(++wc))) return 1;
        do
        { r = wm_cs(name,wc);
          if (r) return r;
        } while (*(++name));
        return 0;
        break;
      case '[' :
        ptr = ++wc;
        while (*ptr && (*ptr != ']') && (*(ptr-1) != '\\')) ptr++;
        if (!(*ptr)) return -1;
        r = 1;
        ptr = wc;
        c1 = *(ptr++);
        if (c1 == '\\') c1 = *(ptr++);
          else if (c1 == ']') return -1;
        if (*(ptr++) == '-')
        { c2 = *(ptr++);
          if (c2 == '\\') c2 = *(ptr++);
            else if (c2 == ']') r = 0;
          if (*ptr != ']') r = 0;
        } else
        { c2 = '\0';
          r = 0;
        }
        if (r)
        { /* format [a-b] */
          if ((*name < c1) || (*name > c2)) return 0;
          wc = ptr;
        } else
        { /* format [abc] */
          while (*wc != ']')
          { if (*wc == '\\') wc++;
            if (*name == *wc)
            { while ((*wc != ']') && (*(wc-1) != '\\')) wc++;
            } else wc++;
          }
        }
        break;
      default  :
        if (*wc == '\\') if (!(*(++wc))) return -1;
        if (*wc != *name) return 0;
        break;
    }
    name++;
    wc++;
  }
  return (*name == *wc) || !strcmp(wc,"*") ? 1 : 0;
}

int wildcard_matching(char* wc, char* name, int csens)
{ if ((readme_file && !strcmp(name,readme_file)) ||
    !strcmp(name,permission_file)) return 0;
  if (!*wc) return 1;
  if ((*wc == '*') && (*name == '.')) return 0;
  if (csens) return wm_cs((unsigned char*)name,(unsigned char*)wc); else
    return wm_ci((unsigned char*)name,(unsigned char*)wc);
}

void endfind(DIR* d)
{ if (d) closedir(d);
}

struct dirent* findnext(DIR* d, char* wc)
{ struct dirent* f;

  while ((f = readdir(d)) != NULL)
    if (wildcard_matching(wc,f->d_name,1)) return f;
  return f;
}

struct dirent* findfirst(DIR** d, char* dir, char* wc)
{ if ((*d = opendir(dir && *dir ? dir : ".")) == NULL) return NULL;
  return findnext(*d,wc);
}

int eval_options(char *p, int o)
{ while (*p)
  { switch (*p)
    { case 'A' :
      case 'a' : o |= LS_DOT; break;
      case 'c' : o = (o & ~SORT_MASK) | SORT_CTIME; break;
      case 'd' : o |= LS_DIRECTORY; break;
      case 'F' :
      case 'p' : o |= LS_TYPECHAR; break;
      case 'f' : o = (o & ~SORT_MASK) | SORT_NONE ; break;
      case 'l' : o |= LS_LONG; break;
      case 'R' : o |= LS_RECURSIVE; break;
      case 'r' : o |= LS_REVERSE; break;
      case 'S' : o = (o & ~SORT_MASK) | SORT_SIZE; break;
      case 't' : o = (o & ~SORT_MASK) | SORT_MTIME; break;
      case 'u' : o = (o & ~SORT_MASK) | SORT_ATIME; break;
    }
    p++;
  }
  return o;
}

int name_less(char* buffer, char* name, struct stat* st, int options)
{ struct fileinfo_s* fi;
  int                l;

  l = (strlen(buffer) + 8) & ~0x07;
  fi = (struct fileinfo_s*)(buffer+l);
  if (options & LS_REVERSE) switch (options & SORT_MASK)
  { case SORT_NAME  : return (strcmp(buffer,name) >= 0);
    case SORT_MTIME : return fi->time <= st->st_mtime;
    case SORT_ATIME : return fi->time <= st->st_atime;
    case SORT_CTIME : return fi->time <= st->st_ctime;
    case SORT_SIZE  : return fi->size <= st->st_size;
    default         : return 0;
  } else switch (options & SORT_MASK)
  { case SORT_NAME  : return (strcmp(buffer,name) <= 0);
    case SORT_MTIME : return fi->time >= st->st_mtime;
    case SORT_ATIME : return fi->time >= st->st_atime;
    case SORT_CTIME : return fi->time >= st->st_ctime;
    case SORT_SIZE  : return fi->size >= st->st_size;
    default         : return 1;
  }
}

int add_name(char** buffer, int* size, int* used, char* name, struct stat* st,
  int options)
{ struct fileinfo_s* fi;
  int                l   = strlen(name);
  int                ofs = 0;
  int                ofsd;

  if (*used+l+sizeof(struct fileinfo_s)+8 >= *size)
  { *size += 1024;
    *buffer = alloc(*buffer,*size);
    if (!*buffer) return -1;
  }
  while ((ofs < *used) && name_less(*buffer+ofs,name,st,options))
  { ofs += strlen(*buffer+ofs) + sizeof(struct fileinfo_s) + 8;
    ofs &= ~0x07;
  }
  ofsd = (ofs + l + sizeof(struct fileinfo_s) + 8) & ~0x07;
  if (ofs < *used) memmove(*buffer+ofsd,*buffer+ofs,*used-ofs);
  strcpy(*buffer+ofs,name);
  ofs += (l + 8) & ~0x07;
  fi = (struct fileinfo_s*)(*buffer+ofs);
  fi->size = st->st_size;
  switch (options & SORT_MASK)
  { case SORT_ATIME : fi->time = st->st_atime; break;
    case SORT_CTIME : fi->time = st->st_ctime; break;
    default         : fi->time = st->st_mtime; break;
  }
  fi->nlink = st->st_nlink;
  fi->mode = st->st_mode;
  *used += (l + sizeof(struct fileinfo_s) + 8) & ~0x07;
  return 0;
}

void permission_str(int dir, int32 perm, char* str)
{ int i;
  int p;

  if (dir) *(str++) = 'd'; else *(str++) = '-';
  for (i=0; i<3; i++)
  { p = (perm >> ((2-i)*8)) & 0xFF;
    if (dir)
    { if (p & P_LIST) *(str++) = 'r'; else *(str++) = '-';
      if (p & (P_MKDIR+P_WRITE)) *(str++) = 'w'; else *(str++) = '-';
      if (p & P_CHDIR) *(str++) =  'x'; else *(str++) = '-';
    } else
    { if (p & FP_READ) *(str++) = 'r'; else *(str++) = '-';
      if (p & FP_WRITE) *(str++) = 'w'; else *(str++) = '-';
      *(str++) = '-';
    }
  }
  *str = '\0';
}

void uid_gid_str(int32 uid, int32 gid, char* str)
{ char* name;

  name = get_username(uid);
  if (name) sprintf(str," %-9s",name); else sprintf(str," %-9"I32,uid);
  name = get_groupname(gid);
  if (name) sprintf(str+10," %-9s",name); else sprintf(str+10," %-9"I32,gid);
}

void file_info(struct fileinfo_s* st, int32 uid, int32 gid, int32 perm,
  char* str, char* dir, char* name, int year, struct file_s* fp, int n)
{ int32          dgid;
  int32          dlc;
  int32          dperm;
  int32          duid;
  struct file_s* f;
  int            found;
  struct tm*     tm;

  if (st->mode & S_IFDIR)
  { if (read_permission_file(dir,&duid,&dgid,&dperm,&dlc,NULL,NULL) == -1)
    { duid = dgid = -1;
      dlc = dperm = 0;
    }
    permission_str(1,dperm,str);
  } else
  { f = find_name(fp,name,n,&found);
    if (found)
    { duid = f->uid;
      dgid = f->gid;
      dlc = f->dlc;
      permission_str(0,f->perm,str);
    } else
    { duid = uid;
      dgid = gid;
      dlc = 0;
      permission_str(0,dirperm_to_fileperm(perm),str);
    }
  }
  sprintf(str+10," %3"I32,dlc);
  uid_gid_str(duid,dgid,str+strlen(str));
  sprintf(str+strlen(str)," %8li ",st->size);
  tm = localtime(&st->time);
  if (tm->tm_year == year) strftime(str+strlen(str),14,"%b %d %H:%M ",tm);
    else strftime(str+strlen(str),14,"%b %d  %Y ",tm);
}

long send_list_recursive(FILE* ds, char* dir, char* name, char* wc, int options,
  char** gbuffer)
{ long           blocks;
  char*          buffer = *gbuffer;
  DIR*           d;
  struct dirent* de;
  struct file_s* f;
  fd_set         fdset_r;
  fd_set         fdset_w;
  int32          gid;
  long           l;
  int            maxfd;
  int            n;
  int            nfiles;
  int            ofs;
  char*          p[8];
  int32          perm;
  int            size;
  struct stat    st;
  char           str[128];
  struct timeval t;
  time_t         tim;
  int32          uid;
  int            used;
  int            year;

  if (!superuser() && (permission_part_ok(dir,P_LIST,0) == -1))
    return fputs(". unreadable\r\n",ds);
  if (read_permission_file(dir,&uid,&gid,&perm,NULL,(char**)&f,&nfiles) == -1)
  { uid = gid = -1;
    perm = 0;
  }
  if (strcmp(dir,"/"))
  { d = opendir(dir+1);
    *(name++) = '/';
  } else d = opendir(".");
  if (!d) return -1;
  blocks = 0;
  buffer = NULL;
  size = used = 0;
  while ((de = readdir(d)) != NULL)
  { if (((options & LS_DOT) || (*de->d_name != '.')) &&
      wildcard_matching(wc,de->d_name,1))
    { strcpy(name,de->d_name);
      if (!stat(dir+1,&st))
      { if (S_ISDIR(st.st_mode) && strcmp(name,".") && strcmp(name,".."))
          n = directory_hidden(dir); else n = 0;
        if (!n)
        { blocks += st.st_blocks >> 1;
          if (add_name(&buffer,&size,&used,de->d_name,&st,options) == -1)
          { closedir(d);
            errno = ENOMEM;
            return -1;
          }
        }
      }
    }
  }
  closedir(d);
  if ((options & LS_LONG) && (!wc || !*wc))
  { sprintf(str,"total %li\r\n",blocks);
    if (fputs(str,ds) == EOF) return -1;
  }
  if (!buffer) return 0;
  time(&tim);
  year = localtime(&tim)->tm_year;
  FD_ZERO(&fdset_r);
  FD_ZERO(&fdset_w);
  maxfd = sock > fileno(ds) ? sock + 1 : fileno(ds) + 1;
  ofs = 0;
  while (ofs < used)
  { FD_SET(sock,&fdset_r);
    FD_SET(fileno(ds),&fdset_w);
    t.tv_sec  = timeout;
    t.tv_usec = 0;
    if (select(maxfd,&fdset_r,&fdset_w,NULL,&t) <= 0) return -1;
    if (FD_ISSET(sock,&fdset_r))
    { switch (read_ftp_command(str,sizeof(str),p,8,&n))
      { case -1 : return -1;
        case  0 :
          if (strcasecmp(*p,"ABOR"))
          { if (send_reply(503,"Unexpected command.") == -1) return -1;
          } else
          { send_reply(426,"Transfer interrupted on request.");
            return 1;
          }
      }
    } else
    { l = strlen(buffer+ofs) + 1;
      if (options & LS_LONG)
      { strcpy(name,buffer+ofs);
        file_info((struct fileinfo_s*)(buffer+((ofs+l+7) & ~0x07)),uid,gid,
          perm,str,dir,name,year,f,nfiles);
        if (!fputs(str,ds) == EOF) return -1;
      }
      if (fputs(buffer+ofs,ds) == EOF) return -1;
      l = (l + 7) & ~0x07;
      if (options & LS_TYPECHAR)
      { blocks = ((struct fileinfo_s*)(buffer+ofs+l))->mode;
        if (blocks & S_IFDIR) n = fputs("/",ds); else
        if (blocks & S_IFIFO) n = fputs("|",ds); else n = 0;
        if (n == EOF) return -1;
      }
      if (fputs("\r\n",ds) == EOF) return -1;
      ofs += l + sizeof(struct fileinfo_s);
    }
  }
  ofs = 0;
  if (options & LS_RECURSIVE) while (ofs < used)
  { l = (strlen(buffer+ofs) + 8) & ~0x07;
    if (strcmp(buffer+ofs,".") && strcmp(buffer+ofs,"..") &&
      (((struct fileinfo_s*)(buffer+ofs+l))->mode & S_IFDIR))
    { strcpy(name,buffer+ofs);
      n = fprintf(ds,"\r\n%s:\r\n",dir);
      if (n > 0)
      { p[0] = NULL;
        n = send_list_recursive(ds,dir,dir+strlen(dir),wc,options,p);
        if (p[0]) free(p[0]);
      }
      if (n == -1) return -1;
    }
    ofs += l + sizeof(struct fileinfo_s);
  }
  return blocks;
}

int do_list(char* fspec, int options)
{ char* buffer;
  int   ds;
  FILE* dsf;
  char* dir;
  int   n;
  char* name;
  int   perm;
  char  wd[256];

  split_dir(fspec,&dir,&name,!(options & LS_DIRECTORY));
  n = add_dir(cwd,dir,wd,sizeof(wd));
  if ((n == -1) || ((perm = permission_ok(dir,P_LIST)) == -1))
    return send_error(550,"list",errno);
  if (pasvsock == -1)
  { if ((ds = socket(PF_INET,SOCK_STREAM,0)) == -1)
      return send_error(426,"socket",errno);
    if (connect_to(ds,(struct sockaddr*)&addr,sizeof(addr),30) == -1)
    { close(ds);
      return send_error(426,"connect",errno);
    }
  } else
  { n = sizeof(addr);
    ds = accept_to(pasvsock,(struct sockaddr*)&addr,&n,30);
    if (ds == -1) return send_error(426,"accept",errno);
  }
  if ((dsf = fdopen(ds,"wt")) == NULL)
  { n = errno;
    close(ds);
    return send_error(426,"fdopen",n);
  }
  log_action(perm,LG_LS,"LIST","%s",wd);
  if (send_reply(150,"Opening ASCII mode data connection for /bin/ls.") == -1)
  { fclose(dsf);
    return -1;
  }
  buffer = NULL;
  n = send_list_recursive(dsf,wd,wd+strlen(wd),name,options,&buffer);
  n = n < 0 ? errno : 0;
  if (buffer) free(buffer);
  fclose(dsf);
  return n ? send_error(426,"list",n) : send_message(226,1,list_msg.msg,dir);
}

int cmd_list(char** p, int n, char* buffer, int bufferlen)
{ int i       = -1;
  int options = LS_LONG | SORT_NAME;

  if (n > 2) return send_reply(501,"Syntax: LIST [<options>] [<filespec>]");
  if (n >= 1) { if (**p == '-')
  { if (n == 2) i = 1;
    options = eval_options(*p,options);
  } else i = 0; }
  return do_list(i != -1 ? *(p+i) : "",options);
}

int cmd_nlst(char** p, int n, char* buffer, int bufferlen)
{ int i       = -1;
  int options = SORT_NAME;

  if (n > 2) return send_reply(501,"Syntax: NLST [<options>] [<filespec>]");
  if (n >= 1) { if (**p == '-')
  { if (n == 2) i = 1;
    options = eval_options(*p,options);
  } else i = 0; }
  return do_list(i != -1 ? *(p+i) : "",options);
}

/* EOF */
