/*   login.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-1999 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "externs.h"
#include "common.h"
#include "config.h"
#include "ftps.h"
#include "log.h"
#include "perm.h"
#include "rfc931.h"
#include "utils.h"

void get_group_list(void)
{ struct group_s* g;
  int             i;

  printf("GRP_LSR\n");
  if (!scanf("%i",&group_count)) return;
  fgetc(stdin);
  if (!group_count) return;
  if (group_count > group_size)
  { group_size = group_count;
    group = (struct group_s*)alloc(group,group_size*sizeof(struct group_s));
    if (!group)
    { errlog("not enough memory for group list");
      group_size = group_count = 0;
      printf("-\n");
      return;
    }
  }
  printf("+\n");
  g = group;
  for (i=0; i<group_count; i++)
  { if (scanf("%s "SI32,g->name,&g->gid) != 2)
    { errlog("error reading group list");
      group_count = i;
      return;
    }
    fgetc(stdin);
    g++;
  }
}

void get_user_list(void)
{ int            i;
  struct user_s* u;

  printf("USR_LSR\n");
  if (!scanf("%i",&user_count)) return;
  fgetc(stdin);
  if (!user_count) return;
  if (user_count > user_size)
  { user_size = user_count;
    user = (struct user_s*)alloc(user,user_size*sizeof(struct user_s));
    if (!user)
    { errlog("not enough memory for user list");
      user_count = user_size = 0;
      printf("-\n");
      return;
    }
  }
  printf("+\n");
  u = user;
  for (i=0; i<user_count; i++)
  { if (scanf("%s "SI32,u->name,&u->uid) != 2)
    { errlog("error reading user list");
      user_count = i;
      return;
    }
    fgetc(stdin);
    u++;
  }
}

void get_dupecheck(void)
{ char str[512];

  printf("DUP\n");
  readline(str,sizeof(str),stdin);
  if (*str == '+') dupechk = strdup(str+1);
}

void get_shortcut(void)
{ char str[512];

  printf("SSC\n");
  readline(str,sizeof(str),stdin);
  if (*str == '+') shortcut = strdup(str+1);
}

int get_one_top(struct top_s* top)
{ int  i;
  char str[256];

  for (i=0; i<10; i++)
  { fgets(str,sizeof(str),stdin);
    if (sscanf(str,"%s "SI64" %s %f %s %f %s "SI64,top[i].d_name,&top[i].d_amount,
      top[i].e_name,&top[i].e_amount,top[i].l_name,&top[i].l_amount,
      top[i].u_name,&top[i].u_amount) != 8)
    { errlog("error reading top list (%s)",str);
      return -1;
    }
    fgets(top[i].d_info,sizeof(top[i].d_info),stdin);
    fgets(top[i].e_info,sizeof(top[i].e_info),stdin);
    fgets(top[i].l_info,sizeof(top[i].l_info),stdin);
    fgets(top[i].u_info,sizeof(top[i].u_info),stdin);
  }
  return 0;
}

void get_top(void)
{ printf("STO\n");
  if (get_one_top(top_m) < 0) return;
  if (get_one_top(top_t) < 0) return;
  if (get_one_top(top_w) < 0) return;
  if (get_one_top(top_y) < 0) return;
}

int get_message(char* cmd, char** message, int* msg_len, int* msg_size)
{ int n;
  int mlen = 0;
  int msize = 0;

  if (!msg_len) msg_len = &mlen;
  if (!msg_size) msg_size = &msize;
  printf("%s\n",cmd);
  scanf("%i",msg_len);
  fgetc(stdin);
  if (!*msg_len) return 1;
  if (*msg_len >= *msg_size)
  { *msg_size = *msg_len+1;
    *message = alloc(*message,*msg_size+2);
  }
  if (!*message)
  { printf("-\n");
    errlog("out of memory (get_message)");
    return 0;
  }
  printf("+\n");
  n = 0;
  while (n < *msg_len)
  { if (!fgets(*message+n,*msg_size-n+2,stdin))
    { errlog("fgets error (get_message)");
      return 0;
    }
    n += strlen(*message+n);
  }
  *(*message+*msg_len) = 0;
  return 1;
}

int cmd_user(char** p, int n)
{ int  anon_login;
  char name[sizeof(usr.name)+1];
  char str[128];
  int fastlogin = 0;

  if (*usr.name)
    return send_reply(530,"Cannot change user id, log out first.");
  if (n != 1) return send_reply(501,"User name required.");
  if (strlen(*p) >= sizeof(name))
    return send_reply(501,"User name too long.");
  strcpy(name,*p);
  *usr.name = '\0';

  /* get anonymous and ident flags */
  printf("UAN %s\n",name);
  fgets(str,sizeof(str),stdin);
  anon_login = (str[0] == '+');
  anonymous = (str[1] == 'a');
  if (anonymous)
  { if (!anon_login)
    { log_action(0,LG_LOGIN,"LOGIN","by %s FAILED: Guest login disabled",name);
      return send_reply(530,"Guest login disabled.");
    }
    strcpy(str,"Guest login initiated, send E-Mail address as password.");
  } else sprintf(str,"Password required for %s.",name);
  if (send_reply_exp(331,str,0) == -1) return -1;

  /* get password */
  switch (read_ftp_command(str,sizeof(str),p,4,&n))
  { case -1 : return -1;
    case  0 : break;
    default : return 0;
  }
  if (strcasecmp(*p,"PASS"))
    return send_reply(503,"Please login with USER and PASS.");
  if (n > 2) return send_reply(501,"Syntax: PASS <password>");
  if (anonymous) if (n == 2) anonpass = strdup(p[1]);
  if ((n == 1) && (!anonymous)) printf("ULP %s %s\n",usr.ident,name);
     else printf("ULP %s %s %s\n",usr.ident,name,(anonymous) ? ((!anonpass) ? "<none>" : anonpass) : p[1]);
  if (!fgets(str,sizeof(str),stdin))
  { /* probably login kick */
    quit(CL_NORMAL);
    return send_reply(421,"Server has closed connection");
  }
  if (*str == '-')
  { strcpy(name,str+3);
    retries++;
    if (login_retries && (retries >= login_retries))
    { quit(CL_NORMAL);
      log_action(0,LG_LOGIN,"LOGIN","by %s FAILED: Too many login failures",name);
      return send_reply(421,"Too many login failures, connection closed.");
    }
    switch(tolower(str[1]))
    { case '-' : return send_reply(530,"Pipe error, login failed.");
      case 'a' :
        return send_reply(530,"You are already logged in maximum times.");
      case 'c' :
        quit(CL_NORMAL);
        return send_reply(421,"Server is temporary closed.");
      case 'd' :
        quit(CL_NORMAL);
        return send_reply(421,"Your account has been disabled.");
      case 'i' :
        return send_reply(530,"You are not allowed to log in from this "
          "Ident@IP.");
      case 'n' :
        return send_reply(530,"Anonymous login is disabled.");
      case 'p' :
        return send_reply(530,"Login incorrect.");
      case 's' :
        quit(CL_NORMAL);
        return send_reply(421,"Shutdown in progress, no logins allowed.");
      case 't' :
        quit(CL_NORMAL);
        strcpy(usr.name,name);
        prepare_msg(&toomany_msg);
        *usr.name = '\0';
        sleep(5);
        return send_message(421,1,toomany_msg.msg,name);
      case 'u' :
        quit(CL_NORMAL);
        return send_reply(421,"No more connections from your IP/domain.");
      default  :
        return send_message(530,1,"Login failed, reason unknown (%s).",str);
    }
  }
  sscanf(str+1,"%s "SI32" "SI32" "SI64" "SI64" %li %i "SI32" "SI64" "SI32" "
    SI32" "SI32" "SI32" "SI32" "SI32" "SI32" "SI32" "SI32" %li "SI32" "SI32" %i",
    usr.name,&usr.uid,&usr.gid,&usr.downloaded,&usr.uploaded,
    &usr.last_login,&timeout,&usr.limit,&usr.credit,usr.gids,usr.gids+1,
    usr.gids+2,usr.gids+3,usr.gids+4,usr.gids+5,usr.gids+6,usr.gids+7,
    &usr.class,&umask_value,&dl_speed,&ul_speed,&fastlogin);
  prepare_msg(&welcome_msg);
  get_top();
  if (!fastlogin) { if (send_message(230,1,welcome_msg.msg,usr.name) == -1) return -1; }
    else { if (send_message(230,1,"User %s logged in.\r\n",usr.name) == -1) return -1; }
  logged_in = 1;
  prepare_msg(&cdup_msg);
  prepare_msg(&chdir_msg);
  prepare_msg(&dele_msg);
  prepare_msg(&download_msg);
  prepare_msg(&goodbye_msg);
  prepare_msg(&list_msg);
  prepare_msg(&mkdir_msg);
  prepare_msg(&rmdir_msg);
  prepare_msg(&toomany_msg);
  prepare_msg(&upload_msg);
  prepare_msg(&welcome_msg);
  get_group_list();
  get_user_list();
  get_dupecheck();
  get_shortcut();
  load_cdpath();
  load_check();
  load_dcheck();
  load_dshortcut();
  return 0;
}

int cmd_pass(char** p, int n)
{ return send_reply(503,"Login with USER first.");
}

int cmd_acct(char** p, int n)
{ return send_reply(502,"ACCT command not implemented.");
}

int cmd_quit(char** p, int n)
{ quit(CL_NORMAL);
  usr.downloaded += bytes_downloaded;
  usr.uploaded += bytes_uploaded;
  return send_message(221,1,*usr.name ?
    goodbye_msg.msg : goodbye_msg.default_msg, usr.name);
}

/* EOF */
