/*   site.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-1999 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "tweak.h"
#include "externs.h"
#include "common.h"
#include "ftps.h"
#include "list.h"
#include "log.h"
#include "perm.h"
#include "utils.h"

struct file_s* find_name(struct file_s*,char*,int,int*);
struct dirent* findfirst(DIR**,char*,char*);
struct dirent* findnext(DIR*,char*);

int cmd_site_adduser_syntax(int rno)
{ return send_reply(rno,
  "Syntax: ADDUSER <name> <group|gid> <password> <1:ratio|limit|leech> <ip0> [<ip1> [<ip2> [...]]]\n"
  "Adds a new user to the userbase.\n");
}

int cmd_site_adduser(char** p, int n)
{ int  i;
  char str[128];

  printf("USR_ADQ");
  for (i=0; i<n; i++) printf(" %s",p[i]);
  printf("\n");
  if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
  return send_reply(*str == '+' ? 200 : 550,str+1);
}

int cmd_site_bind_syntax(int rno)
{ return send_reply(rno,
    "Syntax: BIND [a|p|s] [<port>]\n"
    "Binds FTP server to another port (a=admin, p=primary, s=secondary).\n"
    "Use port 0 to unbind");
}

int cmd_site_bind(char** p, int n)
{ int  port;
  int  port2;
  int  port3;
  char str[128];

  if ((n && (n != 2)) || ((n == 2) && !sscanf(p[1],"%i",&port)))
    return cmd_site_bind_syntax(501);
  if (n)
  { printf("PRT_SET %s %i\n",p[0],port);
    if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
  } else
  { printf("PRT_GET\n");
    if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error"); else
    { sscanf(str+1,"%i %i %i",&port,&port2,&port3);
      sprintf(str,"+Ports (0=not used): admin=%i, primary=%i, secondary=%i",
        port,port2,port3);
    }
  }
  return send_reply(*str == '+' ? 200 : 550,str+1);
}

int cmd_site_chmod_syntax(int rno)
{ return send_reply(rno,
    "Syntax: CHMOD <who>{+|-|=}<permission> [<directory>|<file>|<wildcard>]\n"
    "<who> is one or more of 'ugo*':\n"
    "d=directory, u=user, g=group, o=others, *=user, group and others\n"
    "{+|-|=}: '+'=add right, '-'=remove right, '='=set right\n"
    "For DIRECTORIES, <permission> is one or more of 'agpoclmn*':\n"
    "a=access (chmod+rmdir), g=get (read) file, p=put (create) file, o=overwrite file,\n"
    "c=change to dir, l=list dir, m=make subdir, n=no limit, h=hidden, *=everything\n"
    "For FILES, <permission> is one or more of 'rwx': r=read, w=write, x=chmod\n"
    "Changes the permissions for the specified file or directory\n"
    "See also SITE CHMODR");
 }

int cmd_site_chmod(char** p, int n)
{ char           cmd;
  DIR*           d;
  struct dirent* de;
  int32          dperm;
  char           dir[4] = "";
  int            file = 0;
  int32          gid;
  long           mode = 0;
  char*          name;
  int            notfile = 0;
  char*          path;
  int32          perm;
  char*          q;
  int            se = 0;
  struct stat    st;
  int32          uid;
  char           wd[256];
  int            who = 0;
  char           wildcard[64];

  if (!flag_get(flags,FLAG_CHMODALLOWED) && !superuser())
    return send_reply(550,"You are not allowed to CHMOD.");
  if (n == 1) p[1] = dir;
  if (n<1 || n>2) se = 1; else
  { q = *p;
    while (*q && *q!='+'&& *q!='-' && *q!='=')
    { if (tolower(*q) == 'd') if (who) se = 1; else who = 0x08; else
      if (tolower(*q) == 'u') if (who & 0x01) se = 1; else who |= 0x01; else
      if (tolower(*q) == 'g') if (who & 0x02) se = 1; else who |= 0x02; else
      if (tolower(*q) == 'o') if (who & 0x04) se = 1; else who |= 0x04; else
      if (tolower(*q) == '*') if (who) se = 1; else who |= 0x07; else se = 1;
      q++;
    }
    if (!who) se = 1;
  }
  if (!se)
  { cmd = *q;
    if (cmd!='+' && cmd!='-' && cmd!='=') se = 1;
  }
  if (!se)
  { q++;
    while (*q) switch (tolower(*(q++)))
    { case 'a' :
        notfile = 1;
        if (mode & P_CHMOD) se = 1; else mode |= P_CHMOD;
        break;
      case 'g' :
        notfile = 1;
        if (mode & P_READ) se = 1; else mode |= P_READ;
        break;
      case 'p' :
        notfile = 1;
        if (mode & P_WRITE) se = 1; else mode |= P_WRITE;
        break;
      case 'o' :
        notfile = 1;
        if (mode & P_OVERWRITE) se = 1; else mode |= P_OVERWRITE;
        break;
      case 'c' :
        notfile = 1;
        if (mode & P_CHDIR) se = 1; else mode |= P_CHDIR;
        break;
      case 'l' :
        notfile = 1;
        if (mode & P_LIST) se = 1; else mode |= P_LIST;
        break;
      case 'm' :
        notfile = 1;
        if (mode & P_MKDIR) se = 1; else mode |= P_MKDIR;
        break;
      case 'n' :
        notfile = 1;
        if (mode & P_NOLIMIT) se = 1; else mode |= P_NOLIMIT;
        break;
      case 'h' :
        notfile = 1;
        if (mode || (who != 0x08)) se = 1; else mode = DP_HIDDEN;
        break;
      case '*' :
        notfile = 1;
        if (mode) se = 1; else mode |= (P_ALL & ~P_NOLIMIT);
        break;
      case 'r' :
        file = 1;
        if (mode & FP_READ) se = 1; else mode |= FP_READ;
        break;
      case 'w' :
        file = 1;
        if (mode & FP_WRITE) se = 1; else mode |= FP_WRITE;
        break;
      case 'x' :
        file = 1;
        if (mode & FP_CHMOD) se = 1; else mode |= FP_CHMOD;
        break;
      default  :
        se = 1;
    }
    if (!mode || (file == notfile)) se = 1;
    if (file && (n != 2)) se = 1;
  }
  if (se) return cmd_site_chmod_syntax(501);
  if (find_cdpath(wd)) return send_reply(550,"Read-only path");
  if (add_dir(cwd,p[1],wd,sizeof(wd)) == -1)
    return send_error(550,"chmod",errno);
  log_action(directory_flag_set(wd,DP_HIDDEN),LG_CHMOD,"CHMOD","%s %s",
    p[0],wd);
  split_dir(wd,&path,&name,0);
  if (*name) strcpy(wildcard,name); else strcpy(wildcard,".");
  if (file)
  { if (permission_ok(path,P_CHDIR) == -1)
    { errno = EACCES;
      return send_error(550,"chmod",errno);
    }
  }
  de = findfirst(&d,path ? path+1 : path,wildcard);
  while (de)
  { if (file)
    { se = read_file_permission(path ? path : "/",de->d_name,&uid,&gid,&perm);
      if ((se != -1) && !superuser() && ((uid != usr.uid) ||
        !((perm >> 16) & FP_CHMOD)))
      { errno = EACCES;
        se = -1;
      }
      if (se != -1)
      { strcpy(name,de->d_name);
        *(name-1) = '/';
        if (stat(wd+1,&st) == -1) st.st_mode = S_IFDIR;
        *(name-1) = '\0';
        if (!S_ISDIR(st.st_mode))
        { printf("#1#FCM %i %c %li %s/%s\n",who,cmd,mode,
            path ? path : "",de->d_name);
          fgets(*p, 8, stdin);
          if (**p == '+') send_message(250,0,"%s: success",de->d_name);
            else send_message(250,0,"%s: failed (server)",de->d_name);
        }
      }
    } else
    { strcpy(name,de->d_name);
      *(name-1) = '/';
      if (stat(wd+1,&st) == -1) st.st_mode = 0;
      if (S_ISDIR(st.st_mode))
      { se = read_permission_file(wd,&uid,&gid,&perm,NULL,NULL,NULL);
        if ((se == -1) && superuser())
        { uid = usr.uid;
          gid = usr.gid;
          perm = umask_value;
          se = 0;
        }
        if ((se >= 0) && !superuser() && ((usr.uid != uid) ||
          !(perm & (P_CHMOD << 16)) || (mode & P_NOLIMIT)))
        { errno = EACCES;
          se = -1;
        }
        if (se >= 0)
        { switch (cmd)
          { case '+' :
            case '-' :
              dperm = 0;
              if (who & 0x01) dperm |= mode<<16;
              if (who & 0x02) dperm |= mode<<8;
              if (who & 0x04) dperm |= mode;
              if (who & 0x08) dperm |= mode<<24;
              if (cmd == '+') perm |= dperm; else perm &= ~dperm;
              break;
            case '=' :
              if (who & 0x01) perm = (perm & 0xFF00FFFF) | (mode<<16);
              if (who & 0x02) perm = (perm & 0xFFFF00FF) | (mode<<8);
              if (who & 0x04) perm = (perm & 0xFFFFFF00) | mode;
              if (who & 0x08) perm = (perm & 0x00FFFFFF) | (mode<<24);
              break;
          }
          if (!superuser()) if (!(perm & (P_CHMOD<<16))) perm |= (P_CHMOD<<16);
          se = write_permission_file(wd,uid,gid,perm);
          if (se >= 0) send_message(250,0,"%s: success",name);
        }
      } else se = 0;
      *(name-1) = '\0';
    }
    if (se == -1)
      send_message(250,0,"%s: failed (permission denied)",de->d_name);
    de = findnext(d,wildcard);
  }
  endfind(d);
  return send_reply(250,"CHMOD command finished.");
}

int site_chmod_rec(char* path, int file, int who, char cmd, long mode)
{ DIR*           d;
  struct dirent* de;
  int32          dperm;
  int32          gid;
  char*          name;
  int32          perm;
  int            se;
  struct stat    st;
  char           temp[8];
  int32          uid;

  de = findfirst(&d,strcmp(path,"/") ? path+1 : ".","*");
  if (strcmp(path,"/")) strcat(path,"/");
  name = path + strlen(path);
  while (de)
  { se = 0;
    if (file)
    { *name = '\0';
      se = read_file_permission(path,de->d_name,&uid,&gid,&perm);
      if ((se != -1) && !superuser() && ((uid != usr.uid) ||
        !((perm >> 16) & FP_CHMOD)))
      { errno = EACCES;
        se = -1;
      }
      if (se != -1)
      { strcpy(name,de->d_name);
        if (stat(path+1,&st) == -1) st.st_mode = S_IFDIR;
        if (!S_ISDIR(st.st_mode))
        { printf("#1#FCM %i %c %li %s\n",who,cmd,mode,path);
          fgets(temp, sizeof(temp), stdin);
          if (*temp == '+') send_message(250,0,"%s: success",path);
            else send_message(250,0,"%s: failed (server)",path);
        }
      }
    } else if (strcmp(de->d_name,".."))
    { strcpy(name,de->d_name);
      if (stat(path+1,&st) == -1) st.st_mode = 0;
      if (S_ISDIR(st.st_mode))
      { se = read_permission_file(path,&uid,&gid,&perm,NULL,NULL,NULL);
        if ((se == -1) && superuser())
        { uid = usr.uid;
          gid = usr.gid;
          perm = umask_value;
          se = 0;
        }
        if ((se >= 0) && !superuser() && ((usr.uid != uid) ||
          !(perm & (P_CHMOD << 16)) || (mode & P_NOLIMIT)))
        { errno = EACCES;
          se = -1;
        }
        if (se >= 0)
        { switch (cmd)
          { case '+' :
            case '-' :
              dperm = 0;
              if (who & 0x01) dperm |= mode<<16;
              if (who & 0x02) dperm |= mode<<8;
              if (who & 0x04) dperm |= mode;
              if (who & 0x08) dperm |= mode<<24;
              if (cmd == '+') perm |= dperm; else perm &= ~dperm;
              break;
            case '=' :
              if (who & 0x01) perm = (perm & 0xFF00FFFF) | (mode<<16);
              if (who & 0x02) perm = (perm & 0xFFFF00FF) | (mode<<8);
              if (who & 0x04) perm = (perm & 0xFFFFFF00) | mode;
              if (who & 0x08) perm = (perm & 0x00FFFFFF) | (mode<<24);
              break;
          }
          if (!superuser()) if (!(perm & (P_CHMOD<<16))) perm |= (P_CHMOD<<16);
          se = write_permission_file(path,uid,gid,perm);
          if (se >= 0) send_message(250,0,"%s: success",path);
        }
      } else se = 0;
    }
    if (se == -1)
      send_message(250,0,"%s: failed (permission denied)",path);
    else if (S_ISDIR(st.st_mode) && strcmp(de->d_name,".") &&
      strcmp(de->d_name,"..")) site_chmod_rec(path,file,who,cmd,mode);
    de = findnext(d,"*");
  }
  endfind(d);
  return 0;
}

int cmd_site_chmodr_syntax(int rno)
{ return send_reply(rno,
    "Syntax: CHMODR <who>{+|-|=}<permission> [<directory>]\n"
    "<who> is one or more of 'ugo*':\n"
    "d=directory, u=user, g=group, o=others, *=user, group and others\n"
    "{+|-|=}: '+'=add right, '-'=remove right, '='=set right\n"
    "For DIRECTORIES, <permission> is one or more of 'agpoclmnh*':\n"
    "a=access (chmod+rmdir), g=get (read) file, p=put (create) file, o=overwrite file,\n"
    "c=change to dir, l=list dir, m=make subdir, n=no limit, h=hidden, *=everything\n"
    "For FILES, <permission> is one or more of 'rwx': r=read, w=write, x=chmod\n"
    "Recursively changes the permissions for the specified file or directory\n"
    "See also SITE CHMOD");
}


int cmd_site_chmodr(char** p, int n)
{ char           cmd;
  char           dir[4] = "";
  int            file = 0;
  long           mode = 0;
  int            notfile = 0;
  char*          q;
  int            se = 0;
  char           wd[256];
  int            who = 0;

  if (!flag_get(flags,FLAG_CHMODALLOWED) && !superuser())
    return send_reply(550,"You are not allowed to CHMOD.");
  if (n == 1) p[1] = dir;
  if (n<1 || n>2) se = 1; else
  { q = *p;
    while (*q && *q!='+'&& *q!='-' && *q!='=')
    { if (tolower(*q) == 'u') if (who & 0x01) se = 1; else who |= 0x01; else
      if (tolower(*q) == 'g') if (who & 0x02) se = 1; else who |= 0x02; else
      if (tolower(*q) == 'o') if (who & 0x04) se = 1; else who |= 0x04; else
      if (tolower(*q) == 'd') if (who) se = 1; else who = 0x08; else
      if (tolower(*q) == '*') if (who) se = 1; else who |= 0x07; else se = 1;
      q++;
    }
    if (!who) se = 1;
  }
  if (!se)
  { cmd = *q;
    if (cmd!='+' && cmd!='-' && cmd!='=') se = 1;
  }
  if (!se)
  { q++;
    while (*q) switch (tolower(*(q++)))
    { case 'a' :
        notfile = 1;
        if (mode & P_CHMOD) se = 1; else mode |= P_CHMOD;
        break;
      case 'g' :
        notfile = 1;
        if (mode & P_READ) se = 1; else mode |= P_READ;
        break;
      case 'p' :
        notfile = 1;
        if (mode & P_WRITE) se = 1; else mode |= P_WRITE;
        break;
      case 'o' :
        notfile = 1;
        if (mode & P_OVERWRITE) se = 1; else mode |= P_OVERWRITE;
        break;
      case 'c' :
        notfile = 1;
        if (mode & P_CHDIR) se = 1; else mode |= P_CHDIR;
        break;
      case 'l' :
        notfile = 1;
        if (mode & P_LIST) se = 1; else mode |= P_LIST;
        break;
      case 'm' :
        notfile = 1;
        if (mode & P_MKDIR) se = 1; else mode |= P_MKDIR;
        break;
      case 'n' :
        notfile = 1;
        if (mode & P_NOLIMIT) se = 1; else mode |= P_NOLIMIT;
        break;
      case 'h' :
        notfile = 1;
        if (mode || (who != 0x08)) se = 1; else mode = DP_HIDDEN;
        break;
      case '*' :
        notfile = 1;
        if (mode) se = 1; else mode |= (P_ALL & ~P_NOLIMIT);
        break;
      case 'r' :
        file = 1;
        if (mode & FP_READ) se = 1; else mode |= FP_READ;
        break;
      case 'w' :
        file = 1;
        if (mode & FP_WRITE) se = 1; else mode |= FP_WRITE;
        break;
      case 'x' :
        file = 1;
        if (mode & FP_CHMOD) se = 1; else mode |= FP_CHMOD;
        break;
      default  :
        se = 1;
    }
    if (!mode || (file == notfile)) se = 1;
  }
  if (se) return cmd_site_chmod_syntax(501);
  if (add_dir(cwd,*(p+1),wd,sizeof(wd)) == -1)
    return send_error(550,"chmod",errno);
  if (find_cdpath(wd)) return send_reply(550,"Read-only path");
  log_action(directory_flag_set(wd,DP_HIDDEN),LG_CHMOD,"CHMODR","%s %s",
    p[0],wd);
  site_chmod_rec(wd,file,who,cmd,mode);
  return send_reply(250,"CHMOD command finished.");
}

int cmd_site_chown_syntax(int rno)
{ return send_reply(rno,
    "Syntax: CHOWN [<user>]:[<group>] [<directory>]\n"
    "<user> can be user id or user name, <group> can be group id or group name\n"
    "at least one of <user> and <group> must be given.");
}

int cmd_site_chown(char** p, int n)
{ int32 gid;
  int32 new_gid;
  int32 new_uid;
  int32 perm;
  char* q;
  int   se = 0;
  int32 uid;
  char  wd[256];

  if (!superuser()) return send_reply(550,"Only superuser can do CHOWN.");
  if (n<1 || n>2) se = 1; else
  { q = strchr(*p,':');
    if (q) *q = '\0';
    if (**p)
    { if ((new_uid = get_uid(*p)) == -1)
        return send_reply(501,"User not found.");
    } else new_uid = -1;
    if (q && *(q+1))
    { if ((new_gid = get_gid(q+1)) == -1)
        return send_reply(501,"Group not found.");
    } else new_gid = -1;
    if (new_uid == new_gid == -1) se = 1;
  }
  if (se) return cmd_site_chown_syntax(501);
  if (find_cdpath(n ==2 ? *(p+1) : ""))
    return send_reply(550,"Read-only path");
  se = read_permission_file(n == 2 ? *(p+1) : "",&uid,&gid,&perm,
    NULL,NULL,NULL);
  if ((se == -1) && (errno == ENOTDIR))
  { if (n != 2) return send_error(550,"chown",errno);
    add_dir(cwd,*(p+1),wd,sizeof(wd));
    printf("FCO- "SI32" "SI32" %s\n",new_uid,new_gid,wd);
    fgets(wd, sizeof(wd), stdin);
    return send_reply(*wd == '+' ? 250 : 550,wd+1);
  }
  if (se == -1)
  { uid = gid = 0;
    perm = umask_value;
  }
  if (new_uid != -1) uid = new_uid;
  if (new_gid != -1) gid = new_gid;
  if (n == 2) se = write_permission_file(*(p+1),uid,gid,perm);
    else se = write_permission_file("",uid,gid,perm);
  if (se == -1) return send_error(550,"chown",errno); else
    return send_reply(250,"CHOWN command successful.");
}

int cmd_site_close_syntax(int rno)
{ return send_reply(rno,"Syntax: CLOSE [{normal|transactional|immediate}]\n"
    "Forces all existing connections to close and refuses\n"
    "login attempts until OPEN command is given.");
}

int cmd_site_close(char** p, int n)
{ char str[128];

  if (n > 1) return cmd_site_close_syntax(501);
  if (n)
  { if (!strcasecmp(p[0],"normal")) sprintf(str,"cn"); else
      if (!strcasecmp(p[0],"transactional")) sprintf(str,"ct"); else
      if (!strcasecmp(p[0],"immediate")) sprintf(str,"ci"); else
      return cmd_site_close_syntax(501);
  } else sprintf(str,"ci");
  printf("SRV_SST %s\n",str);
  if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
  return send_reply(*str == '+' ? 200 : 550,str+1);
}

/************************/
/* SITE COLOR extension */
/************************/

int cmd_site_color_syntax(int rno)
{ return send_reply(rno,"Syntax: COLOR\n"
    "Switches between Color- and Monochrome-Mode");
}

int cmd_site_color(char** p, int n)
{ char str[128];

  if (n) return cmd_site_color_syntax(501);
  /* change it in the current session */
  usr.class ^= UF_COLOR;
  /* change it in the users data */
  printf("COL\n");
  if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
  return send_reply(*str == '+' ? 200 : 500,str+1);
}

int cmd_site_count_syntax(int rno)
{ return send_reply(rno,
    "Syntax: COUNT\n"
    "Returns the number of connected users");
}

int cmd_site_count(char** p, int n)
{ char str[128];

  printf("CNT\n");
  if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
  return send_reply(*str == '+' ? 200 : 550,str+1);
}

int cmd_site_version_syntax(int rno)
{ return send_reply(rno,
    "Syntax: VERSION\n"
    "Returns the current FTP4ALL-Version of the server");
}

int cmd_site_version(char** p, int n)
{ char str[128];
  printf("VER\n");
  if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
  return send_reply(*str == '+' ? 200 : 550,str+1);
}

long del_files_r(char* path)
{ struct dirent* d;
  DIR*           dir;
  char*          name;
  long           size = 0;
  struct stat    st;

  dir = opendir(path);
  if (!dir) return 0;
  strcat(path,"/");
  name = path + strlen(path);
  while ((d = readdir(dir)) != NULL)
  { strcpy(name,d->d_name);
    if (strcmp(d->d_name,".") && strcmp(d->d_name,"..") && !stat(path,&st))
    { if (S_ISDIR(st.st_mode))
      { size += del_files_r(path);
        strcpy(name,d->d_name);
        rmdir(path);
      } else
      { unlink(path);
        if (*d->d_name != '.') size += st.st_size;
      }
    }
  }
  closedir(dir);
  return size;
}

int cmd_site_del_syntax(int rno)
{ return send_reply(rno,
    "Syntax: DEL {<username>|<uid>} {<filename>|<directory>} [<factor>]\n"
    "Deletes file or all files in the directory and reduces the upload\n"
    "amount of the specified user by the size of the deleted file(s)\n"
    "multiplied with the <factor>.");
}

int cmd_site_del(char** p, int n)
{ float          factor = 1.0;
  int            ok;
  unsigned long  size = 0;
  struct stat    st;
  char           wd[256];

  if (!superuser()) return send_reply(550,"Only superuser can do SITE DEL.");
  if ((n < 2) || (n > 3)) return cmd_site_del_syntax(501);
  if (n == 3) if (!sscanf(*(p+2),"%f",&factor))
    return cmd_site_del_syntax(501);
  if (get_uid(p[0]) == -1) return send_reply(550,"User does not exist.");
  ok = add_dir(cwd,p[1],wd,sizeof(wd));
  if (ok != -1) ok = stat(wd+1,&st);
  if (ok != -1)
  { if (S_ISDIR(st.st_mode))
    { ok = remove_permission_file(wd);
      if (ok != -1) size = del_files_r(wd+1);
      if (ok != -1) ok = add_dir(cwd,p[1],wd,sizeof(wd));
      if (ok != -1) rmdir(wd+1);
      log_action(DP_HIDDEN,LG_DEL,"SITE DEL","%s %s (total %lu bytes * %3.1f)",
        p[0],wd,size,factor);
    } else
    { size = st.st_size;
      ok = unlink(wd+1);
      log_action(DP_HIDDEN,LG_DEL,"SITE DEL","%s %s (%lu bytes * %3.1f)",
        p[0],wd,size,factor);
    }
  }
  if (ok == -1) log_action(DP_HIDDEN,LG_DEL,"SITE DEL","%s %s FAILED: %s",
    p[0],wd,strerror(errno));
  if (ok == -1) return send_error(550,"delete",errno);
  factor *= size;
  printf("UUR %s %1.0f\n",*p,factor);
  if (!fgets(wd+128,sizeof(wd)-128,stdin)) strcpy(wd+128,"-pipe error");
  if (*(wd+128) == '-') return send_reply(550,wd+129);
  return send_reply(200,wd);
}

int cmd_site_disable_syntax(int rno)
{ return send_reply(rno,"Syntax: DISABLE <username>\n"
    "Disables user account.");
}

int cmd_site_disable(char** p, int n)
{ char str[128];

  if (n != 1) return cmd_site_disable_syntax(501);
  printf("ADI %s\n",*p);
  if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
  return send_reply(*str == '+' ? 200 : 550,str+1);
}

int cmd_site_dump_syntax(int rno)
{ return send_reply(rno,"Syntax: DUMP <area>\n"
    "Dumps ftps data. <area> may be cdpath|check|config|dcheck|dshortcut.");
}

int cmd_site_dump(char** p, int n)
{ int   i;
  int   l;
  char* ptr;
  char  str[256];

  if (n != 1) return cmd_site_dump_syntax(501);
  if (!superuser()) return send_reply(550,"Permission denied");
  if (!strcasecmp(*p,"cdpath"))
  { sprintf(str,"cdpath=0x%08X, len=%i, size=%i",
      (int)cdpath,cdpath_len,cdpath_size);
    send_message(200,cdpath_len ? 0 : 1,str,NULL);
    i = 0;
    while (i < cdpath_len)
    { sprintf(str,"%-40s",cdpath+i);
      l = strlen(cdpath+i);
      i += (l + 0x08) & ~0x07;
      sprintf(str+strlen(str),": uid="SI32", gid="SI32", perm=0x%06X",
        *(int32*)(cdpath+i),
        *(int32*)(cdpath+i+sizeof(int32)),
        *(int32*)(cdpath+i+(2*sizeof(int32))));
      i += 3 * sizeof(int32);
      send_message(200,i < cdpath_len ? 0 : 1,str,NULL);
    }
  } else if (!strcasecmp(*p,"check"))
  { sprintf(str,"check=0x%08X, count=%i, size=%i",
      (int)check,check_count,check_size);
    send_message(200,check_count ? 0 : 1,str,NULL);
    for (i=0; i<check_count; i++)
    { sprintf(str,"%-8s %-48s %-8s",
        check[i].wildcard,check[i].name,check[i].param);
      send_message(200,i < check_count-1 ? 0 : 1,str,NULL);
    }
  } else if (!strcasecmp(*p,"config"))
  { sprintf(str,"server_dir=%s",server_dir);
    send_message(200,0,str,NULL);
    sprintf(str,"permission_file=%s",permission_file);
    send_message(200,0,str,NULL);
    sprintf(str,"readme_file=%s",readme_file);
    send_message(200,0,str,NULL);
    sprintf(str,"shortcut=%s",shortcut);
    send_message(200,0,str,NULL);
    sprintf(str,"dupechk=%s",dupechk);
    send_message(200,0,str,NULL);
    sprintf(str,"dupe_path=%s",dupe_path);
    send_message(200,1,str,NULL);
  } else if (!strcasecmp(*p,"dcheck"))
  { sprintf(str,"dcheck=0x%08X, count=%i, size=%i",
      (int)dcheck,dcheck_count,dcheck_size);
    send_message(200,dcheck_count ? 0 : 1,str,NULL);
    for (i=0; i<dcheck_count; i++)
    { sprintf(str,"%-8s %-48s %-8s",
        dcheck[i].wildcard,dcheck[i].name,dcheck[i].param);
      send_message(200,i < dcheck_count-1 ? 0 : 1,str,NULL);
    }
  } else if (!strcasecmp(*p,"dshortcut"))
  { sprintf(str,"dshortcut=0x%08X, len=%i, size=%i",
      (int)dshortcut,dshortcut_len,dshortcut_size);
    send_message(200,dshortcut_len ? 0 : 1,str,NULL);
    for (i=0; i<dshortcut_len;)
    { ptr = dshortcut + i;
      l = strlen(ptr) + 1;
      sprintf(str,"%-20s %-s",ptr,ptr+l);
      i += l + strlen(ptr+l) + 1;
      send_message(200,i < dshortcut_len ? 0 : 1,str,NULL);
    }
  } else return cmd_site_dump_syntax(501);
  return 0;
}

int cmd_site_enable_syntax(int rno)
{ return send_reply(rno,"Syntax: ENABLE <username>\n"
    "Enables user account.");
}

int cmd_site_enable(char** p, int n)
{ char str[128];

  if (n != 1) return cmd_site_enable_syntax(501);
  printf("AEN %s\n",*p);
  if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
  return send_reply(*str == '+' ? 200 : 550,str+1);
}

int cmd_site_exec_syntax(int rno)
{ return send_reply(rno,"Syntax: EXEC <command> [<parameters>]\n"
    "Executes a shell command.");
}

int cmd_site_exec(char** p, int n)
{ char* c;
  int   err;
  FILE* file;
  pid_t pid;
  int   pipefd[2];
  char  str[128];

  if (!n) return cmd_site_exec_syntax(501);
  if (!superuser()) return send_reply(550,"You are not superuser.");
  err = socketpair(AF_UNIX,SOCK_STREAM,0,pipefd);
  if (err == -1) return send_error(550,"pipe",0);
  pid = fork();
  if (pid == -1) return send_error(550,"fork",errno);
  if (pid)
  { close(pipefd[1]);
    file = fdopen(pipefd[0],"rt"); 
    if (!file) return send_reply(550,"fdopen");
    do
    { c = fgets(str,sizeof(str)-1,file);
      if (c && strlen(str))
        if (send_message(200,0,str,NULL) == -1) return -1;
    } while (c);
    fclose(file);
    close(pipefd[0]);
    pid = wait(&err);
    sprintf(str,"EXEC finished with exitcode %i.",err);
    return send_reply(200,str);
  } else
  { fclose(sockf);
    if (strcmp(cwd,"/")) chdir(cwd+1);
    close(pipefd[0]);
    err = dup2(pipefd[1],1);
    if (err != -1) err = dup2(pipefd[1],2);
    if (err != -1) err = dup2(pipefd[1],0);
    close(pipefd[1]);
    *(p+n) = NULL;
    if (err != -1) err = execvp(*p,p);
    exit(err);
  }
}

int cmd_site_group_syntax(int rno)
{ return send_reply(rno,
    "Implemented GROUP commands, syntax: SITE GROUP <command> <arguments>\n"
    "ADD    <name> [<gid>]                                 - add new group\n"
    "CHANGE <name> {NAME=<s>|GID=<n>|LIMIT=<n>|CREDIT=<n>|\n"
    "               DL_MAX=<n>|DL_MIN=<n>|UL_MAX=<n>|\n"
    "               UL_MIN=<n>|LOGINS=<n>|INFO=<s>|\n"
    "               IDENT=<s>|UMASK=<n>}                   - change group\n"
    "LIST                                                  - list all groups\n"
    "REMOVE [<name>]                                       - remove group");
}

int cmd_site_group(char** p, int n)
{ int64 credit;
  int32 dl_min;
  int32 dl_max;
  int   i;
  char  in[256];
  char* q;
  char* q2;
  int   se = 0;
  char  str[256];
  int32 ul_min;
  int32 ul_max;
  int32 umask;

  if (!n) se = 1; else if (!strcasecmp(*p,"ADD"))
  { if ((n < 2) || (n > 3)) se = 1; else
      printf("GRP_ADD %s %s\n",p[1],n == 3 ? p[2] : "-0");
  } else if (!strcasecmp(*p,"CHANGE"))
  { if (n < 3) se = 1; else
    { printf("GRP_CHG %s",p[1]);
      for (i=2; i<n; i++) printf(" %s",p[i]);
      printf("\n");
    }
  } else if (!strcasecmp(*p,"LIST"))
  { if (n != 1) se = 1; else
    { if (send_message(211,0,
        "NAME      GID LIM MBCRED DMIN DMAX UMIN UMAX   UMASK  LG INFO             IDENT",NULL) == -1) return -1; 
      printf("GRP_LS\n");
      if (!fgets(in,sizeof(in),stdin)) strcpy(in,"-pipe error");
      while ((*in == '+') && strcmp(in,"+\n"))
      { i = 0;
        q = in + 1;
        do
        { p[i++] = q;
          q2 = strchr(q,'\t');
          if (q2) *(q2++) = '\0';
          q = q2;
        } while (q2);
        if (!sscanf(p[3],SI64,&credit)) credit = 0;
        credit /= 1024 * 1024;
        if (!sscanf(p[4],SI32,&dl_min)) dl_min = 0;
        dl_max = (dl_min >> 16) & 0xFFFF;
        dl_min &= 0xFFFF;
        if (!sscanf(p[5],SI32,&ul_min)) ul_min = 0;
        ul_max = (ul_min >> 16) & 0xFFFF;
        ul_min &= 0xFFFF;
	if (!sscanf(p[6],SI32,&umask)) umask=0;
        sprintf(str,"%-9s %3s %3s %6"I64" %4"I32" %4"I32" %4"I32" %4"I32" 0x%06X %2s %-16s %s",
	p[0],p[1],p[2],credit,dl_min,dl_max,ul_min,ul_max,umask, p[7],p[8],p[9]);
        if (send_message(211,0,str,NULL) == -1) return -1;
        if (!fgets(in,sizeof(in),stdin)) strcpy(in,"-");
      }
      if (*in == '+') return send_reply(211,"<end of list>");
      if (strcmp(in,"-")) return send_reply(211,in+1);
        else return send_reply(211,"pipe error, aborting list");
    }
  } else if (!strcasecmp(*p,"REMOVE"))
  { if (n != 2) se = 1; else printf("GRP_RM %s\n",p[1]);
  } else se = 1;
  if (se) return cmd_site_group_syntax(501);
  if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
  return send_reply(*str == '+' ? 200 : 550,str+1);
}

int cmd_site_groups_syntax(int rno)
{ return send_reply(rno,
    "Syntax: GROUPS [<user> [{+|-}<group>]]\n"
    "Without parameters, display the names of the groups you are in. The\n"
    "superuser can also add/remove users to/from groups.");
}

int cmd_site_groups(char** p, int n)
{ int   i;
  char* name;
  char  str[128];

  if (!n)
  { strcpy(str,"Primary group: ");
    name = get_groupname(usr.gid);
    if (name) strcat(str,name); else sprintf(str+strlen(str),SI32,usr.gid);
    if (send_message(200,0,str,NULL) == -1) return -1;
    strcpy(str,"Secondary groups:");
    for (i=0; i<8; i++) if (usr.gids[i] != -1)
    { strcat(str," ");
      name = get_groupname(usr.gids[i]);
      if (name) strcat(str,name); else
        sprintf(str+strlen(str),SI32,usr.gids[i]);
    }
    if (strlen(str) == 17) strcat(str," (none)");
    return send_reply(200,str);
  }
  if (n == 1)
  { printf("GRP %s\n",*p);
    fgets(str, sizeof(str), stdin);
    if (*str == '-') return send_reply(550,str+1);
    if (send_message(200,0,str+1,NULL) == -1) return -1;
    fgets(str, sizeof(str), stdin);
    return send_reply(200,str+1);
  }
  if (n == 2)
  { printf("GRP %s %s\n",*p,*(p+1));
    fgets(str, sizeof(str), stdin);
    if (*str == '-') return send_reply(550,str+1);
    return send_reply(200,str+1);
  }
  return cmd_site_groups_syntax(501);
}

int cmd_site_info_syntax(int rno)
{ return send_reply(rno,"Syntax: INFO <user_information>\n"
    "<user_information> = current user's new user information\n"
    "This command sets the user information of the current user. The user\n"
    "information is displayed in top ten lists. The length is limited.");
}

int cmd_site_idle_syntax(int rno)
{ return send_reply(rno,"Syntax: IDLE <timeout>\nChanges your maximum idletime.");
}

int cmd_site_idle(char **p, int n)
{ return send_reply(550,"Currently not implemented.");
}

int cmd_site_info(char** p, int n)
{ char str[128];

  if (n > 1) return cmd_site_info_syntax(501);
  if (!n) printf("UIN\n"); else printf("UIN %s\n",*p);
  if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
  return send_reply(*str == '+' ? 200 : 550,str+1);
}

int cmd_site_kick_syntax(int rno)
{ return send_reply(rno,"Syntax: KICK {<user>|<pid>} [<signal_no>]\n"
    "<user> = name of logged in user or <pid> = process id\n"
    "<signal_no> = number of signal to send\n"
    "This sends a signal to the user's process, thereby kicking the user from\n"
    "the server. If you give a process id, only this session is killed.\n"
    "Otherwise all sessions of that user are killed. The only signal no that\n"
    "makes sense is 9 and should be used if the normal kick does not work\n"
    "(This should never happen).");
}

int cmd_site_kick(char** p, int n)
{ int  pid;
  char str[128];

  if ((n < 1) || (n > 2)) return cmd_site_kick_syntax(501);
  if (sscanf(p[0],"%i",&pid)) printf("USR_KIL %i %s\n",pid,n == 2 ? p[1] : "");
    else printf("USR_KIC %s %s\n",p[0],n == 2 ? p[1] : "");
  if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
  return send_reply(*str == '+' ? 200 : 550,str+1);
}

int cmd_site_mod_syntax(int rno)
{ return send_reply(rno,"Syntax: MOD [<directory>]\n"
    "Shows the permissions of a directory.");
}

void cat_permissions(char* s, long p)
{ if (p & P_NOLIMIT) strcat(s," (no limit)");
  if ((p & P_ALL) == P_ALL)
  { strcat(s," (full rights)");
    return;
  }
  if (!p)
  { strcat(s," (no rights)");
    return;
  }
  if (p & P_CHDIR) strcat(s," cd");
  if (p & P_LIST) strcat(s," ls");
  if (p & P_MKDIR) strcat(s," mkdir");
  if (p & P_CHMOD) strcat(s," chmod");
  if (p & P_READ) strcat(s," get");
  if (p & P_WRITE) strcat(s," put");
  if (p & P_OVERWRITE) strcat(s," overwrite");
}

void cat_file_permissions(char* s, long p)
{ if (!p)
  { strcat(s," (no rights)");
    return;
  }
  if (p & FP_READ) strcat(s," read");
  if (p & FP_WRITE) strcat(s," write");
  if (p & FP_CHMOD) strcat(s," chmod");
}

int cmd_site_mod(char** p, int n)
{ char*          dir;
  int32          gid;
  char*          name;
  int32          perm;
  int            se;
  char           str[64];
  int32          uid;

  if (n > 1) return cmd_site_mod_syntax(501);
  if (n == 1)
  { se = permission_ok(*p,P_CHDIR);
    if ((se == -1) && (errno == ENOTDIR))
    { split_dir(*p,&dir,&name,0);
      se = permission_ok(dir,P_CHDIR);
      if (se != -1)
      { se = read_file_permission(dir,name,&uid,&gid,&perm);
        if (se != -1)
        { strcpy(str,"owner [");
          name = get_username(uid);
          if (name) sprintf(str+strlen(str),"%-9s",name); else
            sprintf(str+strlen(str),"%-9"I32,uid);
          strcat(str,"] :");
          cat_file_permissions(str,(perm >> 16) & 0xFF);
          if (send_message(250,0,str,NULL) == -1) return -1;
          strcpy(str,"group [");
          name = get_groupname(gid);
          if (name) sprintf(str+strlen(str),"%-9s",name); else
            sprintf(str+strlen(str),"%-9"I32,gid);
          strcat(str,"] :");
          cat_file_permissions(str,(perm >> 8) & 0xFF);
          if (send_message(250,0,str,NULL) == -1) return -1;
          strcpy(str,"others            :");
          cat_file_permissions(str,perm & 0xFF);
          return send_reply(250,str);
        } else errno = EACCES;
      }
    }
    if (se >= 0) se = read_permission_file(*p,&uid,&gid,&perm,NULL,NULL,NULL);
  } else
  { se = permission_ok("",P_CHDIR);
    if (se >= 0) se = read_permission_file("",&uid,&gid,&perm,NULL,NULL,NULL);
  }
  if (se >= 0)
  { strcpy(str,"directory         :");
    if ((perm >> 24) & DP_HIDDEN) strcat(str," hidden");
    if (!((perm >> 24) & 0xFF)) strcat(str," (no flags)");
    if (send_message(250,0,str,NULL) == -1) return -1;
    strcpy(str,"owner [");
    name = get_username(uid);
    if (name) sprintf(str+strlen(str),"%-9s",name); else
      sprintf(str+strlen(str),"%-9"I32,uid);
    strcat(str,"] :");
    cat_permissions(str,(perm >> 16) & 0xFF);
    if (send_message(250,0,str,NULL) == -1) return -1;
    strcpy(str,"group [");
    name = get_groupname(gid);
    if (name) sprintf(str+strlen(str),"%-9s",name); else
      sprintf(str+strlen(str),"%-9"I32,gid);
    strcat(str,"] :");
    cat_permissions(str,(perm >> 8) & 0xFF);
    if (send_message(250,0,str,NULL) == -1) return -1;
    strcpy(str,"others            :");
    cat_permissions(str,perm & 0xFF);
    return send_reply(250,str);
  }
  return send_error(550,"mod",errno);
}

int cmd_site_open_syntax(int rno)
{ return send_reply(rno,"Syntax: OPEN\n"
    "Non-sysop logins will be permitted again.");
}

int cmd_site_open(char** p, int n)
{ char str[128];

  if (n) return cmd_site_open_syntax(501);
  fputs("SRV_SST on\n",stdout);
  if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
  return send_reply(*str == '+' ? 200 : 550,str+1);
}

int cmd_site_passwd_syntax(int rno)
{ return send_reply(rno,"Syntax: PASSWD [<user>] <password>\n"
    "Changes password for you or for specified <user>. You must be superuser\n"
    "in order to change someone others password. Password - means to clear it.");
}

int cmd_site_passwd(char ** p, int n)
{ char str[128];

  if ((n < 1) || (n > 2)) return cmd_site_passwd_syntax(501);
  printf("USR_PWD %s %s\n",n == 1 ? usr.name : p[0],
    strcmp(p[n-1],"-") ? p[n-1] : "");
  if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
  return send_reply(*str == '+' ? 200 : 550,str+1);
}

int cmd_site_reg_syntax(int rno)
{ return send_reply(rno,
    "Implemented REG commands, syntax: SITE REG <command> <arguments>\n"
    "DEL  <key>           - delete entry\n"
    "GET  <key>           - get entry\n"
    "LIST                 - list all entries\n"
    "SET  <key> <value>   - set entry");
}

int cmd_site_reg(char** p, int n)
{ int   i;
  char  in[256];
  int   se = 0;
  char  str[256];

  if (!strcasecmp(p[0],"DEL") || !strcasecmp(p[0],"RM"))
  { if (n != 2) se = 1; else printf("REG_RM %s\n",p[1]);
  } else if (!strcasecmp(p[0],"GET"))
  { if (n != 2) se = 1; else printf("REG_GET %s\n",p[1]);
  } else if (!strcasecmp(p[0],"LIST"))
  { if (n != 1) se = 1; else
    { printf("REG_LS\n");
      if (!fgets(in,sizeof(in),stdin)) strcpy(in,"-pipe error");
      while ((*in == '+') && strcmp(in,"+\n"))
      { if (send_message_exp(211,0,in+1,NULL,0) == -1) return -1;
        if (!fgets(in,sizeof(in),stdin)) strcpy(in,"-");
      }
      if (*in == '+') return send_reply(211,"<end of list>");
      if (strcmp(in,"-")) return send_reply_exp(550,in+1,0);
        else return send_reply(211,"pipe error, aborting list");
    }
  } else if (!strcasecmp(*p,"SET"))
  { if (n < 3) se = 1; else
    { printf("#2#REG_SET %s",p[1]);
      for (i=2; i<n; i++) printf(" %s",p[i]);
      printf("\n");
    }
  } else se = 1;
  if (se) return cmd_site_reg_syntax(501);
  if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
  return send_reply_exp(*str == '+' ? 200 : 550,str+1,0);
}

int cmd_site_run_syntax(int rno)
{ return send_reply(rno,"Syntax: RUN <script> [<parameters>]\n"
    "Runs a user script.");
}

int cmd_site_run_script(char** p, int n, int display_exit)
{ char* c;
  int   err;
  FILE* file;
  pid_t pid;
  int   pipefd[2];
  char  str[256];
  char  str2[256];
  char envbuffer[6][20];
  char* temp;

  if (!superuser() && display_exit) 
    return send_reply(550,"Only superusers may use this command.");
  if (!n) return cmd_site_run_syntax(501);
  err = socketpair(AF_UNIX,SOCK_STREAM,0,pipefd);
  if (err == -1) return send_error(550,"pipe",0);

  sprintf(envbuffer[0],"F4A_USER=%s",usr.name);
  putenv(envbuffer[0]);
  sprintf(envbuffer[1],"F4A_CLASS=%c",class_char());
  putenv(envbuffer[1]);
  sprintf(envbuffer[2],"F4A_INVISIBLE=%c",invisible() ? 'i' : '-');
  putenv(envbuffer[2]);
  sprintf(envbuffer[3],"F4A_COLOR=%c",coloruser() ? 'c' : '-');
  putenv(envbuffer[3]);
  sprintf(envbuffer[4],"F4A_GROUP=%s",get_groupname(usr.gid));
  putenv(envbuffer[4]);
  sprintf(envbuffer[5],"F4A_APASS=%s",anonpass);
  putenv(envbuffer[5]);
  temp = server_dir + strlen(server_dir)+9;
  strcpy(temp,"F4A_STD=");strcat(temp,server_dir);
  putenv(temp);

  pid = fork();
  if (pid == -1) return send_error(550,"fork",errno);
  if (pid)
  { close(pipefd[1]);
    file = fdopen(pipefd[0],"rt");
    if (!file) return send_reply(550,"fdopen");
    if (display_exit) c = (char*)-1; else
    { c = fgets(str,sizeof(str)-1,file);
      if (!c) strcpy(str,"(no output from program)");
    }
    while (c)
    { c = fgets(str2,sizeof(str2)-1,file);
      if (c && strlen(str2))
      { if (display_exit)
        { if (send_message(200,0,str2,NULL) == -1) return -1;
        } else
        { if (send_message(200,0,str,NULL) == -1) return -1;
          strcpy(str,str2);
        }
      }
    }
    fclose(file);
    close(pipefd[0]);
    pid = wait(&err);
    if (WIFEXITED(err))
    { if (WEXITSTATUS(err) == 255) sprintf(str,"Script does not exist.");
        else if (display_exit)
          sprintf(str,"Script finished with exitcode %i.",WEXITSTATUS(err));
    } else if (display_exit)
    { sprintf(str,"Script terminated on signal %i.",WTERMSIG(err));
    }
    return send_reply(200,str);
  } else
  { fclose(sockf);
    if (strcmp(cwd,"/")) chdir(cwd+1);
    close(pipefd[0]);
    err = dup2(pipefd[1],1);
    if (err != -1) err = dup2(pipefd[1],2);
    if (err != -1) err = dup2(pipefd[1],0);
    close(pipefd[1]);
    *(p+n) = NULL;
    if (err != -1)
    { strcpy(str,server_dir);
      strcat(str,DFL_BINDIR);
      strcat(str,*p);
      execv(str,p);
      strcpy(str,server_dir);
      strcat(str,DFL_BINDIR);
      c = get_groupname(usr.gid);
      if (c)
      { strcat(str,c);
        strcat(str,"/");
        strcat(str,*p);
        err = execv(str,p);
      }
    }
    exit(-1);
  }
}

int cmd_site_run(char** p, int n)
{ return cmd_site_run_script(p,n,1);
}

int cmd_site_shutdown_syntax(int rno)
{ return send_reply(rno,
    "Syntax: SHUTDOWN [{normal|transactional|immediate}]\n"
    "Force all connections to close and ftp server process to terminate.\n"
    "Note that you will be disconnected immediately, too.");
}

int cmd_site_shutdown(char** p, int n)
{ char str[128];

  if (n > 1) return cmd_site_shutdown_syntax(501);
  if (n)
  { if (!strcasecmp(p[0],"normal")) sprintf(str,"sn"); else
      if (!strcasecmp(p[0],"transactional")) sprintf(str,"st"); else
      if (!strcasecmp(p[0],"immediate")) sprintf(str,"si"); else
      return cmd_site_shutdown_syntax(501);
  } else sprintf(str,"si");
  printf("SRV_SST %s\n",str);
  if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
  if (*str == '-') return send_reply(550,str+1);
  quit(CL_SHUTDOWN);
  return send_reply(421,str+1);
}

int cmd_site_user_syntax(int rno)
{ return send_reply(rno,
    "Implemented USER commands, syntax: SITE USER <command> [<arguments>]\n"
    "                               or: SITE USR <command> [<arguments>]\n"
    "ADD <name> [<template>]                 - add user\n"
    "CHANGE <name> NAME=<s>|UID=<n>|GID=<n>|LIMIT=<n>|IP[0-9]=<r>|\n"
    "              INFO=<s>|CREDIT[+|-]=<n>|MBCRED[+|-]=<n>|\n"
    "              MBUP/MBDOWN [+|-] = <megs> for total stats\n"
    "              MUP/MDOWN   [+|-] = <megs> for month stats\n"
    "              WKUP/WKDOWN [+|-] = <megs> for week stats\n"
    "              YUP/YDOWN   [+|-] = <megs> for year stats\n"
    "              CLASS=[a|g|n|o|r|s|t]|LOGINS=<n>|FLAGS[+|-]={c|d|i|p|*|-}|\n"
    "              UL_MIN=<n>|UL_MAX=<n>|DL_MIN=<n>|DL_MAX=<n>|PASSWD=<s>\n"
    "                                        - change user attributes\n"
    "LIST [<options>] [<user>|<group>]       - list all or specified user(s)\n"
    "LISTA                                   - list all logged in users\n"
    "REMOVE <name>                           - remove user\n"
    "RESET {up|down|credit|all} [{week|month|year|total|all}]\n"
    "                                        - resets counter for all users");
}

int cmd_site_user(char** p, int n)
{ int  se = 0;
  char str[1024];

  if (!n) se = 1; else if (!strcasecmp(*p,"ADD"))
  { if ((n < 2) || (n > 3)) se = 1; else
      printf("USR_ADD %s %s\n",p[1],n == 3 ? p[2] : "");
  } else if (!strcasecmp(*p,"CHANGE"))
  { if (n < 3) se = 1; else
    { printf("USR_CHG %s %s",p[1],p[2]);
      n -= 3;
      p += 3;
      while (n--)
      { printf(" %s",*p);
        p++;
      }
      printf("\n");
    }
  } else if(!strcasecmp(*p,"LIST") || !strcasecmp(*p,"LISTA"))
  { if (n > 3) se = 1; else
    { if (!strcasecmp(*p,"LIST"))
        printf("ULI %s %s\n",n>=2 ? *(p+1) : "", n==3 ? *(p+2) : "");
      else printf("ULL %s %s\n",n==2 ? *(p+1) : "", n==3 ? *(p+2) : "");
      if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
      while (*str=='+' && strcmp(str,"+\n"))
      { if (send_message(211,0,str+1,NULL) == -1) return -1;
        if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-");
      }
      if (*str == '+') return send_reply(211,"<end of list>");
      if (strcmp(str,"-")) return send_reply(550,str+1);
        else return send_reply(211,"pipe error, aborting list");
    }
  } else if (!strcasecmp(*p,"REMOVE"))
  { if (n != 2) se = 1; else printf("USR_RM %s\n",p[1]);
  } else if (!strcasecmp(*p,"RESET"))
  { if ((n < 2) || (n > 3)) se = 1; else printf("USR_RST %s %s\n",p[1],
      n == 3 ? p[2] : "");
  } else se = 1;
  if (se) return cmd_site_user_syntax(501);
  if (!fgets(str,sizeof(str),stdin)) strcpy(str,"-pipe error");
  return send_reply(*str == '+' ? 200 : 550,str+1);
}

int cmd_site_w_syntax(int rno)
{ return send_reply(rno,
    "Syntax: SITE W\n"
    "    or: SITE WHO\n" 
    "Lists user activity.");
}

int cmd_site_w(char** p, int n)
{ char  str[512];
  char  str2[512];
  char* strptr;

  if (n) return cmd_site_w_syntax(501);
  printf("WHO\n");
  strptr = str;
  if (!fgets(strptr,sizeof(str),stdin))
    strcpy(strptr,"pipe error, aborting list");
  if (*strptr == '-') return send_reply(550,strptr+1);
  n = 0;
  while (strcmp(strptr,"+\n"))
  { strptr = (strptr == str) ? str2 : str;
    if (!fgets(strptr,sizeof(str),stdin))
      strcpy(strptr,"-pipe error, aborting list");
    if (send_message(211,!strcmp(strptr,"+\n"),
      strptr == str ? str2+1 : str+1,NULL) == -1) return -1;
    if (*strptr == '-') return send_reply(211,strptr+1);
    n++;
  }
  return n ? 0 : send_reply(211,"<empty list>");
}

int cmd_site_help_syntax(int rno)
{ if (superuser() || groupop() || siteop()) return send_reply(rno,
    "Implemented SITE commands, syntax: SITE <command> [<argument(s)>]\n"
    "ADDUSER   BIND      CHMOD     CHMODR    CHOWN     CLOSE     COLOR\n"
    "COUNT     DEL       DISABLE   DUMP      ENABLE    EXEC      FF\n"
    "GROUP     GROUPS    HELP      IDLE      INFO      KICK      MOD\n"
    "OPEN      PASSWD    REG       RUN       SHUTDOWN  USR/USER  VERSION\n"
    "W/WHO\n"
    "Type SITE HELP <command> to get more help.");
  return send_reply(rno,
    "Implemented SITE commands, syntax: SITE <command> [<argument(s)>]\n"
    "CHMOD     CHMODR    COLOR     COUNT     FF        GROUPS    HELP\n"
    "IDLE      INFO      MOD       PASSWD    RUN       VERSION   W/WHO\n"
    "Type SITE HELP <command> to get more help.");
}

int cmd_site_ff(char**,int);
int cmd_site_ff_syntax(int);
int cmd_site_help(char**,int);

struct sitecmd_s
{ char command[10];
  int  (*function)(char**,int);
  int  (*syntaxfnc)(int);
} site_command[] =
{ { "ADDUSER",  cmd_site_adduser,  cmd_site_adduser_syntax  },
  { "BIND",     cmd_site_bind,     cmd_site_bind_syntax     },
  { "CHMOD",    cmd_site_chmod,    cmd_site_chmod_syntax    },
  { "CHMODR",   cmd_site_chmodr,   cmd_site_chmodr_syntax   },
  { "CHOWN",    cmd_site_chown,    cmd_site_chown_syntax    },
  { "CLOSE",    cmd_site_close,    cmd_site_close_syntax    },
  { "COLOR",    cmd_site_color,    cmd_site_color_syntax    },
  { "COUNT",    cmd_site_count,    cmd_site_count_syntax    },
  { "DEL",      cmd_site_del,      cmd_site_del_syntax      },
  { "DISABLE",  cmd_site_disable,  cmd_site_disable_syntax  },
  { "DUMP",     cmd_site_dump,     cmd_site_dump_syntax     },
  { "ENABLE",   cmd_site_enable,   cmd_site_enable_syntax   },
  { "EXEC",     cmd_site_exec,     cmd_site_exec_syntax     },
  { "FF",       cmd_site_ff,       cmd_site_ff_syntax       },
  { "GROUP",    cmd_site_group,    cmd_site_group_syntax    },
  { "GROUPS",   cmd_site_groups,   cmd_site_groups_syntax   },
  { "HELP",     cmd_site_help,     cmd_site_help_syntax     },
  { "IDLE",     cmd_site_idle,     cmd_site_idle_syntax     },
  { "INFO",     cmd_site_info,     cmd_site_info_syntax     },
  { "KICK",     cmd_site_kick,     cmd_site_kick_syntax     },
  { "MOD",      cmd_site_mod,      cmd_site_mod_syntax      },
  { "OPEN",     cmd_site_open,     cmd_site_open_syntax     },
  { "PASSWD",   cmd_site_passwd,   cmd_site_passwd_syntax   },
  { "REG",      cmd_site_reg,      cmd_site_reg_syntax      },
  { "RUN",      cmd_site_run,      cmd_site_run_syntax      },
  { "SHUTDOWN", cmd_site_shutdown, cmd_site_shutdown_syntax },
  { "USER",     cmd_site_user,     cmd_site_user_syntax     },
  { "USR",      cmd_site_user,     cmd_site_user_syntax     },
  { "VERSION",  cmd_site_version,  cmd_site_version_syntax  },
  { "W",        cmd_site_w,        cmd_site_w_syntax        },
  { "WHO",      cmd_site_w,        cmd_site_w_syntax        } };

int cmd_site_help(char** p, int n)
{ int i = 0;
  int j;
  int k = (sizeof(site_command) / sizeof(struct sitecmd_s)) - 1;
  int r;

  if (n != 1) return cmd_site_help_syntax(501);
  while (i <= k)
  { j = (i + k) >> 1;
    r = strcasecmp(*p,site_command[j].command);
    if (r < 0) k = j - 1; else if (r > 0) i = j + 1; else
      return site_command[j].syntaxfnc(214);
  }
  return cmd_site_help_syntax(501);
}

int cmd_site(char** p, int n)
{ int   i = 0;
  int   j;
  int   k = (sizeof(site_command) / sizeof(struct sitecmd_s)) - 1;
  char* q;
  int   r;
  char* s;

  if (!n) return cmd_site_help_syntax(501);
  if (shortcut && strcasecmp(*p,"SITE"))
  { r = strlen(*p);
    q = shortcut;
    do
    { skipspace(&q);
      s = q;
      skipnspace(&q);
      if ((r == q-s) && !strncasecmp(*p,s,r))
      { return cmd_site_run_script(p,n,0);
      }
    } while (*q);
  }
  if ((n >= 2) && !strcasecmp(*p,"SITE"))
  { p++;
    n--;
  }
  while (i <= k)
  { j = (i + k) >> 1;
    r = strcasecmp(*p,site_command[j].command);
    if (r < 0) k = j - 1; else if (r > 0) i = j + 1; else
    { return site_command[j].function(p+1,n-1);
    }
  }
  return send_reply(501,"Unknown SITE command.");
}

/* EOF */
